unit FastShareMem;

(*
 * Shared Memory Allocator for Delphi DLL's
 * Version: 1.01
 *
 * Features:
 *   no runtime dll required.
 *   no performance degradation.
 *
 * Usage:
 *   Must be the first unit listed in the project file's USES section
 *   for both dll and exe projects. If you install a memory manager for
 *   leak detection, it should be listed immediately after this unit.
 *
 * Author: Emil M. Santos
 *   You may use and modify this software as you wish, but this section
 *   must be kept intact. Please see Readme.txt for copyright and disclaimer.
 *
 * Send bugs/comments to fastsharemem@codexterity.com
 * On the web: http://www.codexterity.com

   Revision History:
   2002 Aug 14: Rewrote address-computation code to better match windows 98
                allocation. VirtualAlloc may round down requested address *twice*.
                Replaced ASSERTs with (lower-level) Win32 MessageBox calls.
                (Thanks to Darryl Strickland (DStrickland@carolina.rr.com))
 *)

interface

implementation
uses Windows;

const SignatureBytes1 = $BABE01234567FEED;
      SignatureBytes2 = $F00D76543210B00B;

type
    TMemMgrPack = record
         Signature1: int64;
         MemMgr: TMemoryManager;
         Signature2: int64;
    end;
    PMemMgrPack = ^TMemMgrPack;


    procedure ValidatePack( p: PMemMgrPack );
    var pid: DWORD;
    begin
         // use pid for additional safety;
         pid := GetCurrentProcessId;
         p^.Signature1 := SignatureBytes1 xor pid;
         p^.Signature2 := SignatureBytes2 xor pid;
    end;

    function IsPackValid( p: PMemMgrPack ): boolean;
    var pid: DWORD;
    begin
         pid := GetCurrentProcessId;
         Result := (p^.Signature1 = SignatureBytes1 xor pid) and
                   (p^.Signature2 = SignatureBytes2 xor pid);
    end;



var si: TSYSTEMINFO;
    OldMemMgr: TMemoryManager;
    requested, allocated: PMemMgrPack;

    err: integer;
    lpMsgBuf: PChar;
    granul: integer;
initialization
     GetSystemInfo( si );
     requested := si.lpMaximumApplicationAddress;
     requested := pointer( (integer(requested) div $10000) * $10000 ); // align on start of last full 64k page
     requested := pointer( (integer(requested) div si.dwPageSize) * si.dwPageSize ); // align on start of last full 64k page
     allocated := VirtualAlloc( requested, si.dwPageSize, MEM_RESERVE or MEM_COMMIT, PAGE_READWRITE );

     if allocated <> nil then
     begin
          if requested <> allocated then begin
              MessageBox( 0, 'Shared Memory Allocator setup failed: Address was relocated.', 'FastShareMem', 0 );
              Halt;
          end;

          GetMemoryManager( allocated^.MemMgr );
          ValidatePack( allocated );
     end
     else begin
          if not IsPackValid( requested ) then begin
              MessageBox( 0, 'Shared Memory Allocator setup failed: Address already reserved.', 'FastShareMem', 0 );
              Halt;
          end;

          GetMemoryManager( OldMemMgr );
          SetMemoryManager( requested^.MemMgr );
     end;

finalization
     (* cleanup *)
     if (allocated <> nil) and IsPackValid( requested ) then
     begin
          VirtualFree( allocated, 0, MEM_DECOMMIT or MEM_RELEASE );
     end
     else begin
          SetMemoryManager( OldMemMgr );
     end;
end.
