unit FileShortCut;

interface

uses
  Windows, Messages, SysUtils, Classes,Forms,Dialogs,DesignEditors,designIntf;

type
  TSelectEvent=procedure(Sender:TObject; Path:String) of object;

  TSCPathDialog = class(TComponent)
  private
    FHandle:THandle;
    FTitle:String;
    FDirectory:String;
    FShowStatus:Boolean;
    FStatusText:String;

    FOnShow:TNotifyEvent;
    FOnSelect:TSelectEvent;

    function getVisible:Boolean;
    procedure SetDirectory(Dir:String);
    procedure SetStatusText(Text:String);
  public
    { Shows the dialog. Returns false if the user clicked Cancel }
    { or if an error occurred }
    function Execute:Boolean;
    { Sets the status of the OK-Button to either enabled or disabled    }
    { This function should only be called in one of the 2 Eventhandlers }
    function setOKButton(enabled:Boolean):Boolean;
    { whether the dialog is visible }
    property visible:Boolean read getVisible;
    { Window Handle of the dialog is only valid when visible }
    { ...should not be used...}
    property Handle:THandle read FHandle;
  published
    { should be set before calling Execute }
    property Title:String read FTitle write FTitle;
    { is not valid while the Dialog is visible }
    property Directory:String read FDirectory write setDirectory;
    { must be set before executing the dialog }
    property ShowStatus:Boolean read FShowStatus write FShowStatus default false;
    { is only used if ShowStatus is true }
    property StatusText:String read FStatusText write SetStatusText;

    property Tag;
    // Events
    property OnShow:TNotifyEvent read FOnShow write FOnShow;
    property OnSelect:TSelectEvent read FOnSelect write FOnSelect;
  end;

  TSCTargetFilenameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TSCFilenameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TSCIcon = Class;

  TSCShortCutNameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TSCWorkingDirectoryProperty = class(TStringProperty)
  private
    FSDirOpen: TSCPathDialog;
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    procedure WDSelect(Sender: TObject; Path: String);
  end;


  TFileShortCut = class(TComponent)
  private
    { Private declarations }
    FPathObj:TFileName;
    FPathLink:TFileName;
    FDesc:string;
    FArguments:string;
    FWindowState:TWindowState;
    FWorkingDirectory:TFileName;
    FSCIcon:TSCIcon;
    procedure SetPathObj(value:TFileName);
    procedure SetSCIcon(value:TSCIcon);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(owner: TComponent); override;
    destructor Destroy; override;
    function execute:boolean;
  published
    { Published declarations }
    property Arguments:string read FArguments write FArguments;
    property Description:string read FDesc write FDesc;
    property ShortCutIcon:TSCIcon read FSCIcon write SetSCIcon;
    property ShortCutName:TFileName read FPathLink write FPathLink;
    property TargetFile:TFileName read FPathObj write SetPathObj;
    property WindowState:TWindowState read FWindowState write FWindowState;
    property WorkingDirectory:TFileName read FWorkingDirectory write FWorkingDirectory;

  end;

  TSCIcon = class(TPersistent)
    private
       FOwner:TFileShortCut;
       FIconIndex:integer;
       FIconLocation:TFileName;
       procedure SetIconLocation(value:TFileName);
       procedure SetIconIndex(value: integer);
    public
       constructor Create;
    published
       property IconLocation:TFileName read FIconLocation write SetIconLocation;
       property IconIndex:integer read FIconIndex write SetIconIndex default 0;
  end;

procedure Register;

implementation
uses ShlObj,ActiveX,menus,Controls,IconSelector;

const
  {$EXTERNALSYM IID_IPersistFile}
  IID_IPersistFile: TGUID = (
    D1:$0000010B;D2:$0000;D3:$0000;D4:($C0,$00,$00,$00,$00,$00,$00,$46));

// CreateLink - uses the shell's IShellLink and IPersistFile interfaces
//   to create and store a shortcut to the specified object.
// Returns the result of calling the member functions of the interfaces.
// lpszPathObj - address of a buffer containing the path of the object
// lpszPathLink - address of a buffer containing the path where the
//   shell link is to be stored
// lpszDesc - address of a buffer containing the description of the
//   shell link


function CreateLink(lpszPathObj,lpszPathLink,lpszDesc,lpszArguments,lpszWorkingDirectory,lpszIconLocation:string;iIcon:integer;iWindowState:TWindowState):HRESULT;
var
    hres:HRESULT;
    psl:IShellLink;
    ppf:IPersistFile;
    wsz:array [0..MAX_PATH] of WideChar;
    swWindowState:integer;
begin

    // Get a pointer to the IShellLink interface.
    hres := CoCreateInstance(CLSID_ShellLink, nil,
        CLSCTX_INPROC_SERVER, IID_IShellLinkA, psl);
    if (SUCCEEDED(hres)) then
    begin

        // Set the path to the shortcut target, and add the
        // description.
        psl.SetPath(PChar(lpszPathObj));
        swWindowState:=SW_SHOWNORMAL;
        case iWindowState of
             wsNormal:swWindowState:=SW_SHOWNORMAL;
             wsMaximized:swWindowState:=SW_SHOWMAXIMIZED;
             wsMinimized:swWindowState:=SW_SHOWMINIMIZED;
        end;
        psl.SetShowCmd(swWindowState);
        if lpszDesc <> '' then
           psl.SetDescription(PChar(lpszDesc));
        if lpszArguments <> '' then
           psl.SetArguments(PChar(lpszArguments));
        if lpszWorkingDirectory <> '' then
           psl.SetWorkingDirectory(PChar(lpszWorkingDirectory));
        if lpszIconLocation <> '' then
           psl.SetIconLocation(PChar(lpszIconLocation),iIcon);

       // Query IShellLink for the IPersistFile interface for saving the
       // shortcut in persistent storage.
        hres := psl.QueryInterface(IID_IPersistFile,ppf);

        if (SUCCEEDED(hres))then
        begin

            // Ensure that the string is ANSI.
            MultiByteToWideChar(CP_ACP, 0, PChar(lpszPathLink), -1,
                wsz, MAX_PATH);


            // Save the link by calling IPersistFile::Save.
            hres := ppf.Save(wsz, TRUE);
        end;
    end;
    result:=hres;
end;

function TFileShortCut.execute:boolean;
begin
     if FPathObj='' then
     begin
          result:=false;
//          ShowMessage('TFileShotcut: Target file not especified');
     end
     else if FPathLink='' then
     begin
          result:=false;
//          ShowMessage('TFileShotcut: Link filename not especified');
     end
     else
         result:=(CreateLink(FPathObj,FPathLink,FDesc,FArguments,FWorkingDirectory,FSCIcon.IconLocation,FSCIcon.IconIndex,FWindowState)=S_OK);
end;

constructor TFileShortCut.Create(owner: TComponent);
begin
     inherited Create(owner);
     FSCIcon:=TSCIcon.Create;
     FSCIcon.FOwner:=self;
     FWindowState:=wsNormal;
     FormIconSelector:=nil;
end;

destructor TFileShortCut.destroy;
begin
     FSCIcon.Free;
     inherited destroy;
end;

procedure TFileShortCut.SetPathObj(value:TFileName);
begin
     if FPathObj <> value then
     begin
          FPathObj:=value;
          if FWorkingDirectory = '' then
             WorkingDirectory:=ExtractFilePath(value);
     end;
end;

procedure TFileShortCut.SetSCIcon(value:TSCIcon);
begin
     FSCIcon.Assign(value);
end;

///////////////////////////////////////////////////

procedure TSCTargetFilenameProperty.Edit;
var
  FSFileOpen: TOpenDialog;

begin
  FSFileOpen := TOpenDialog.Create(Application);
  FSFileOpen.Filename := GetValue;
  FSFileOpen.Filter := 'All files (*.*)|*.*';
  FSFileOpen.Options := FSFileOpen.Options + [ofShowHelp, ofPathMustExist, ofFileMustExist];
  FSFileOpen.Title := 'Select a file';
  try
    if FSFileOpen.Execute then
      SetValue(FSFileOpen.Filename);
  finally
    FSFileOpen.Free;
  end;
end;

function TSCTargetFilenameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

///////////////////////////////////////////////////

procedure TSCFilenameProperty.Edit;
var
  FSFileOpen: TOpenDialog;
begin
  FSFileOpen := TOpenDialog.Create(Application);
  FSFileOpen.Filename := GetValue;
  FSFileOpen.DefaultExt:='lnk';
  FSFileOpen.Filter := 'All files (*.*)|*.*';
  FSFileOpen.Options := FSFileOpen.Options + [ofShowHelp, ofPathMustExist];
  FSFileOpen.Title := 'Select a file';
  try
    if FSFileOpen.Execute then
      SetValue(FSFileOpen.Filename);
  finally
    FSFileOpen.Free;
  end;
end;

function TSCFilenameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

///////////////////////////////////////////////////

procedure TSCShortCutNameProperty.Edit;
begin
     FormIconSelector:=TFormIconSelector.Create(Application);
     try
        if FormIconSelector.Execute(TSCIcon(GetComponent(0))) then
        begin
             SetValue(IntToStr(FormIconSelector.actual)+'#'+FormIconSelector.FileName);
        end;
     finally
            FormIconSelector.Free;
     end;
end;

function TSCShortCutNameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

///////////////////////////////////////////////////

procedure TSCWorkingDirectoryProperty.Edit;
begin
  FSDirOpen := TSCPathDialog.Create(Application);
  FSDirOpen.Directory := GetValue;
  FSDirOpen.StatusText:='Select working directory';
  FSDirOpen.Title := 'Select the working directory';
  FSDirOpen.ShowStatus:=true;
  FSDirOpen.OnSelect:=WDSelect;
  try
    if FSDirOpen.Execute then
      SetValue(FSDirOpen.Directory);
  finally
    FSDirOpen.Free;
  end;
end;

function TSCWorkingDirectoryProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

procedure TSCWorkingDirectoryProperty.WDSelect(Sender: TObject; Path: String);
begin
     FSDirOpen.StatusText:=Path;
end;

///////////////////////////////////////////////////////////
/// TSCPathDialog ///////////////////////////////////////////
///////////////////////////////////////////////////////////

function CallBack(Wnd: HWND; uMsg: UINT; lParam, lpData: LPARAM): Integer stdcall;
var dir:String;
begin
 try
  with TSCPathDialog(lpData) do
  case uMsg of
   BFFM_INITIALIZED:
    begin
     FHandle:=Wnd;
     SetDirectory(FDirectory);
     SetStatusText(FStatusText);
     try
      if assigned(FOnShow) then
       FOnShow(TSCPathDialog(lpData));
     except
      On e:Exception do
       ShowMessage(e.Message);
     end;
    end;
   BFFM_SELCHANGED: // lpParam is a pointer to the item identifier list for the newly selected folder.
    begin
     if assigned(FOnSelect) then
     try
      SetString(dir,nil,MAX_PATH);
      if SHGetPathFromIDList(PItemIDList(lParam),PChar(Dir)) then
       FOnSelect(TSCPathDialog(lpData),PChar(Dir))
      else
       FOnSelect(TSCPathDialog(lpData),'');
     except
      On e:Exception do
       ShowMessage(e.Message);
     end;
    end;
  end;
 except
 end;
 result:=0;
end;

function TSCPathDialog.Execute:Boolean;
var iList,Root:PItemIDList;
    bi:TBrowseInfo;
    DispName:String;
    malloc:IMalloc;
begin
 result:=false;
 if (Owner is TWinControl) then
  bi.hwndOwner:=TWinControl(Owner).Handle
 else bi.hwndOwner:=Application.Handle; //exit
 if SHGetSpecialFolderLocation(Handle, CSIDL_DRIVES,Root)=NOERROR then
 try
  SHGetMalloc(malloc);
  SetString(DispName,nil,MAX_PATH);
  with bi do
  begin
   pidlRoot := root;
   pszDisplayName := PChar(DispName);
   lpszTitle := PChar(FTitle);
   ulFlags:=BIF_RETURNONLYFSDIRS;
   if FShowStatus then
    ulFlags:=ulFlags or BIF_STATUSTEXT;
   lpfn:=@CallBack;
   lParam:=Integer(self);

  end;
  iList:=SHBrowseForFolder(bi);
  FHandle:=0;
  if iList<>nil then
  try
   if SHGetPathFromIDList(iList,PChar(DispName)) then
   begin
    FDirectory:=PChar(DispName);
    result:=true;
   end;
  finally
   malloc.Free(iList);
  end;
 finally
  malloc.Free(root);
 end;
end;

function TSCPathDialog.setOKButton(enabled:Boolean):Boolean;
begin
 result:=false;
 if (FHandle<>0) then
 begin
  result:=true;
  if enabled then
   SendMessage(FHandle,BFFM_ENABLEOK,1,0)
  else
   SendMessage(FHandle,BFFM_ENABLEOK,0,0);
 end;
end;

function TSCPathDialog.getVisible:Boolean;
begin
 result:=FHandle<>0;
end;

procedure TSCPathDialog.SetDirectory(Dir:String);
begin
 if (Dir<>'') and (Dir[length(Dir)]='\') then
  FDirectory:=copy(Dir,1,length(Dir)-1)
 else
  FDirectory:=Dir;
 if (FHandle<>0) and (FDirectory<>'') then
  SendMessage(FHandle,BFFM_SETSELECTION,Integer(LongBool(true)),Integer(PChar(FDirectory)));
end;

procedure TSCPathDialog.SetStatusText(Text:String);
begin
 FStatusText:=Text;
 if (FHandle<>0) and FShowStatus then
  SendMessage(FHandle,BFFM_SETSTATUSTEXT,0,Integer(PChar(FStatusText)));
end;

/////////////////////////////////////////////////////////
// TSCIcon //////////////////////////////////////////////
/////////////////////////////////////////////////////////

constructor TSCIcon.Create;
begin
     inherited;
     FIconIndex:=0;
     FIconLocation:='';
end;

procedure TSCIcon.SetIconLocation(value:TFileName);
var
   idx:integer;
begin
     if Pos('#',value) > 0 then
     begin
          idx:=StrToInt(Copy(value,1,Pos('#',value)-1));
          delete(value,1,Pos('#',value));
          FIconLocation:=value;
          FIconIndex:=idx;

     end
     else
         if value <> FIconLocation then
         begin
              FIconLocation:=value;
              FIconIndex:=0;
         end;
end;

procedure TSCIcon.SetIconIndex(value: integer);
begin
     if FIconIndex <> value then
        FIconIndex:=value;
end;
    
procedure Register;
begin
  RegisterComponents('Win32', [TFileShortCut]);
  RegisterPropertyEditor(TypeInfo(TFileName), TFileShortCut, 'TargetFile', TSCTargetFileNameProperty);
  RegisterPropertyEditor(TypeInfo(TFileName), TFileShortCut, 'ShortCutName', TSCFileNameProperty);
  RegisterPropertyEditor(TypeInfo(TFileName), TFileShortCut, 'WorkingDirectory', TSCWorkingDirectoryProperty);
  RegisterPropertyEditor(TypeInfo(TFileName), TSCIcon, 'IconLocation', TSCShortCutNameProperty);
end;

end.
