// Direct64 Unit, part of Graph64 project
// Interface between Graph64 and others targets:
// DirectX(TDirectSurface)
// DelphiX required

// Code by: Pavol Stugel
// (c) NETGraphics 2002

unit Direct64;

interface

uses
  Graphics,
  DirectX, DXDraws, DXClass,
  SysUtils, Types, Graph64;

type
  TBufferType = (btSelf, btLinkedBitmap);
  TScreen64 = class(TBitmap64)
  private
    FSurface: TDirectDrawSurface;
    FWasAssigned: Boolean;
    FLockedSurface: IDirectDrawSurface4;
    FLockedSurfaceDesc: TDDSurfaceDesc2;
    FLockedRect: TRect;

  public
    function GetScanLine(Row: Integer): Pointer;override;
    constructor Create;
    procedure LinkToSurface(Surface: TDirectDrawSurface);
    destructor Destroy;override;

    function Lock: Boolean;
    procedure UnLock;
  end;


implementation

{ TScreen64 }

constructor TScreen64.Create;
begin
  FWasAssigned := False;
  inherited Create;
end;

destructor TScreen64.Destroy;
begin
  inherited;
end;

function TScreen64.GetScanLine(Row: Integer): Pointer;
begin
  Result :=  Pointer(Integer(FLockedSurfaceDesc.lpsurface) +
    Row * FLockedSurfaceDesc.lpitch);
end;

procedure TScreen64.LinkToSurface(Surface: TDirectDrawSurface);
var
  pf: Integer;
begin
    pf := Surface.BitCount;
    if (pf = 8) or {(pf=16) or} (pf = 24) then
      Error(Format('Sorry this %d pixel format isn''t supported!',[pf]))
    else
    begin
      // now it's supported only 15/32 bit bitmaps
      PixelFormat := pf32bit;
      if pf = 32 then PixelFormat := pf32bit
      else
        PixelFormat := pf15bit;
      try
        FSurface := Surface;
        FLockedSurface := Surface.ISurface4;
        FWidth := Surface.Width;
        FHeight := Surface.Height;

        ReallocateBuffer; //realocate filter buffer
        ClipRect := Rect(0, 0, FWidth, FHeight);
      except
        Error('Lost bitmap!');
      end;
      FWasAssigned := True;
    end
end;

function TScreen64.Lock: Boolean;
begin
  FLockedSurface := FSurface.ISurface4;
  FLockedSurfaceDesc.dwSize := SizeOf(TDDSurfaceDesc);
  FLockedRect := Rect(0, 0, FLockedSurfaceDesc.dwWidth, FLockedSurfaceDesc.dwHeight);
  if FLockedSurface.Lock(@FLockedRect, FLockedSurfaceDesc,
    DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT, 0) <> DD_OK then Result := False
  else Result := True;
end;

procedure TScreen64.UnLock;
begin
  FLockedSurface.Unlock(@FLockedRect);
  FLockedSurface := nil;
end;

end.

