// Graph64 GE Unit
// Code by: Pavol Stugel
// (c) Pavol Stugel 1999-2002
//
// Version GE 0.1f (snapshot version) last update: 10.august 2002
// http://www.graph64.netgraphics.sk
// e-mail:// graph64@netgraphics.sk
//
// If you have any ideas to add some features into this unit you are welcome!

{
  About:
  Graph64 is graphics library :)
  chceck homepage for more infos

  History:

  27.feb.2002 - idea to rewrite my very old Graph64 library;)
}

unit Graph64;

interface

//if you have problems with compiling this library enable this feature:
{ $define oldcompiler}

uses
  {$ifndef oldcompiler}
  Types,
  {$endif}
   SysUtils;

const
  // blending and filtering methods
  bmCopy: Integer = $0001;
  bmLens: Integer = $0002;
  bmBlend: Integer =$0003;
  bmBlend_Sub: Integer =$0004;
  bmAlpha: Integer = $0005;
  bmAlphaTransform: Integer = $0006;
  bmGrayScale: Integer = $0007;
  bmLightness: Integer = $0008;
  bmDarkness: Integer = $0009;
  bmDisolve: Integer = $000A;
  bmShadow: Integer = $000B;
  bmPixShadow: Integer = $000C;
  bmAxialBlend: Integer = $000D;
  

  fmNone: Integer = $0000;
  fmDisolve: Integer = $0001;
  fmAlphaTransform: Integer = $0002;

type

  {$ifdef oldcompiler}
  PCardinal = ^TCardinal;
  PPoint = ^TPoint;
  TPoint = packed record
    X: Longint;
    Y: Longint;
  end;

  PRect = ^TRect;
  TRect = packed record
    case Integer of
      0: (Left, Top, Right, Bottom: Longint);
      1: (TopLeft, BottomRight: TPoint);
  end;
  {$endif}


  EGraph64 = class(Exception);

  TPixelFormat = (pf8bit, pf15bit, pf16bit, pf24bit, pf32bit);
  TAlphaFormat = (afNone, afPixel, afSelf);

  TOtherInfo = packed record
    X, Y,             // bitmap top left position
    tX, tY: Integer;  // true position of curent line
                      // tX - X = first painted position in Bitmap
                      // ... same for Y
    Width,
    Height: Integer;
  end;

  POtherInfo = ^TOtherInfo;

  TMoveMemProc = procedure (SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
    
  TFillMemProc = procedure (DestationPtr: Pointer; Size: Integer; Color: Integer;
    OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
    {
      Note about AlphaFormat:
      afNone - no alpha channel presents
      afPixel - only for pf32bit (alpha is last byte in each pixel (RGBA))
      afSelf - means alpha is stored in separate memory (8bit)
    }
  TBitmap64 = class;
      
  TBlendSpec = packed record
    sfactor, dfactor: Single;
  end;

  TTransformSpec = packed record
    alpha : TBitmap64;
    frame : Byte;
    x,y: Integer;  //top left pixel of start
  end;

  TRGBSpec = packed record
    Red, Green, Blue: Word;
  end;

  TAxialSpec = packed record
    AStart,
    AMiddle,
    AEnd: Byte;
  end;

  TScanLine = function (Row: Integer): Pointer;
  TBlendMode = Integer;
  TPaintMode = Integer;

  TBitmap64 = class(TObject)  {declaration for TBitmap64}

  private

    FWasAssigned: Boolean;
    FAlphaFormat: TAlphaFormat;  //alpha format of bitmap
    FPixelFormat: TPixelFormat;  //pixel format of bitmap
    FClipRect: TRect;
    FCenter: TPoint;

    FFilterMode,
    FBlendMode: TBlendMode;  //drawing mode
    FPaintMode: TPaintMode;  //filling mode

    FFillData,
    FCopyData,
    FFilterData: Pointer;  //memory for other specific variables

    FFilterMem: Pointer;  //memory for paint filter

    RBitMask, GBitMask, BBitMask,
    ABitMask: Cardinal;
    RStart, GStart, BStart,
    AStart: Byte;

    bmBits: Pointer;
    bmBpp: Integer;

    TopDown: Boolean;  //True, if bitmap have top down orientation
    linePitch: Integer;  //line size in bytes

    FFillMemProc: TFillMemProc;
    FFilterMemProc,
    FMoveMemProc: TMoveMemProc;
    BlendFilter: Boolean;    

    procedure SetupBits;
    procedure SetupBitmap(Width, Height: Integer; PixelFormat: TPixelFormat;
      AlphaFormat: TAlphaFormat);

    procedure SetPixelFormat(Value: TPixelFormat);
    procedure SetSize(NewWidth, NewHeight: Integer);
    procedure SetClipper(r: TRect);
    procedure SetPixel(X, Y, Color: Integer);
    function  GetPixel(X, Y: Integer): Integer;
    procedure SetBlendMode(BlendMode: TBlendMode);
    procedure SetPaintMode(PaintMode: TPaintMode);
    procedure SetBlendFilterMode(FilterMode: TBlendMode);

    procedure SetBlendParams(Params: Pointer);
    procedure SetPaintParams(Params: Pointer);

    procedure SetBlendFilterParams(Params: Pointer);

    function Clip(const bitmap: TBitmap64; var x,y, SizeX, SizeY, StartX,
      StartY: Longint): Boolean;


    procedure FillMemProc(DestationPtr: Pointer; Size: Integer; Color: Integer;
      OtherData: Pointer; OtherInfo: POtherInfo);

    procedure MoveMemProc(SourcePtr, DestationPtr: Pointer; Size: Integer;
      TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
      OtherInfo: POtherInfo);
    procedure SetupFilterMemory;


  public

    Transparent: Boolean;
    TransparentColor,
    FWidth, FHeight: Integer; // 1x1 == one x one pixels
    Blend: TBlendSpec;


    function GetScanLine(Row: Integer): Pointer;virtual;

    constructor Create;overload;
    constructor Create(Width, Height: Integer; PixelFormat: TPixelFormat;
      AlphaFormat: TAlphaFormat);overload;

    destructor Destroy;override;

    property BlendMode: TBlendMode read FBlendMode write SetBlendMode;
    property PaintMode: TPaintMode read FPaintMode write SetPaintMode;
    property BlendFilterMode: TBlendMode read FFilterMode write SetBlendFilterMode;

    property BlendParams: Pointer read FCopyData write SetBlendParams;
    property PaintParams: Pointer read FCopyData write SetPaintParams;
    property BlendFilterParams: Pointer read FFilterData write SetBlendFilterParams;
    property ClipRect: TRect read FClipRect write SetClipper;
    property Pixels[X, Y: Integer]: Integer read GetPixel write SetPixel;
    property ScanLine[Row: Integer]: Pointer read GetScanLine;
    property Center: TPoint read FCenter write FCenter;
    property PixelFormat: TPixelFormat read FPixelFormat write SetPixelFormat;
    property Width: Integer read FWidth;
    property Height: Integer read FHeight;


    //for internal use
    procedure ReallocateBuffer;
    procedure MoveCenter(RelativeX, RelativeY: Integer);

    //functions primitives 2D, painting using fill procedures

    function  rgb(r, g, b: Byte; a: Byte = 0): Integer;
    procedure HLine(X, Y, X1: Integer; Color: Integer);
    procedure VLine(X, Y, Y1: Integer; Color: Integer);
    procedure Line(X, Y, X1, Y1: Integer; Color: Integer; First: Boolean = True);
    procedure Polyline(const Points: array of TPoint; Color: Integer;
      StartIndex: Integer = 0; NumPts: Integer = -1);
    procedure Rectangle(X, Y, X1, Y1: Integer; Color: Integer);overload;
    procedure Rectangle(const Rect: TRect; Color: Integer);overload;
    procedure FillRect(const Rect: TRect; Color: Integer);
    procedure Ellipse(x1, y1, x2, y2, Color: Integer);
    procedure FillEllipse(x1, y1, x2, y2, Color: Integer);
    procedure Triangle(x1, y1, x2, y2, x3, y3,  Color: Integer);
    procedure FillTriangle(x1, y1, x2, y2, x3,y3, Color:Integer);

    //functions drawing / bliting, painting using move procedures
    procedure Draw(x, y: Integer; b: TBitmap64);

  end;

  procedure RegisterBlendMode(Mode: Integer;
    BlendFunc8, BlendFunc15, BlendFunc16, BlendFunc24, BlendFunc32: TMoveMemProc;
    FillFunc8, FillFunc15, FillFunc16, FillFunc24, FillFunc32: TFillMemProc);

  procedure Error(const s: String);

const
  g64DefaultPixelFormat = pf32bit; //default pixel and alpha format
  g64DefaultAlphaFormat = afSelf;  //unused yet
  DisolveTempDataSize = 1600*5;    //buffer for disolve effect

const
  //defines maximal number of modes
  MaxBlendModes = 60;
  MaxFilterModes = 60;
  MaxFillModes = 60;

var
  // 8, 15, 16, 24, 32
  BlendModes:  array[0..4,0..MaxBlendModes] of TMoveMemProc;
  FilterModes: array[0..4,0..MaxFilterModes] of TMoveMemProc;
  FillModes: array[0..4,0..MaxFillModes] of TFillMemProc;

  DisolveTempData: PByte;
  
  _p: PByte;
  _i: Integer;

function DetectCPU: Integer;

implementation

{$ifdef oldcompiler}
  function Rect(Left, Top, Right, Bottom: Integer): TRect;
  begin
    Result.Left := Left;
    Result.Top := Top;
    Result.Bottom := Bottom;
    Result.Right := Right;
  end;

  function Point(X, Y: Integer): TPoint;
  begin
    Result.X := X;
    Result.Y := Y;
  end;

{$endif}



  procedure Error(const s: String);
  begin
    raise EGraph64.Create('Graph64: ' + s);
  end;

  procedure SwapI(var x, y: Integer);
  asm
    mov ecx, [eax]
    xchg ecx, [edx]
    mov [eax], ecx
  end;

  procedure SwapMinI(var x, y: Integer);
  asm
    mov ecx, [eax]
    cmp ecx, [edx]
    jle @finish
    xchg ecx, [edx]
    mov [eax], ecx
  @finish:
  end;

  procedure PlusRect(var Rect: TRect; Point: TPoint);
  begin
    Rect.Left := Rect.Left + Point.X;
    Rect.Top := Rect.Top + Point.Y;
    Rect.Right := Rect.Right + Point.X;
    Rect.Bottom := Rect.Bottom + Point.Y;
  end;

// --------- core low-level functions --------------
procedure FillMem_8(DestationPtr: Pointer; Size: Integer; Color: Integer; OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
// TODO: FillMem_8 need optimize for rep stosd,sw
begin
  asm
    push edi
    mov edi, DestationPtr
    mov ecx, Size
    mov eax, color
    rep stosb
    pop edi
  end;
end;

procedure FillMem_15_16(DestationPtr: Pointer; Size: Integer; Color: Integer; OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
// TODO: FillMem_15_16 need optimize for rep stosd
begin
  asm
    push edi
    mov edi, DestationPtr
    mov ecx, Size
    mov eax, color
    rep stosw
    pop edi
  end;
end;

procedure FillMem_32(DestationPtr: Pointer; Size: Integer; Color: Integer;
 OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
begin
  asm
    push edi
    mov edi, DestationPtr
    mov ecx, Size
    mov eax, color
//    add eax, y       //only test
    rep stosd
    pop edi
  end;
end;

procedure FillMem_Lens_15(DestationPtr: Pointer; Size: Integer; Color: Integer;
 OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
begin
  if Size > 0 then
  asm
    push edi
    mov edi, DestationPtr
    mov ecx, Size

    mov edx, Color
    shr dx, 1
    push ebx
    mov bx, 0011110111101111b //565 format
    and dx, bx //compute this color only once

  @Loop1:
    mov ax, [edi]
    shr ax, 1
    and ax, bx
    add ax, dx
    stosw
    loop @Loop1

    pop ebx
    pop edi
  end;
end;

procedure FillMem_Lens_16(DestationPtr: Pointer; Size: Integer; Color: Integer;
 OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
begin
  if Size > 0 then
  asm
    push edi
    mov edi, DestationPtr
    mov ecx, Size

    mov edx, Color
    shr dx, 1
    push ebx
    mov bx, 1111101111101111b //565 format
    and dx, bx //compute this color only once

  @Loop1:
    mov ax, [edi]
    shr ax, 1
    and ax, bx
    add ax, dx
    stosw
    loop @Loop1

    pop ebx
    pop edi
  end;
end;

procedure FillMem_Lens_32(DestationPtr: Pointer; Size: Integer; Color: Integer;
  OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
// TODO: FillMem_Lens need to test on linux due to EBX register
begin
  if Size > 0 then
  asm
    push edi
    mov edi, DestationPtr
    mov ecx, Size

    mov edx, Color
    shr edx, 1
    push ebx
    mov ebx, 011111110111111101111111b
    and edx, ebx //compute this color only once

  @Loop1:
    mov eax, [edi]
    shr eax, 1
    and eax, ebx
    add eax, edx
    stosd
    loop @Loop1

    pop ebx
    pop edi
  end;
end;

procedure FillMem_Blend_15(DestationPtr: Pointer; Size: Integer; Color: Integer;
  OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
var
  xx, sfact, dfact: Integer;
  p: PSingle;
  pb: PWord;
  r,g,b,
  rr,gg,bb: Word;
begin

  if Size <= 0 then exit;
  if not Assigned(OtherData) then
  begin
    Error('FillMem_Blend_32 without SourceFactor and DestationFactor parameter!');
    exit;
  end;

  p := OtherData;
  sfact := trunc(p^ * 128);Inc(p);
  dfact := trunc(p^ * 128);
  if (sfact = 64) and (dfact = 64) then
  begin
    FillMem_Lens_15(DestationPtr, Size, Color, Nil);
    Exit;
  end;

  r := sfact * ((Color and $7c00) shr 10);
  g := sfact * ((Color and $03e0) shr 5);
  b := sfact * (Color and $001f);

  pb := DestationPtr;
  for xx := 0 to Size - 1 do
  begin
    rr := (r + ((pb^ and $7c00) shr 10) * dfact) shr 7;
    gg := (g + ((pb^ and $03e0) shr  5) * dfact) shr 7;
    bb := (b + (pb^ and $001f)  * dfact) shr 7;

    if rr > 31 then rr:= 31;
    if gg > 31 then gg:= 31;
    if bb > 31 then bb:= 31;

    pb^ := (rr shl 10) or (gg shl 5) or bb;
    Inc(pb); //goto next pixel
  end;
end;

procedure FillMem_Blend_32(DestationPtr: Pointer; Size: Integer; Color: Integer;
  OtherData: Pointer; X: Integer = 0; Y: Integer = 0);
var
  xx, sfact, dfact: Integer;
  p: PSingle;
  pb: PByte;
  r,g,b,
  rr,gg,bb: Word;
  pi: PCardinal;
begin
  if Size <= 0 then exit;
  if not Assigned(OtherData) then
  begin
    Error('FillMem_Blend_32 without SourceFactor and DestationFactor parameter!');
    exit;
  end;
  p := OtherData;
  sfact := trunc(p^ * 128);Inc(p);
  dfact := trunc(p^ * 128);
  if (sfact = 64) and (dfact = 64) then
  begin
    Fillmem_Lens_32(DestationPtr, Size, Color, Nil);
    exit;
  end;
  r := sfact * ((Color shr 16) and $ff);
  g := sfact * ((Color shr 8) and $ff);
  b := sfact * ((Color) and $ff);

  pb := DestationPtr;
  for xx := 0 to Size - 1 do
  begin
    bb:= (b + pb^ * dfact) shr 7;Inc(pb);
    gg:= (g + pb^ * dfact) shr 7;Inc(pb);
    rr:= (r + pb^ * dfact) shr 7;Inc(pb);
    Inc(pb); //skip alpha

    if rr > 255 then rr:= 255;
    if gg > 255 then gg:= 255;
    if bb > 255 then bb:= 255;
    asm mov eax, pb; sub eax, 4; mov pi, eax; end;
    pi^ := (rr shl 16) or (gg shl 8) or bb;
  end;
end;

/////////////////////////////////// moving func ////////////////////////////////
procedure MoveMem_8(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
var
  i: Integer;
  p,p1: PByte;
begin
  if Size < 0 then exit;
  if (TransFilter and 1) = 1 then
  begin
    p := SourcePtr;
    p1 := DestationPtr;
    for i := 0 to Size - 1 do
    begin
      if p^ = TransColor then
      else p^ := p1^;
      Inc(p);
      Inc(p1);
    end;
  end
  else Move(SourcePtr^, DestationPtr^, Size);
end;

procedure MoveMem_15_16(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
var
  i: Integer;
  p,p1: PWord;
begin
  if Size < 0 then exit;
  if (TransFilter and 1) = 1 then
  begin
    p := SourcePtr;
    p1 := DestationPtr;
    for i := 0 to Size - 1 do
    begin
      if p^ = TransColor then
      else p1^ := p^;
      Inc(p);
      Inc(p1);
    end;
  end
  else Move(SourcePtr^, DestationPtr^, Size * 2);
end;

procedure MoveMem_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
var
  i: Integer;
  p,p1: PInteger;
begin
  if Size < 0 then exit;
  if (TransFilter and 1) = 1 then
  begin
    p := SourcePtr;
    p1 := DestationPtr;
    for i := 0 to Size - 1 do
    begin
      if p^ <> TransColor then  p1^ := p^;
      Inc(p);
      Inc(p1);
    end;
  end
  else Move(SourcePtr^, DestationPtr^, Size * 4);
  
end;

procedure MoveMem_Lens_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);  
begin
  if Size < 0 then exit;
  
  if (TransFilter and 1) = 1 then
  asm
    push edi
    push esi
    push ebx

    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    xor eax, eax
    mov ebx, TransColor          //Transparent Color is in EBX
    push ebp                     //now, we doesn't need EBP register
    mov bp, 0011110111101111b    // EBX = AndMask (16bpp)

  @1:
    mov ax, [esi]               //  EAX = color from bitmap
    cmp ax, bx                 // if EAX = TransCol then skip
    jz @skip

    mov dx, [edi]               // EDX = color from destation bitmap
    shr ax, 1                   // EAX = EAX / 2
    shr dx, 1                   // EDX = EDX / 2
    and ax, bp                 // clear overflow bits in EAX
    and dx, bp                 // clear overflow bits in EDX
    add ax, dx                 // EAX = EAX + EDX
    mov [edi], ax               // Write EAX to destation bitmap/memory

  @skip:
    add edi, 2                   // goto next pixel
    add esi, 2
    loop @1

    pop ebp

    pop ebx
    pop esi
    pop edi
  end
  else
  asm
    push edi
    push esi
    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    xor eax, eax

    push ebx
    mov bx, 0011110111101111b

  @1:
    mov ax, [esi]
    mov dx, [edi]
    shr ax, 1
    shr dx, 1
    and ax, bx
    and dx, bx
    add ax, dx
    stosw
    add esi, 2
    loop @1

    pop ebx
    pop esi
    pop edi
  end;
  
end;

procedure MoveMem_Lens_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
begin
  if Size < 0 then exit;
  if (TransFilter and 1) = 1 then
  asm
    push edi
    push esi
    push ebx

    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    xor eax, eax
    mov ebx, TransColor          //Transparent Color is in EBX
    push ebp                     //now, we doesn't need EBP register
    mov ebp, 011111110111111101111111b  // EBX = AndMask (32bpp)

  @1:
    mov eax, [esi]               //  EAX = color from bitmap
    cmp eax, ebx                 // if EAX = TransCol then skip
    jz @skip

    mov edx, [edi]               // EDX = color from destation bitmap
    shr eax, 1                   // EAX = EAX / 2
    shr edx, 1                   // EDX = EDX / 2
    and eax, ebp                 // clear overflow bits in EAX
    and edx, ebp                 // clear overflow bits in EDX
    add eax, edx                 // EAX = EAX + EDX
    mov [edi], eax               // Write EAX to destation bitmap/memory

  @skip:
    add edi, 4                   // goto next pixel
    add esi, 4
    loop @1

    pop ebp
    
    pop ebx
    pop esi
    pop edi
  end
  else
  asm
    push edi
    push esi
    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    xor eax, eax

    push ebx
    mov ebx, 011111110111111101111111b

  @1:
    mov eax, [esi]
    mov edx, [edi]
    shr eax, 1
    shr edx, 1
    and eax, ebx
    and edx, ebx
    add eax, edx
    stosd
    add esi, 4
    loop @1

    pop ebx
    pop esi
    pop edi
  end;
end;

procedure MoveMem_Shadow_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
var
  Color: Integer;
begin
  if Size < 0 then exit;
  Color := Integer(OtherData^);  //get from BlendParams color type
  // only transparent mode is supported
  if (TransFilter and 1) = 1 then
  asm
    push edi
    push esi
    push ebx

    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    xor eax, eax
    mov ebx, TransColor         // Transparent Color is in EBX
    mov edx, Color
    shr dx, 1
    push ebp                    // now, we doesn't need EBP register
    mov bp, 0011110111101111b   // EBX = AndMask (16bpp)
    and dx, bp

  @1:
    mov ax, [esi]               // color in AX
    cmp ax, bx                  // if is transparent color then skip
    jz @skip

    mov ax, [edi]               // EDX = color from destation bitmap
    shr ax, 1                   // EAX = EAX / 2
    and ax, bp                  // clear overflow bits in EAX
    add ax, dx                  // EAX = EAX + EDX
    mov [edi], ax               // Write EAX to destation bitmap/memory

  @skip:
    add edi, 2                  // goto next pixel
    add esi, 2
    loop @1

    pop ebp

    pop ebx
    pop esi
    pop edi
  end;
end;

procedure MoveMem_Shadow_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
var
  Color: Integer;
begin
  if Size < 0 then exit;
  Color := Integer(OtherData^);  //get from BlendParams color type
  // only transparent mode is supported
  if (TransFilter and 1) = 1 then
  asm
    push edi
    push esi
    push ebx

    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    xor eax, eax
    mov ebx, TransColor         // Transparent Color is in EBX
    mov edx, Color
    shr edx, 1
    push ebp                    // now, we doesn't need EBP register
    mov ebp, 011111110111111101111111b  // EBX = AndMask (32bpp)
    and edx, ebp

  @1:
    mov eax, [esi]               // color in AX
    cmp eax, ebx                  // if is transparent color then skip
    jz @skip

    mov eax, [edi]               // EDX = color from destation bitmap
    shr eax, 1                   // EAX = EAX / 2
    and eax, ebp                  // clear overflow bits in EAX
    add eax, edx                  // EAX = EAX + EDX
    mov [edi], eax               // Write EAX to destation bitmap/memory

  @skip:
    add edi, 4                  // goto next pixel
    add esi, 4
    loop @1

    pop ebp

    pop ebx
    pop esi
    pop edi
  end;
end;

procedure MoveMem_Lightness_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
  OtherInfo: POtherInfo);
// hehe procedure :)))
var
  s,d: PWord;
  w,
  r,g,b: Word;
  i: Integer;
  ra,            //red amount
  ga,            //green amount
  ba: Word;      //blue amount
  Transparent: Boolean;
begin
  s := SourcePtr;
  d := DestationPtr;
  ra := Word(OtherData^) div 8;Inc(PWord(OtherData));
  ga := Word(OtherData^) div 8;Inc(PWord(OtherData));
  ba := Word(OtherData^) div 8;
  Transparent := (TransFilter and 1) = 1;
  for i := 0 to Size - 1 do
  begin
    w := s^;
    if (not Transparent) or (w <> TransColor) then
    begin
      r := (w and $7c00) shr 10;
      g := (w and $03e0) shr 5;
      b := (w and $001f);
      r := r + ((31 - r) * ra) div 31;
      g := g + ((31 - g) * ga) div 31;
      b := b + ((31 - b) * ba) div 31;
      d^ := (r shl 10) or (g shl 5) or b;
    end;
    Inc(s);
    Inc(d);
  end;
end;

procedure MoveMem_Lightness_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
  OtherInfo: POtherInfo);
// hehe procedure :)))
var
  s,d: PLongint;
  w: Longint;
  r,g,b: Word;
  i: Integer;
  ra,            //red amount
  ga,            //green amount
  ba: Word;      //blue amount
  Transparent: Boolean;
begin
  s := SourcePtr;
  d := DestationPtr;

  ra := Word(OtherData^);Inc(PWord(OtherData));
  ga := Word(OtherData^);Inc(PWord(OtherData));
  ba := Word(OtherData^);
  Transparent := (TransFilter and 1) = 1;
  for i := 0 to Size - 1 do
  begin
    w := s^;
    if (not Transparent) or (w <> TransColor) then
    begin
      r := (w shr 16) and $ff;
      g := (w shr 8) and $ff;
      b := w and $ff;
      
      r := r + ((255 - r) * ra) div 255;
      g := g + ((255 - g) * ga) div 255;
      b := b + ((255 - b) * ba) div 255;
      
      d^ := (r shl 16) or (g shl 8) or b;
    end;
    Inc(s);
    Inc(d);
  end;
end;

procedure MoveMem_Darkness_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
  OtherInfo: POtherInfo);
// hehe2 procedure :)))
var
  s,d: PWord;
  w,
  r,g,b: Word;
  i: Integer;
  ra,            //red amount
  ga,            //green amount
  ba: Word;      //blue amount
  Transparent: Boolean;
begin
  s := SourcePtr;
  d := DestationPtr;
  ra := Word(OtherData^) div 8;Inc(PWord(OtherData));
  ga := Word(OtherData^) div 8;Inc(PWord(OtherData));
  ba := Word(OtherData^) div 8;
  Transparent := (TransFilter and 1) = 1;
  for i := 0 to Size - 1 do
  begin
    w := s^;
    if (not Transparent) or (w <> TransColor) then
    begin
      r := (w and $7c00) shr 10;
      g := (w and $03e0) shr 5;
      b := (w and $001f);
      r := r - (r * ra) div 31;
      g := g - (g * ga) div 31;
      b := b - (b * ba) div 31;
      d^ := (r shl 10) or (g shl 5) or b;
    end;
    Inc(s);
    Inc(d);
  end;
end;

procedure MoveMem_Darkness_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
  OtherInfo: POtherInfo);
// hehe2 procedure :)))
var
  s,d: PLongint;
  w: Longint;
  r,g,b: Word;
  i: Integer;
  ra,            // red amount
  ga,            // green amount
  ba: Word;      // blue amount
  Transparent: Boolean;

begin
  s := SourcePtr;
  d := DestationPtr;
  ra := Word(OtherData^);Inc(PWord(OtherData));
  ga := Word(OtherData^);Inc(PWord(OtherData));
  ba := Word(OtherData^);
  Transparent := (TransFilter and 1) = 1;
  for i := 0 to Size - 1 do
  begin
    w := s^;
    if (not Transparent) or (w <> TransColor) then
    begin
      r := (w shr 16) and $ff;
      g := (w shr 8) and $ff;
      b := w and $ff;
      r := r - (r * ra) div 255;
      g := g - (g * ga) div 255;
      b := b - (b * ba) div 255;
      d^ := (r shl 16) or (g shl 8) or b;
    end;
    Inc(s);
    Inc(d);
  end;
end;

procedure MoveMem_disolve_15_filter(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo:
  POtherInfo);
var
  SeekSize: Integer;
  b: PByte;
begin
  if Size < 0 then exit; //if nothing to paint go away
  b := DisolveTempData;
  Inc(b, OtherInfo^.tY - OtherInfo^.Y);
  SeekSize := Integer(b^) + (OtherInfo^.tX - OtherInfo^.X);
  if (TransFilter and 2) = 2 then  //filter implementation
  asm
    push edi
    push esi

    mov esi, OtherData
    mov dl, [esi]

    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    mov eax, SeekSize
    push ebx
    mov ebx, TransColor
    push ebp
    mov ebp, DisolveTempData
    add ebp, eax             //seek to random start
  @1:
    mov ax, bx      //fill pixel with transparent color
    mov dh, [ebp]   //get value from Disolve map
    cmp dh, dl
    jnl @skip       // skip if nothing
    mov ax, [esi]   //get pixel from source pixel

  @skip:

    mov [edi], ax
    inc ebp
    add esi, 2
    add edi, 2

    loop @1
    pop ebp
    pop ebx
    pop esi
    pop edi
  end;
end;

procedure MoveMem_disolve_32_filter(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo:
  POtherInfo);
var
  SeekSize: Integer;
  b: PByte;
begin
  if Size < 0 then exit; //if nothing to paint go away
  b := DisolveTempData;
  Inc(b, OtherInfo^.tY - OtherInfo^.Y);
  SeekSize := Integer(b^) + (OtherInfo^.tX - OtherInfo^.X);
  if (TransFilter and 2) = 2 then  //filter implementation
  asm
    push edi
    push esi

    mov esi, OtherData
    mov dl, [esi]

    mov edi, DestationPtr
    mov esi, SourcePtr
    mov ecx, Size
    mov eax, SeekSize
    push ebx
    mov ebx, TransColor
    push ebp
    mov ebp, DisolveTempData
    add ebp, eax             //seek to random start
  @1:
    mov eax, ebx      //fill pixel with transparent color
    mov dh, [ebp]   //get value from Disolve map
    cmp dh, dl
    jnl @skip       // skip if nothing
    mov eax, [esi]   //get pixel from source pixel

  @skip:

    mov [edi], eax
    inc ebp
    add esi, 4
    add edi, 4

    loop @1
    pop ebp
    pop ebx
    pop esi
    pop edi
  end;
end;

procedure MoveMem_disolve_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo:
  POtherInfo);
var
  SeekSize: Integer;
  b: PByte;
begin
  if Size < 0 then exit; //if nothing to paint go away
  b := DisolveTempData;
  Inc(b, OtherInfo^.tY - OtherInfo^.Y);
  SeekSize := Integer(b^) + (OtherInfo^.tX - OtherInfo^.X);
  if (TransFilter and 1) = 1 then
  asm
    push edi
    push esi

    mov esi, OtherData
    mov dl, [esi]

    mov edi, DestationPtr
    mov esi, SourcePtr

    mov ecx, Size

    mov eax, SeekSize

    push ebx
    mov ebx, TransColor

    push ebp
    
    mov ebp, DisolveTempData
    add ebp, eax             //seek to random start

  @1:
    mov al, [ebp]
    cmp al, dl
    jnl @skip

    mov ax, [esi]
    cmp ax, bx
    jz @skip

    mov [edi], ax
  @skip:

    inc ebp
    add esi, 2
    add edi, 2

    loop @1
    pop ebp
    pop ebx
    pop esi
    pop edi
  end
  else
  asm //non transparent version
    push edi
    push esi

    mov esi, OtherData
    mov dl, [esi]
    
    mov edi, DestationPtr
    mov esi, SourcePtr

    mov ecx, Size

    mov eax, SeekSize

    push ebp
    mov ebp, DisolveTempData
    add ebp, eax             //seek to random start

  @1:
    mov al, [ebp]
    cmp al, dl
    jnl @skip

    mov ax, [esi]
    mov [edi], ax
  @skip:

    inc ebp
    add esi, 2
    add edi, 2

    loop @1

    pop ebp
    pop esi
    pop edi
  end;
end;

procedure MoveMem_disolve_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
  OtherInfo: POtherInfo);
var
  SeekSize: Integer;
  b: PByte;
begin
  if Size < 0 then exit; //if nothing to paint go away
  b := DisolveTempData;Inc(b, OtherInfo^.tY - OtherInfo^.Y);
  SeekSize := Integer(b^) + (OtherInfo^.tX - OtherInfo^.X);
  //seek to random start of prerendered DisloveBuffer,
  //It's nessenary generate same output image for same bitmap
  //can't be generated for each bitmap!
  if (TransFilter and 1) = 1 then
  asm
    push edi
    push esi

    mov esi, OtherData
    mov dl, [esi]

    mov edi, DestationPtr
    mov esi, SourcePtr

    mov ecx, Size

    mov eax, SeekSize

    push ebx
    mov ebx, TransColor

    push ebp
    mov ebp, DisolveTempData
    add ebp, eax             //seek to random start


  @1:
    mov al, [ebp]
    cmp al, dl
    jnl @skip

    mov eax, [esi]
    cmp eax, ebx
    jz @skip


    mov [edi], eax
  @skip:

    inc ebp
    add esi, 4
    add edi, 4

    loop @1
    pop ebp
    pop ebx
    pop esi
    pop edi
  end
  else
  asm //non transparent version
    push edi
    push esi

    mov esi, OtherData
    mov dl, [esi]
    
    mov edi, DestationPtr
    mov esi, SourcePtr

    mov ecx, Size

    mov eax, SeekSize

    push ebp                 //is this dangerous for compiler?
    mov ebp, DisolveTempData
    add ebp, eax             //seek to random start

  @1:
    mov al, [ebp]
    cmp al, dl
    jnl @skip

    mov eax, [esi]
    mov [edi], eax
  @skip:

    inc ebp
    add esi, 4
    add edi, 4

    loop @1

    pop ebp
    pop esi
    pop edi
  end;
end;

procedure MoveMem_Alpha_Transform_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
//todo: Some speed optimalization ...
var
  s: ^TTransformSpec;
  src,dest: PWord;
  AlphaData: PByte;
  Opacity: Byte;
  i,
  DeltaY,
  DeltaX: Integer;
  Transparent: Boolean;

begin
  if Size < 0 then exit; //nothing to draw
  if not Assigned(OtherData) then exit; //we need to have alpha in OtherData
  s := OtherData;
  Opacity := s^.frame;
  DeltaX := (OtherInfo^.tX - OtherInfo^.X) mod s^.alpha.Width;
  DeltaY := (OtherInfo^.tY - OtherInfo^.Y) mod s^.alpha.Height;
  AlphaData := s^.alpha.ScanLine[DeltaY];
  Inc(AlphaData, DeltaX);
  src := SourcePtr;
  dest := DestationPtr;
  Transparent := (TransFilter and 1) = 1;

  for i := 0 to Size - 1 do
  begin
    if AlphaData^ > Opacity then
    if (not Transparent) or (src^ <> TransColor) then
      dest^ := src^;
    Inc(src);
    Inc(dest);
    Inc(DeltaX);
    Inc(AlphaData);
    if DeltaX >= s^.alpha.Width then
    begin
      DeltaX := 0;
      AlphaData := s^.alpha.ScanLine[DeltaY];
    end;
  end;
end;

procedure MoveMem_Alpha_Transform_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
//this procedure is copy of Alpha_Transform_15 but p,p1 type was changed to PCardinal (due to 32bit)
var
  s: ^TTransformSpec;
  src,dest: PInteger;
  AlphaData: PByte;

  Opacity: Byte;
  i,
  DeltaY,
  DeltaX: Integer;
  Transparent: Boolean;

begin
  if Size < 0 then exit; //nothing to draw
  if not Assigned(OtherData) then exit; //we need to have alpha in OtherData
  s := OtherData;
  Opacity := s^.frame;
  DeltaX := (OtherInfo^.tX - OtherInfo^.X) mod s^.alpha.Width;
  DeltaY := (OtherInfo^.tY - OtherInfo^.Y) mod s^.alpha.Height;
  AlphaData := s^.alpha.ScanLine[DeltaY];
  Inc(AlphaData, DeltaX);
  src := SourcePtr;
  dest := DestationPtr;
  Transparent := (TransFilter and 1) = 1;

  for i := 0 to Size - 1 do
  begin
    if AlphaData^ > Opacity then
    if (not Transparent) or (src^ <> TransColor) then
      dest^ := src^;
    Inc(src);
    Inc(dest);
    Inc(DeltaX);
    Inc(AlphaData);
    if DeltaX >= s^.alpha.Width then
    begin
      DeltaX := 0;
      AlphaData := s^.alpha.ScanLine[DeltaY];
    end;
  end;
end;

procedure MoveMem_Alpha_Transform_15_filter(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
//todo: Some speed optimalization ...
var
  s: ^TTransformSpec;
  src,dest: PWord;
  AlphaData: PByte;
  Opacity: Byte;
  i,
  DeltaY,
  DeltaX: Integer;
  Transparent: Boolean;
  dw: Word;

begin
  if Size < 0 then exit; //nothing to draw
  if not Assigned(OtherData) then exit; //we need to have alpha in OtherData
  s := OtherData;
  Opacity := s^.frame;
  DeltaX := (OtherInfo^.tX - OtherInfo^.X) mod s^.alpha.Width;
  DeltaY := (OtherInfo^.tY - OtherInfo^.Y) mod s^.alpha.Height;
  AlphaData := s^.alpha.ScanLine[DeltaY];
  Inc(AlphaData, DeltaX);
  src := SourcePtr;
  dest := DestationPtr;
  Transparent := (TransFilter and 1) = 1;

  for i := 0 to Size - 1 do
  begin
    dw := Word(TransColor);
    if AlphaData^ > Opacity then
    if (not Transparent) or (src^ <> TransColor) then
      dw := src^;
    dest^ := dw;
    
    Inc(src);
    Inc(dest);
    Inc(DeltaX);
    Inc(AlphaData);
    if DeltaX >= s^.alpha.Width then
    begin
      DeltaX := 0;
      AlphaData := s^.alpha.ScanLine[DeltaY];
    end;
  end;
end;

procedure MoveMem_Alpha_Transform_32_filter(SourcePtr, DestationPtr: Pointer; Size: Integer;
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
//this procedure is copy of Alpha_Transform_15 but p,p1 type was changed to PCardinal (due to 32bit)
var
  s: ^TTransformSpec;
  src,dest: PInteger;
  AlphaData: PByte;

  Opacity: Byte;
  i,
  DeltaY,
  DeltaX: Integer;
  Transparent: Boolean;
  dw: Integer;

begin
  if Size < 0 then exit; //nothing to draw
  if not Assigned(OtherData) then exit; //we need to have alpha in OtherData
  s := OtherData;
  Opacity := s^.frame;
  DeltaX := (OtherInfo^.tX - OtherInfo^.X) mod s^.alpha.Width;
  DeltaY := (OtherInfo^.tY - OtherInfo^.Y) mod s^.alpha.Height;
  AlphaData := s^.alpha.ScanLine[DeltaY];
  Inc(AlphaData, DeltaX);
  src := SourcePtr;
  dest := DestationPtr;
  Transparent := (TransFilter and 1) = 1;

  for i := 0 to Size - 1 do
  begin
    dw := TransColor;
    if AlphaData^ > Opacity then
      if (not Transparent) or (src^ <> TransColor) then
        dw := src^;
    dest^ := dw;
    Inc(src);
    Inc(dest);
    Inc(DeltaX);
    Inc(AlphaData);
    if DeltaX >= s^.alpha.Width then
    begin
      DeltaX := 0;
      AlphaData := s^.alpha.ScanLine[DeltaY];
    end;
  end;
end;

procedure MoveMem_AxialBlend_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
var
  xx, sfact, dfact: Integer;
  p,
  sb,         //source bytes
  pb: PByte;  //destation bytes
  rr,gg,bb: Word;
  pe,
  pi: PCardinal;

  AStart,
  AMiddle,
  AEnd: Byte; //transparency in start,middle and end of image
  Pp,Half: Integer;
begin

  if Size <= 0 then exit;
  if not Assigned(OtherData) then
  begin
    Error('MoveMem_AxialBlend_32 without paramters!');
    exit;
  end;

  p := OtherData;
  AStart := p^; Inc(p);   //start transparency
  AMiddle := p^; Inc(p);  //midle transparency 50% of height
  AEnd := p^;             //end transparency
  Half := OtherInfo^.Height div 2;
  pp := OtherInfo^.tY - OtherInfo^.Y;

  if pp > Half then  //if we are in second part of image
  begin
    Dec(pp, Half);
    dfact := Trunc((pp * ((AEnd - AMiddle) / Half)) * 1.28);
    Inc(dfact, Trunc(AMiddle *1.28));
  end
  else
  begin
    dfact := Trunc((pp * ((AMiddle - AStart) / Half)) * 1.28);
    Inc(dfact, Trunc(AStart *1.28));
  end;

  if dfact > 128 then dfact := 128;
  sfact := 128 - dfact;
  sb := SourcePtr;
  pb := DestationPtr;

  if (TransFilter and 1) = 0 then
  begin
    for xx := 0 to Size - 1 do
    begin
      bb:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
      gg:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
      rr:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
      Inc(pb); //skip alpha
      Inc(sb);
      if rr > 255 then rr:= 255;
      if gg > 255 then gg:= 255;
      if bb > 255 then bb:= 255;
      asm mov eax, pb; sub eax, 4; mov pi, eax; end;
      pi^ := (rr shl 16) or (gg shl 8) or bb;
    end;
  end
  else
  begin  //for transparent mode
    for xx := 0 to Size - 1 do
    begin
      Pointer(pe) := Pointer(sb);
      if pe^ <> TransColor then
      begin
        bb:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
        gg:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
        rr:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
        Inc(pb); //skip alpha
        Inc(sb);
        if rr > 255 then rr:= 255;
        if gg > 255 then gg:= 255;
        if bb > 255 then bb:= 255;
        asm mov eax, pb; sub eax, 4; mov pi, eax; end;
        pi^ := (rr shl 16) or (gg shl 8) or bb;
      end
      else
      begin
        Inc(pb, 4);
        Inc(sb, 4);
      end;
    end;
  end;
end;

procedure MoveMem_Blend_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
var
  xx, sfact, dfact: Integer;
  p: PSingle;
  rr,gg,bb: Word;
  sp,
  dp: PWord;
begin

  if Size < 0 then exit;
  if not Assigned(OtherData) then
  begin
    Error('MoveMem_Blend_15 without SourceFactor and DestationFactor parameter!');
    exit;
  end;

  p := OtherData;
  sfact := trunc(p^ * 128);Inc(p);
  dfact := trunc(p^ * 128);
  sp := SourcePtr;
  dp := DestationPtr;

  if (TransFilter and 1) = 0 then
  begin
    for xx := 0 to Size - 1 do
    begin
      rr:= (((sp^ and $7c00) shr 10) * sfact + ((dp^ and $7c00) shr 10) * dfact) shr 7;
      gg:= (((sp^ and $03e0) shr  5) * sfact + ((dp^ and $03e0) shr  5) * dfact) shr 7;
      bb:= ((sp^ and $001f) * sfact + (dp^ and $001f) * dfact) shr 7;
      if rr > 31 then rr:= 31;
      if gg > 31 then gg:= 31;
      if bb > 31 then bb:= 31;
      dp^ := (rr shl 10) or (gg shl 5) or bb;
      Inc(sp);Inc(dp); //goto next pixels      
    end;
  end
  else
  begin  //for transparent mode
    for xx := 0 to Size - 1 do
    begin
      if sp^ <> TransColor then
      begin
        rr:= (((sp^ and $7c00) shr 10) * sfact + ((dp^ and $7c00) shr 10) * dfact) shr 7;
        gg:= (((sp^ and $03e0) shr  5) * sfact + ((dp^ and $03e0) shr  5) * dfact) shr 7;
        bb:= ((sp^ and $001f) * sfact + (dp^ and $001f) * dfact) shr 7;
        if rr > 31 then rr:= 31;
        if gg > 31 then gg:= 31;
        if bb > 31 then bb:= 31;
        dp^ := (rr shl 10) or (gg shl 5) or bb;
        Inc(sp);Inc(dp); //goto next pixels
      end
      else
      begin
        Inc(sp);Inc(dp);
      end;
    end;

  end;
end;

procedure MoveMem_Blend_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
var
  xx, sfact, dfact: Integer;
  p: PSingle;
  sb,         //source bytes
  pb: PByte;  //destation bytes
  rr,gg,bb: Word;
  pe,
  pi: PCardinal;
begin

  if Size <= 0 then exit;
  if not Assigned(OtherData) then
  begin
    Error('MoveMem_Blend_32 without SourceFactor and DestationFactor parameter!');
    exit;
  end;
  
  p := OtherData;
  sfact := trunc(p^ * 128);Inc(p);
  dfact := trunc(p^ * 128);
  sb := SourcePtr;
  pb := DestationPtr;
  if (TransFilter and 1) = 0 then
  begin
    for xx := 0 to Size - 1 do
    begin
      bb:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
      gg:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
      rr:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
      Inc(pb); //skip alpha
      Inc(sb);
      if rr > 255 then rr:= 255;
      if gg > 255 then gg:= 255;
      if bb > 255 then bb:= 255;
      asm mov eax, pb; sub eax, 4; mov pi, eax; end;
      pi^ := (rr shl 16) or (gg shl 8) or bb;
    end;
  end
  else
  begin  //for transparent mode
    for xx := 0 to Size - 1 do
    begin
      Pointer(pe) := Pointer(sb);
      if pe^ <> TransColor then
      begin
        bb:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
        gg:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
        rr:= (sb^ * sfact + pb^ * dfact) shr 7;Inc(pb);Inc(sb);
        Inc(pb); //skip alpha
        Inc(sb);             
        if rr > 255 then rr:= 255;
        if gg > 255 then gg:= 255;
        if bb > 255 then bb:= 255;
        asm mov eax, pb; sub eax, 4; mov pi, eax; end;
        pi^ := (rr shl 16) or (gg shl 8) or bb;
      end
      else
      begin
        Inc(pb, 4);
        Inc(sb, 4);
      end;
    end;
  end;
end;

procedure MoveMem_Grayscale_15(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
begin
  if (TransFilter and 1) = 1 then
  asm
    push ebx
    push edi
    push esi   //save registers

    mov esi, SourcePtr
    mov edi, DestationPtr
    mov ebx, Size  //setup registers

    mov eax, TransColor
    push ebp
    mov bp, ax

  @1:
    //******** Pixel cycle ***********
    mov ax, [esi]  //get one pixel from buffer
    cmp ax, bp
    jz @Skip
    //separate all values into ah=Red, al=Green, dl=Blue
    // ax == 0RRRRRGGGGGBBBBB
    mov dl, al
    and dl, $1f      // dl = blue color

    shr ax, 2       // ah == 000RRRRR,  al == GGGGGBBB
    shr al, 3       // al = green color, ah = red color

    shl ax, 1        // convert into Intensity= G*2+R*2+B*8
    shl dl, 2

    add al,ah     //
    add al,dl     // al = r + g + b 

    shr al, 3     //Intensity div 8
//    xor dx, dx
    xor ah, ah
    mov dl, al    //intensity into dl

    shl dx, 5
    or  ax, dx
    shl dx, 5
    or ax, dx
    mov [edi], ax

  @Skip:
  
    inc edi
    inc edi
    inc esi
    inc esi

    dec ebx
    jnz @1

    pop ebp

    pop esi
    pop edi
    pop ebx
  end
  else
  asm  // for not transparent
    push ebx
    push edi
    push esi   //save registers

    mov esi, SourcePtr
    mov edi, DestationPtr
    mov ebx, Size  //setup registers

  @1:
    //******** Pixel cycle ***********
    mov ax, [esi]  //get one pixel from buffer
    //separate all values into ah=Red, al=Green, dl=Blue
    // ax == 0RRRRRGGGGGBBBBB
    mov dl, al
    and dl, $1f      // dl = blue color

    shr ax, 2       // ah == 000RRRRR,  al == GGGGGBBB
    shr al, 3       // al = green color, ah = red color

    shl ax, 1        // convert into Intensity= G*2+R*2+B*8
    shl dl, 2

    add al,ah     //
    add al,dl     // al = r + g + b

    shr al, 3     //Intensity div 8
    xor ah, ah
    mov dl, al    //intensity into dl

    shl dx, 5
    or  ax, dx
    shl dx, 5
    or ax, dx
    mov [edi], ax

  @Skip:

    inc edi
    inc edi
    inc esi
    inc esi

    dec ebx
    jnz @1

    pop esi
    pop edi
    pop ebx
  end;
end;

procedure MoveMem_Grayscale_32(SourcePtr, DestationPtr: Pointer; Size: Integer;
    TransColor: Integer; TransFilter: Byte; OtherData: Pointer;
    OtherInfo: POtherInfo);
begin
  if (TransFilter and 1) = 1 then
  asm
    push ebx
    push edi
    push esi  //save registers

    mov ecx, TransColor
    mov esi, SourcePtr
    mov edi, DestationPtr //setup data registers
    mov ebx, Size

 @1:mov eax, [esi]
    cmp eax, ecx
    jz @skip

    xor edx, edx
    shr eax, 1
    and eax, 011111110111111101111111b

    shr al, 6
    mov dl, al
    shr eax, 8
    add dl, al
    add dl, ah

    xor eax,eax
    mov al, dl
    shl edx, 8
    or eax, edx
    shl edx, 8
    or eax, edx
    mov [edi], eax

  @skip:
    add edi, 4
    add esi, 4
    dec ebx
    jnz @1

    pop esi
    pop edi
    pop ebx
  end else
  asm
    push ebx
    push edi
    push esi  //save registers

    mov esi, SourcePtr
    mov edi, DestationPtr //setup data registers
    mov ebx, Size

 @1:mov eax, [esi]
//   cmp eax, t1
//   jz @skip

    xor edx, edx
    shr eax, 1
    and eax, 011111110111111101111111b

    shr al, 6
    mov dl, al
    shr eax, 8
    add dl, al
    add dl, ah

    xor eax,eax
    mov al, dl
    shl edx, 8
    or eax, edx
    shl edx, 8
    or eax, edx
    mov [edi], eax

//@skip:
    add edi, 4
    add esi, 4
    dec ebx
    jnz @1

    pop esi
    pop edi
    pop ebx
  end;  
end;

function TBitmap64.Clip(const bitmap: TBitmap64; var x,y, SizeX, SizeY, StartX, StartY: Integer): Boolean;
{ Clipping function for Bitmaps.}
begin
  if (x > FClipRect.Right) or ( x + bitmap.FWidth < FClipRect.Left) or
    (y > FClipRect.Bottom) or ( y + bitmap.FHeight < FClipRect.Top) then
  begin  //nothing to draw -> nothing to clip
    Result := false;
    exit;
  end;

  Result := True;
  if (X + bitmap.FWidth) > FClipRect.Right then
    SizeX :=  1 + FClipRect.Right- x else SizeX := bitmap.FWidth;
  if (Y + bitmap.FHeight) > FClipRect.Bottom then
    SizeY := 1 + FCLipRect.Bottom - y else SizeY := bitmap.FHeight;

  if X < FCLipRect.Left then   //clip x start
  begin
    StartX:= FClipRect.Left-x;
    X:= FClipRect.Left;
  end
  else StartX:= 0;

  if Y < FClipRect.Top then   //clip y start
  begin
    StartY := FClipRect.Top -y;
    Y:= FClipRect.Top;
  end
  else StartY:= 0;

  SizeX := SizeX- StartX;
  SizeY := SizeY- StartY;
  if (SizeX = 0) or (SizeY = 0) then Result := false;
end;

{ TBitmap64 }
constructor TBitmap64.Create(Width, Height: Integer;
  PixelFormat: TPixelFormat; AlphaFormat: TAlphaFormat);
begin
  FWasAssigned := False;
  try
    bmBits := Nil;
    SetupBitmap(Width, Height, PixelFormat, AlphaFormat); //first setup settings
//    SetSize(Width, Height); //than Allocate memory for Bitmap
    SetupFilterMemory;      // ... for Filters
  finally
  end;
{  FPixelFormat := PixelFormat;
  FAlphaFormat := AlphaFormat;
  SetupBits;  //setup colors bits for next use
  FClipRect := Rect(0, 0, Width, Height);  //set clipper to full bitmap size
  case FPixelFormat of
    pf8bit: bmBpp := 1;
    pf15bit,pf16bit: bmBpp := 2;
    pf24bit: bmBpp := 3;
    pf32bit: bmBpp := 4;
  end;

  TopDown := True; //image is classic top down orientation
  SetSize(Width, Height);  //allocate memory

  BlendMode := bmCopy;
  PaintMode := pmCopy;
  Transparent := False;
  TransparentColor := -1;
  FCenter := Point(0, 0);  //center of screen is top left point}
end;

constructor TBitmap64.Create;
begin
  Create(0, 0, g64DefaultPixelFormat, g64DefaultAlphaFormat);
end;

procedure TBitmap64.SetupBitmap(Width, Height: Integer; PixelFormat: TPixelFormat;
  AlphaFormat: TAlphaFormat);
// setup bitmap  :)
// no memory allocations!!!
begin
//  FPixelFormat := PixelFormat;
  Self.PixelFormat := PixelFormat;
  FAlphaFormat := AlphaFormat;
  SetupBits;  //setup colors bits for next use
  FClipRect := Rect(0, 0, Width, Height);  //set clipper to full bitmap size
  TopDown := True; //image is classic top down orientation

  BlendMode := bmCopy;
  PaintMode := bmCopy;
  Transparent := False;
  TransparentColor := -1;
  FCenter := Point(0, 0);  //center of screen is top left point
end;


procedure TBitmap64.SetupBits;
begin
  case FPixelFormat of
  pf15bit:
    begin       {normal 555 format for 15bpp resolutions}
      RBitMask := $7c00;
      GBitMask := $03e0;
      BBitMask := $001f;
      ABitMask := $ffff;
      RStart := 10;
      GStart := 5;
      BStart := 0;
      AStart := 0;
    end;
  pf16bit:
    begin       {normal 565 format for 16bpp resolutions}
      RBitMask := $f800;
      GBitMask := $07e0;
      BBitMask := $001f;
      ABitMask := $ffff;
      RStart := 11;
      GStart := 5;
      BStart := 0;
      AStart := 0;
    end;
  pf24bit,pf32bit:
    begin        { 888 format for 24bpp, a888 format for 32bit, a-alpha,not used yet}
      RBitMask := $ff000000;
      GBitMask := $00ff0000;
      BBitMask := $0000ff00;
      ABitMask := $000000ff;
      RStart := 16;
      GStart := 8;
      BStart := 0;
      AStart := 24;
    end;
  end;
end;

procedure TBitmap64.SetPixelFormat(Value: TPixelFormat);
begin
  FPixelFormat := Value;
  case FPixelFormat of
    pf8bit: bmBpp := 1;
    pf15bit,pf16bit: bmBpp := 2;
    pf24bit: bmBpp := 3;
    pf32bit: bmBpp := 4;
  end;
end;

procedure TBitmap64.SetSize(NewWidth, NewHeight: Integer);
begin
  if (NewWidth = FWidth) and (NewHeight = FHeight) then exit; //nothing to change
  if Assigned(bmBits) then FreeMem(bmBits);   //release old picture data
  FWidth := NewWidth;
  FHeight := NewHeight;
  linePitch := bmBpp * FWidth;
  try
    GetMem(bmBits, FWidth * FHeight * bmBpp); //new size
  except
    on E: EOutOfMemory do
    begin
      Error(E.Message);
      bmBits := Nil;
    end;
  end;
end;

destructor TBitmap64.Destroy;
begin
  if Assigned(FFilterMem) then
    FreeMem(FFilterMem);
  if Assigned(bmBits) then
    FreeMem(bmBits);
  inherited Destroy;
end;


procedure TBitmap64.FillMemProc(DestationPtr: Pointer; Size,
  Color: Integer; OtherData: Pointer; OtherInfo: POtherInfo);
begin
{  if Filter then
  begin
    FFillMemProc(FFilterMem, Size, Color, OtherData);
    FFilterMemProc(FFilterMem, DestationPtr, Size, 0, False, FFilterData, Nil);
  end
  else}
    FFillMemProc(DestationPtr, Size, Color, OtherData);
end;

procedure TBitmap64.MoveMemProc(SourcePtr, DestationPtr: Pointer; Size,
  TransColor: Integer; TransFilter: Byte; OtherData: Pointer; OtherInfo: POtherInfo);
begin
  if BlendFilter then
  begin
    FFilterMemProc(SourcePtr, FFilterMem, Size, TransColor, TransFilter or 2,
      FFilterData, OtherInfo);
   FMoveMemProc(FFilterMem, DestationPtr, Size, TransColor, TransFilter,
      OtherData, OtherInfo);
{    FMoveMemProc(SourcePtr, FFilterMem, Size, TransColor, TransFilter,
      OtherData, OtherInfo);
    FFilterMemProc(FFilterMem, DestationPtr, Size, TransColor, TransFilter,
      FFilterData, OtherInfo);}

  end
  else
    FMoveMemProc(SourcePtr, DestationPtr, Size, TransColor, TransFilter,
      OtherData, OtherInfo);
end;

function TBitmap64.GetScanLine(Row: Integer): Pointer;
// note: GetScanLine doesn't accept ClipRect! You can get scanline outside ClipRect!
{$Warnings Off}
begin
  if bmBits = Nil then exit; //optremove
  if (Row < 0) or (Row >= FHeight) then
  begin
    Error(Format('GetScanLine: You want row number %d, but bitmap have only %d rows(Bitmap height is %d)',[Row, FHeight-1, FHeight]));
  end
  else
  if TopDown then
    Result := Pointer(Longint(bmBits) + row * linePitch)
  else
    Result := Pointer(Longint(bmBits) + (FHeight - 1 - row) * linePitch)
end;
{$Warnings on}

procedure TBitmap64.SetClipper(r: TRect);
begin
  SwapMinI(r.Left, r.Right);  //left is left, right is right ;)
  SwapMinI(r.Top, r.Bottom);

  // Clip the clipper first
  if r.Right >= FWidth then r.Right := FWidth - 1;
  if r.Bottom >= FHeight then r.Bottom := FHeight -1;

  if r.Left < 0 then r.Left := 0;
  if r.Top < 0 then r.Top := 0;

  FClipRect := r;
end;

procedure TBitmap64.SetBlendParams(Params: Pointer);
begin
  FCopyData := Params;
end;

procedure TBitmap64.SetPaintParams(Params: Pointer);
begin
  FFillData := Params;
end;

procedure TBitmap64.SetBlendFilterParams(Params: Pointer);
begin
  FFilterData := Params;
end;

Function PixelFormat2Int(pf: TPixelFormat): Integer;
begin

  case pf of
    pf8bit: Result := 0;
    pf15bit: Result := 1;
    pf16bit: Result := 2;
    pf24bit: Result := 3;
    pf32bit: Result := 4;
  else
    Result := 4
  end;
end;

procedure TBitmap64.SetBlendMode(BlendMode: TBlendMode);
begin
  FMoveMemProc := BlendModes[PixelFormat2Int(PixelFormat), BlendMode];
end;

procedure TBitmap64.SetPaintMode(PaintMode: TPaintMode);
begin
  FFillMemProc := FillModes[PixelFormat2Int(PixelFormat), PaintMode];
end;

procedure TBitmap64.SetBlendFilterMode(FilterMode: TBlendMode);
begin
  if FilterMode = fmNone then
    BlendFilter := False
  else
  begin
    BlendFilter := True;
    FFilterMemProc := FilterModes[PixelFormat2Int(PixelFormat), FilterMode];
  end;
end;

// ---------------------------features -------------------------
function TBitmap64.rgb(r, g, b: Byte; a: Byte): Integer;
begin
  case FPixelFormat of
    pf8bit:  result := 255;
    pf15bit: result := ((r div 8) shl 10) + ((g div 8) shl 5) + ((b div 8));
    pf16bit: result := ((r div 8) shl 11) + ((g div 4) shl 5) + ((b div 8));
    pf24bit,pf32bit: result := (a shl 24) + (r shl 16) + (g shl 8) + b;
  else
  begin
    Error('Unsupported pixel format!');
    Result := 0;
  end;
  end;
end;


procedure TBitmap64.SetPixel(X, Y, Color: Integer);
begin
  Inc(X, Center.X);
  Inc(Y, Center.Y);

  if (X >= ClipRect.Left) and (X <= ClipRect.Right) and (Y >= ClipRect.Top)
    and (Y <= ClipRect.Bottom) then
    FillMemProc(Ptr(Integer(ScanLine[y]) + X * bmBpp), 1, Color, FFillData, Nil);
end;

function TBitmap64.GetPixel(X, Y: Integer): Integer;
var
  p: Pointer;
begin
  Result := -1;
  Inc(X, Center.X);
  Inc(Y, Center.Y);

  if (X >= ClipRect.Left) and (X <= ClipRect.Right) and (Y >= ClipRect.Top)
    and (Y <= ClipRect.Bottom) then
  begin
    p := Ptr(Integer(ScanLine[Y]) + X * bmBpp);
    case PixelFormat of
      pf8bit: Result := Byte(p^);
      pf15bit,pf16bit: Result := Word(p^);
      pf32bit: Result := Integer(p^);
    end;
  end;
end;


procedure TBitmap64.HLine(X, Y, X1, Color: Integer);
begin
  Inc(X, FCenter.X);
  Inc(X1, FCenter.X);
  Inc(Y, FCenter.Y);

  SwapMinI(X, X1);  // X is left, and X1 is right
  if (X1 < ClipRect.Left) or (X >= ClipRect.Right) or (Y > ClipRect.Bottom)
    or (Y < ClipRect.Top) then exit;  //if line is totaly out of clip region

  if X < ClipRect.Left then X := ClipRect.Left;
  if X1 > ClipRect.Right then X1 := ClipRect.Right;
    FillMemProc(Ptr(Integer(ScanLine[Y]) + X * bmBpp), 1 + X1 - X, Color,
      FFillData , Nil);
end;

procedure TBitmap64.VLine(X, Y, Y1, Color: Integer);
// note: this is not fasted way to draw vertical lines :( ,it's univerzal
var
  i: Integer;
begin
  Inc(X, FCenter.X);
  Inc(Y, FCenter.Y);
  Inc(Y1, FCenter.Y);

  SwapMinI(Y, Y1);  // Y is top, and Y1 is bottom
  if (X < ClipRect.Left) or (X > ClipRect.Right) or (Y1 < ClipRect.Top)
    or (Y > ClipRect.Bottom) then exit;  //line is out of "space"

  if Y < ClipRect.Top then Y := ClipRect.Top;
  if Y1 > ClipRect.Bottom then Y1 := ClipRect.Bottom;
  for i := Y to Y1 do
    FillMemProc(Ptr(Integer(ScanLine[i]) + X * bmBpp), 1, Color, FFillData, Nil);
end;

procedure TBitmap64.Rectangle(X, Y, X1, Y1, Color: Integer);
begin
  SwapMinI(Y, Y1);
  SwapMinI(X, X1);
  HLine(X, Y, X1, Color);
  HLine(X, Y1, X1, Color);
  if (Y1 - Y) > 2 then
  begin
    VLine(X, Y + 1, Y1 - 1, Color);
    VLine(X1, Y + 1, Y1 - 1, Color);
  end;
end;

procedure TBitmap64.Rectangle(const Rect: TRect; Color: Integer);
begin
  Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom, Color);
end;

procedure TBitmap64.FillRect(const Rect: TRect; Color: Integer);
// todo: FillRect, some optimalization for direct clipping and
// direct call FFillMem function instead of call Hline
var
  i: Integer;
  r: TRect;
begin
  r := Rect;
  SwapMinI(r.Left, r.Right);
  SwapMinI(r.Top, r.Bottom);
  for i := r.Top to r.Bottom do
    HLine(r.Left, i, r.Right, Color);
end;

function sgn(a: Integer): Integer; //for line usage
begin
  if a > 0  then Result := + 1 else
  if a < 0 then Result := - 1 else Result := 0;
end;

procedure TBitmap64.Line(X, Y, X1, Y1, Color: Integer; First: Boolean);
// todo: Line, slow pixeling used!
// If you don't want to draw first pixel set First to False!
var
  i,s,d1x,d1y,d2x,d2y,u,v,m,n: Integer;
begin
  u := X1 - X;
  v := Y1 - Y;
  d1x := SGN(u);
  d1y := SGN(v);
  d2x := SGN(u);
  d2y := 0;
  m := ABS(u);
  n := ABS(v);
  if not (M > N) then
  begin
    d2x := 0 ;
    d2y := SGN(v);
    m := ABS(v);
    n := ABS(u);
  end;
  s := m shr 1;
  for i := 0 to m do //main loop
  begin
    if (i = 0) then
    begin
      if First then SetPixel(X, Y, Color);
    end
    else
     SetPixel(X, Y, Color); // this is realy slow :(
    s := s + n;
    if not (s < m) then
    begin
      s := s - m;
      X := X + d1x;
      Y := Y + d1y;
    end
    else
    begin
      X := X + d2x;
      Y := Y + d2y;
    end;
  end;
end;

procedure TBitmap64.Polyline(const Points: array of TPoint; Color: Integer; StartIndex,
  NumPts: Integer);
//note: Polyline isn't cliped, because function line is clipped, so what?
var
  a,b,x: Integer;
begin
  a := StartIndex;

  if NumPts = -1 then
    b := High(Points)
  else
    b := StartIndex + NumPts;
  // instead of calling exception, fix range
  if a < Low(Points) then a := Low(Points);
  if b > High(Points) then b := High(Points);

  if b > a then
  for x := a to b - 1 do
    Line(Points[x].X, Points[x].Y, Points[x + 1].X, Points[x + 1].Y, Color,
      (x = a));
      
end;

procedure TBitmap64.Ellipse(x1, y1, x2, y2, Color: Integer);
// x1,y1 is top left, x2,y2 is bottom right
// note: it's slow setpixel used :(
var
 px, py,
 dx, dy,
 r, rx, ry: Integer;
begin
  SwapMinI(x1, x2);
  SwapMinI(y1, y2);

  px := (x2 - x1) div 2;
  py := (y2 - y1) div 2;

  y1 := y1 + py;
  y2 := y1;

  dx := (px * px) shl 1;
  dy := (py * py) shl 1;
  r  := px * py * py;
  rx := r shl 1;
  ry := 0;

  SetPixel(x1, y1, Color);
  SetPixel(x2, y1, Color);

  while px > 0 do
  begin
    if r > 0 then
    begin
      Inc(y1);
      Dec(y2);
      Inc(ry, dx);
      Dec(r, ry);
    end;
    if r <= 0 then
    begin
      Dec(px);
      Inc(x1);
      Dec(x2);
      Dec(rx, dy);
      Inc(r, rx);
    end;
    SetPixel(x1, y1, Color);
    SetPixel(x1, y2, Color);
    SetPixel(x2, y1, Color);
    SetPixel(x2, y2, Color);
  end;
  
end;

procedure TBitmap64.FillEllipse(x1, y1, x2, y2, Color: Integer);
// x1,y1 is top left, x2,y2 is bottom right, same as Ellipse ;)
var
  oy,
  px, py,
  dx, dy,
  r, rx, ry: Integer;
begin
  SwapMinI(x1, x2);
  SwapMinI(y1, y2);

  px := (x2 - x1) div 2;
  py := (y2 - y1) div 2;

  y1 := y1 + py;
  y2 := y1;

  dx := (px * px) shl 1;
  dy := (py * py) shl 1;
  r  := px * py * py;
  rx := r shl 1;
  ry := 0;

  HLine(x1, y1, x2, Color);
  oy := y1;

  while px > 0 do
  begin
    if r > 0 then
    begin
      Inc(y1);
      Dec(y2);
      Inc(ry, dx);
      Dec(r, ry);
    end;
    if r <= 0 then
    begin
      Dec(px);
      Inc(x1);
      Dec(x2);
      Dec(rx, dy);
      Inc(r, rx);
    end;
    if (oy <> y1) then
    begin
      HLine(x1, y1, x2, Color);
      HLine(x1, y2, x2, Color);
    end;
    oy := y1;
  end;
end;

procedure TBitmap64.Triangle(x1, y1, x2, y2, x3, y3, Color: Integer);
begin
  Line(x1, y1, x2, y2, Color, False);
  Line(x2, y2, x3, y3, Color, False);
  Line(x3, y3, x1, y1, Color, False);
end;

procedure TBitmap64.FillTriangle(x1, y1, x2, y2, x3,y3, color:Integer);
//copy paste old Triangle function :(
var
  First,Last,xx,ax,bx,yy,p1,q1,p2,q2,p3,q3: Integer;
begin
  {First we must find first and last line}
  First := y1; Last := y1;
  if y2 < First then First :=y2;
  if y2 > Last then Last :=y2;
  if y3 < First then First :=y3;
  if y3 > Last then Last :=y3;

  p1 := x1 - x3; q1 := y1 - y3;
  p2 := x2 - x1; q2 := y2 - y1;
  p3 := x3 - x2; q3 := y3 - y2;
  for yy:=First to Last do
    begin
      ax := FWidth;
      bx:=-1;
      if (y3>=yy) or (y1>=yy) then
        if (y3<=yy) or (y1<=yy) then
          if not(y3=y1) then
            begin
              xx:=(yy-y3)*p1 div q1+x3;
              if xx<ax then ax:=xx;
              if xx>bx then bx:=xx;
            end;
      if (y1>=yy) or (y2>=yy) then
        if (y1<=yy) or (y2<=yy) then
          if not(y1=y2) then
            begin
              xx:=(yy-y1)*p2 div q2+x1;
              if xx<ax then ax:=xx;
              if xx>bx then bx:=xx;
            end;
      if (y2>=yy) or (y3>=yy) then
        if (y2<=yy) or (y3<=yy) then
          if not(y2=y3) then
            begin
              xx:=(yy-y2)*p3 div q3+x2;
              if xx<ax then ax:=xx;
              if xx>bx then bx:=xx;
            end;
      if ax<=bx then HLine(ax, yy, bx, color);
    end;
end;

procedure TBitmap64.Draw(x, y: Integer; b: TBitmap64);
var
  SizeX, SizeY,
  StartX, StartY: Integer;
  i: Integer;
  oi: POtherInfo;
begin
  Inc(x, Center.X);
  Inc(y, Center.y);
  New(oi);
  oi^.X := x;
  oi^.Y := y;
  oi^.Width := b.Width;
  oi^.Height := b.Height;

  if not Clip(b, x,y, SizeX, SizeY, StartX, StartY) then
  begin
    Dispose(oi);
    exit;
  end;

  oi^.tX := x;
  oi^.tY := y;

  if SizeY > 0 then
  for i := 0 to SizeY - 1 do
  begin
    MoveMemProc(Ptr(Integer(b.ScanLine[i + StartY]) + StartX * bmBpp),
      Ptr(Integer(ScanLine[i + y]) + x * b.bmBpp), SizeX, b.TransparentColor,
       Byte(b.Transparent), FCopyData, oi{, x, i + y});
    Inc(oi^.tY);
  end;
  Dispose(oi);
end;


procedure TBitmap64.ReallocateBuffer;
begin

  try
    if Assigned(FFilterMem) then
    FreeMem(FFilterMem);
  except
    Error('Can''t reallocate memory buffer !!!');
  end;

  try
    GetMem(FFilterMem, FWidth * bmBpp);  //one line for used filters
  except
    on E: EOutOfMemory do
    begin
      Error(E.Message);
      bmBits := Nil;
    end;
  end;

end;

procedure TBitmap64.MoveCenter(RelativeX, RelativeY: Integer);
var
  p: TPoint;
begin
  p := FCenter;
  p.X := p.X + RelativeX;
  p.Y := p.Y + RelativeY;
  Center := p;
end;

procedure TBitmap64.SetupFilterMemory;
begin
  if FWidth > 0 then
  try
   if Assigned(FFilterMem) then
   try
     FreeMem(FFilterMem);
   except
     Error('Cant'' free memory used by FFilterMem!');
   end;
    GetMem(FFilterMem, FWidth * bmBpp);  //one line  for used filters
  except
    on E: EOutOfMemory do
    begin
      Error(E.Message);
    end;
  end;
end;

procedure AssignBlendProc(Depth: Integer; Mode: Integer; Proc: TMoveMemProc;
  DefaultProc: TMoveMemProc);
begin
  if Assigned(Proc) then
    BlendModes[Depth, Mode] := Proc
  else
    BlendModes[Depth, Mode] := DefaultProc;
end;

procedure AssignFillProc(Depth: Integer; Mode: Integer; Proc: TFillMemProc;
  DefaultProc: TFillMemProc);
begin
  if Assigned(Proc) then
    FillModes[Depth, Mode] := Proc
  else
    FillModes[Depth, Mode] := DefaultProc;
end;

procedure RegisterBlendMode(Mode: Integer;
  BlendFunc8, BlendFunc15, BlendFunc16, BlendFunc24, BlendFunc32: TMoveMemProc;
  FillFunc8, FillFunc15, FillFunc16, FillFunc24, FillFunc32: TFillMemProc);
// so stupid function :)
begin
  AssignBlendProc(0, Mode, BlendFunc8, MoveMem_8);
  AssignBlendProc(1, Mode, BlendFunc15, MoveMem_15_16);
  AssignBlendProc(2, Mode, BlendFunc16, MoveMem_15_16);
  AssignBlendProc(3, Mode, BlendFunc24, MoveMem_32);
  AssignBlendProc(4, Mode, BlendFunc32, MoveMem_32);

  AssignFillProc(0, Mode, FillFunc8, FillMem_8);
  AssignFillProc(1, Mode, FillFunc15, FillMem_15_16);
  AssignFillProc(2, Mode, FillFunc16, FillMem_15_16);
  AssignFillProc(3, Mode, FillFunc24, FillMem_32);
  AssignFillProc(4, Mode, FillFunc32, FillMem_32);
end;

procedure AssignBlendFilterProc(Depth: Integer; Mode: Integer; Proc: TMoveMemProc;
  DefaultProc: TMoveMemProc);
begin
  if Assigned(Proc) then
    FilterModes[Depth, Mode] := Proc
  else
    FilterModes[Depth, Mode] := DefaultProc;
end;

procedure RegisterBlendFilterMode(Mode: Integer;
  BlendFunc8, BlendFunc15, BlendFunc16, BlendFunc24, BlendFunc32: TMoveMemProc);
// so, next stupid function :))))
begin
  AssignBlendFilterProc(0, Mode, BlendFunc8, MoveMem_8);
  AssignBlendFilterProc(1, Mode, BlendFunc15, MoveMem_15_16);
  AssignBlendFilterProc(2, Mode, BlendFunc16, MoveMem_15_16);
  AssignBlendFilterProc(3, Mode, BlendFunc24, MoveMem_32);
  AssignBlendFilterProc(4, Mode, BlendFunc32, MoveMem_32);
end;


function DetectCPU: Integer;
begin
  asm
    pushfd //; save EFLAGS
    pop eax //; store EFLAGS in EAX
    mov ebx, eax //; save in EBX for later testing
    xor eax, 00200000h //; toggle bit 21
    push eax //; put to stack
    popfd //; save changed EAX to EFLAGS
    pushfd //; push EFLAGS to TOS
    pop eax //; store EFLAGS in EAX
    cmp eax, ebx //; see if bit 21 has changed
//    jz NO_CPUID //; if no change, no CPUID
  end;
end;

initialization

  GetMem(DisolveTempData, DisolveTempDataSize);  //alocate memory for Disolve filter

  _p := DisolveTempData;
  Randomize;
  for _i := 0 to DisolveTempDataSize - 1 do  //calculate Disolve map
  begin
    _p^ := Random(100);
    Inc(_p);
  end;

  //register blending modes
  RegisterBlendMode(bmCopy,
    MoveMem_8, MoveMem_15_16, MoveMem_15_16, Nil, MoveMem_32,
    FillMem_8, FillMem_15_16, FillMem_15_16, Nil, FillMem_32);

  RegisterBlendMode(bmLens,
    Nil, MoveMem_lens_15, Nil, Nil, MoveMem_lens_32,
    Nil, FillMem_lens_15, Nil, Nil, FillMem_lens_32);

  RegisterBlendMode(bmDisolve,
    Nil, MoveMem_Disolve_15, Nil, Nil, MoveMem_Disolve_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmBlend,
    Nil, MoveMem_Blend_15, Nil, Nil, MoveMem_Blend_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmGrayScale,
    Nil, MoveMem_GrayScale_15, Nil, Nil, MoveMem_GrayScale_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmLightness,
    Nil, MoveMem_Lightness_15, Nil, Nil, MoveMem_Lightness_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmDarkness,
    Nil, MoveMem_Darkness_15, Nil, Nil, MoveMem_Darkness_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmShadow,
    Nil, MoveMem_Shadow_15, Nil, Nil, MoveMem_Shadow_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmAlphaTransform,
    Nil, MoveMem_Alpha_Transform_15, Nil, Nil, MoveMem_Alpha_Transform_32,
    Nil, Nil, Nil, Nil, Nil);

  RegisterBlendMode(bmAxialBlend, Nil, Nil, Nil, Nil, MoveMem_AxialBlend_32,
  Nil, Nil, Nil, Nil, Nil);

  //register filter modes for blending
  RegisterBlendFilterMode(fmDisolve,
    Nil, MoveMem_disolve_15_filter, Nil, Nil, MoveMem_disolve_32_filter);

  RegisterBlendFilterMode(fmAlphaTransform,
    Nil, MoveMem_Alpha_Transform_15_filter, Nil, Nil, MoveMem_Alpha_Transform_32_filter);

finalization

  FreeMem(DisolveTempData, DisolveTempDataSize);

end.

