// Screen64 Unit, part of Graph64 project
// Interface between Graph64 and others targets:
// GDI(TBitmap)

// Designed for Delphi 6 Professional
// Code by: Pavol Stugel
// (c) NETGraphics 2002

{ $Define CLX}

// define CLX if you create multiplatform app or for linux
// BUT: CLX doesn't support 15bit !!!! I don't now why!???! Why Borland?!?!?


unit Screen64;

interface

uses
{$ifdef CLX}
  QGraphics,
{$else}
  Graphics,
{$endif}
  SysUtils,Types, Graph64;

type
  TBufferType = (btSelf, btLinkedBitmap);
  TScreen64 = class(TBitmap64)
  private
    FBuffer: TBitmap;
    FBufferType: TBufferType;
    FWasAssigned: Boolean;

  public
    function GetScanLine(Row: Integer): Pointer;override;
    constructor Create;
    procedure LinkToTBitmap(Bitmap: TBitmap);
    procedure LoadFromBMP(FileName:  String);
    destructor Destroy;override;
  end;

const
{$ifdef CLX}
  _pf8bit= QGraphics.pf8bit;
//  _pf15bit= QGraphics.pf15bit; !!!!!!!!!
  _pf16bit= QGraphics.pf16bit;
  _pf24bit= QGraphics.pf24bit;
  _pf32bit= QGraphics.pf32bit;
{$else}
  _pf8bit= Graphics.pf8bit;
  _pf15bit= Graphics.pf15bit;
  _pf16bit= Graphics.pf16bit;
  _pf24bit= Graphics.pf24bit;
  _pf32bit= Graphics.pf32bit;
{$endif}

implementation

{ TScreen64 }

function pfToWin(PixelFormat: TPixelFormat): Graphics.TPixelFormat;
// Graph64 pixel format to Windows pixel format
begin
  case PixelFormat of
    pf8bit : Result := _pf8bit;
    pf15bit : Result := _pf15bit;
    pf16bit : Result := _pf16bit;
    pf32bit : Result := _pf32bit;
  else
  Result := _pf32bit;
  end
end;

constructor TScreen64.Create;
begin
  FWasAssigned := False;
  FBuffer := Nil;
  inherited Create;
  FBufferType := btSelf;
end;

procedure TScreen64.LoadFromBMP(FileName:  String);
begin
  if Assigned(FBuffer) then
  begin
    try
      FBuffer.LoadFromFile(FileName);
    except
      Error(Format('Can''t load %s', [FileName]));
    end;
    FBuffer.PixelFormat := pfToWin(PixelFormat);
    FWidth := FBuffer.Width;
    FHeight := FBuffer.Height;
    ClipRect := Rect(0, 0, FWidth, FHeight);
  end;
end;

destructor TScreen64.Destroy;
begin
  if not FWasAssigned then
    FBuffer.Free;
  inherited;
end;

function TScreen64.GetScanLine(Row: Integer): Pointer;
begin
  if Assigned(FBuffer) then Result := FBuffer.ScanLine[Row]
    else Result := Nil;
end;

procedure TScreen64.LinkToTBitmap(Bitmap: TBitmap);
var
{$ifdef CLX}
  pf: QGraphics.TPixelFormat;
{$else}
  pf: Graphics.TPixelFormat;
{$endif}

begin
  FWasAssigned := False;
  if Assigned(Bitmap) then
  begin
    pf := Bitmap.PixelFormat;
    if (pf = _pf16bit) or (pf = _pf24bit) then
      Error('Sorry this pixel format isn''t supported!')
    else
    begin
      // now it's supported only 15/32 bit bitmaps
      if pf = _pf32bit then PixelFormat := pf32bit
      else
{      if pf = _pf15bit then }PixelFormat := pf15bit;
      try
        FBuffer := Bitmap; // make link to existing bitmap
        FWidth := Bitmap.Width;
        FHeight := Bitmap.Height;
        ReallocateBuffer; //realocate filter buffer
        ClipRect := Rect(0, 0, FWidth, FHeight);
      except
        Error('Lost bitmap!');
        FBuffer := Nil;
      end;
      FWasAssigned := True;
    end;
  end
  else
    Error('Can''t assign NULL bitmap!');
end;

end.
