unit RTTI;

{
  All the code in this unit is copyrighted by Mert Sakarya.
  Please don't remove this comment.
  Feel free to ask anything you want.
  My email address is, mertsakarya@hotmail.com
  version 0.01:)
}
interface

uses
  Windows, Messages, SysUtils, Classes, typInfo, Grids,
  Controls, StdCtrls, MSXML2_TLB, Graphics, FastStrings;

type
  PMethod = ^TMethod;

  TRTTI = class(TComponent)
  private
    fFields : TStrings;
    procedure SetFields(const Value: Tstrings);
    function Process(node : IXMLDOMNode; name : string; var o:TObject; parent : TWinControl = nil): string;
    function createComponent(cType, cName : string; parent : TWinControl) : TControl;
    function GetRegisteredClasses(const Name: String): TControlClass;
    procedure SetRegisteredClasses(const Name: String; const Value: TControlClass);
    procedure setControl(o : TWinControl; controlType, controlName:string);
    function findParent(name : string): TWinControl;
  protected
    { Protected declarations }
  public
    fList : TStrings;
    procedure RegisterClass(clsRef : TControlClass);
    property RegisteredClasses[const Name: String]: TControlClass read GetRegisteredClasses write SetRegisteredClasses; default;
    function RTTIText(showDefaults : boolean = false): String;
    procedure saveToFile(fileName:string; showDefaults : boolean = false);
    function loadFromFile(fileName:string):string;
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    property Fields : TStrings read fFields write SetFields;
  end;
procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Tools', [TRTTI]);
end;

{ TRTTI }

constructor TRTTI.Create(AOwner: TComponent); begin inherited; flist := TStringList.Create; fFields := TStringList.Create; end;
destructor TRTTI.Destroy; begin fList.Free; fFields.Free; inherited; end;
procedure TRTTI.SetFields(const Value: Tstrings); begin fFields.Text := Value.Text; end;
procedure TRTTI.saveToFile(fileName: string; showDefaults : boolean); begin with TStringList.Create do begin Text := RTTIText(showDefaults); saveToFile(fileName); free; end; end;
function TRTTI.GetRegisteredClasses(const Name: String): TControlClass; begin Result := TControlClass(fList.Objects[fList.IndexOf(Name)]); end;
procedure TRTTI.RegisterClass(clsRef: TControlClass); begin fList.AddObject(clsRef.ClassName, TObject(clsRef)); end;
procedure TRTTI.SetRegisteredClasses(const Name: String; const Value: TControlClass);
var
  i : integer;
begin
  i := fList.IndexOf(Name);
  if i < 0 then fList.AddObject(Name, TObject(Value)) else fList.Objects[i] := TObject(Value);
end;

function getData(o:TObject; pName:string; showDefaults : boolean): string;
var
  i,count:integer;
  typname:ttypekind;
  s:TStrings;
  tmpObj : TObject;
  propList : PPropList;
  value, txt, header, propName : string;
  pm : PMethod;
  isShowable : boolean;
  function canShow(bShowDef:boolean; obj:TObject; p:PPropInfo): boolean;
  begin
    result := bShowDef;
    if not result then begin
      if p.PropType^.Kind in [tkInteger, tkChar, tkWChar, tkEnumeration, tkSet] then begin
        result := (p.Default <> getOrdProp(o,p.Name));
      end else if p.PropType^.Kind = tkMethod then begin
        result := (getOrdProp(o,p.Name)<>0);
      end else begin
        result := (p.Default = Low(Integer));
      end;
    end;
  end;
begin
  count := getPropList(o, PropList);
  txt := '';
  for i:= 0 to count -1 do try
    value := '';
    propName := propList[i]^.Name;
    typName :=propList[i].PropType^.Kind;
    isShowable := canShow(showDefaults,o,PropList[i]);
    if isShowable then
      if typName in [tkInteger, tkChar, tkWChar] then
        value := intToStr(GetOrdProp(o,propList[i]))
      else if typName = tkEnumeration then
        value := GetEnumProp(o,propList[i])
      else if typName = tkSet then
        value := GetSetProp(o,propList[i])
      else if typName = tkFloat then
        value := floatToStr(GetFloatProp(o,propList[i]))
      else if typName in [tkString, tkLString, tkWString] then begin
        value := fastReplace(GetStrProp(o,propList[i]),'&','&amp;');
        value := fastReplace(value,'<','&lt;');
        value := fastReplace(value,'>','&gt;');
      end else if typName = tkMethod then try
          GetMem(pm,SizeOf(TMethod));
          pm^ := GetMethodProp(o,propList[i]);
          value := TObject(pm.Data).MethodName(pm.Code);
      except
           //
      end else if typName = tkClass then begin
        tmpObj := TObject(getObjectProp(o,propName));
        if tmpObj is TStrings then begin
          s:=TStrings(tmpObj);
          txt := txt + '<component name="' + propName+'"><![CDATA['+s.Text + ']]></component>';
        end else begin
          if tmpObj is TFont then begin // if ParentFont is true then don't save font
            if getEnumProp(o, 'ParentFont') <> 'True' then
              txt := txt + getData(tmpObj, propName, showDefaults);
          end else
            txt := txt + getData(tmpObj, propName, showDefaults);
        end;
      end;

      if isShowable then
      if (typName = tkMethod) and (value <> '') then
        txt := txt + '<property name="' + propName+'" type="'+propList[i].PropType^.Name+'"' + '>'+value + '</property>' +#13#10
      else if typName <> tkClass then
        txt := txt + '<property name="' + propName+'">'+value + '</property>' +#13#10;
  except
    //txt := txt + '<Exception/>';
  end;

  if txt <> '' then begin
    header := '<component name="'+pName+'" type="'+o.ClassName+'"';
    if (o is TControl) and ((o as TControl).HasParent) then
      header := header + ' parent ="'+(TControl(o).Parent.Name)+'"'
    else if (o is TComponent) then
      header := header + ' parent ="'+(TComponent(o).Owner.Name)+'"';
    header := header + '>'+#13#10+txt;
    if (o is TWinControl) and ((o as TWinControl).ControlCount > 0) then begin
      header := header + '<controls>'+#13#10;
      for i := 0 to (o as TWinControl).controlCount -1 do begin
        header := header + '<'+(o as TWinControl).Controls[i].ClassName+' name="'+(o as TWinControl).Controls[i].Name+'" />'+#13#10;
      end;
      header := header + '</controls>'+#13#10;
    end;
    header := header +'</component>'+#13#10;
  end else header := '';
  result:=header;
end;

function TRTTI.RTTIText(showDefaults : boolean): String;
var
  i : integer;
  txt : string;
  c : TComponent;
begin
  txt := '<components>';
  for i := 0 to fFields.Count-1 do begin
    c := Owner.FindComponent(fFields[i]);
    txt := txt + getData(c, c.Name, showDefaults);
  end;
  result := txt + '</components>';
end;

function TRTTI.Process(node : IXMLDOMNode; name : string; var o:TObject; parent:TWinControl): string;
var
   i, j : integer;
   controlNode, curNode : IXMLDOMNode;
   controlNodes : IXMLDOMNodeList;
   controlName, controlType, fieldName : string;
   value : Variant;
   propInfo : PPropInfo;
   tmpObj : TObject;
   pm : Pointer;
   tm : TMethod;
   font : TFont;
begin
  result := '';
  if Assigned(parent) then try TControl(o).Parent := parent; except end;
  for i := 0 to node.childNodes.length - 1 do try
    curNode := node.childNodes[i];
    if curNode.NodeName = 'property' then begin
      fieldName := curNode.attributes[0].nodeValue;
      value := curNode.text;
      propInfo := getPropInfo(o, fieldName);
      result := result + fieldName + ' = ' + value + #13#10;
      if propInfo.PropType^.Kind = tkMethod then begin
        if value <> '' then begin
            pm := owner.MethodAddress(value);
            tm.Code := pm; tm.Data := Pointer(owner);
            setMethodProp(o,fieldName,tm);
        end;
      end
      else if propInfo.PropType^.Kind in [tkInteger, tkChar, tkWChar] then setOrdProp(o,fieldName, strToInt(value))
      else if propInfo.PropType^.Kind in [tkString, tkLString, tkWString] then setStrProp(o,fieldName,value)
      else if propInfo.PropType^.Kind = tkEnumeration then SetEnumProp(o,fieldName,value)
      else if propInfo.PropType^.Kind = tkFloat then SetFloatProp(o,fieldName,strToFloat(value))
      else if propInfo.PropType^.Kind = tkSet then SetSetProp(o,fieldName,value)
      else if propInfo.PropType^.Kind = tkVariant then SetVariantProp(o,fieldName, value)
      else if propInfo.PropType^.Kind = tkInt64 then SetInt64Prop(o,fieldName, strToInt64(value));
    end else if curNode.NodeName = 'component' then begin
      fieldName := curNode.attributes[0].nodeValue;
      result := result + fieldName + ' diving...' + #13#10;
      tmpObj := TObject(getObjectProp(o,fieldName));
      if tmpObj is TStrings then begin
        (tmpObj as TStrings).Text := curNode.text;
        result := result+curNode.text;
      end else begin
        if tmpObj is TFont then try
          setEnumProp(o, 'ParentFont', 'False'); except
        end;
        result := result + Process(curNode, fieldName, tmpObj);
      end;
    end else if curNode.NodeName = 'controls' then begin
      controlNodes := curNode.childNodes;
      for j := 0 to controlNodes.length-1 do begin
        controlType := controlNodes.item[j].nodeName;
        controlName := controlNodes.item[j].attributes.getNamedItem('name').nodeValue;
        setControl(TWinControl(o), controlType, controlName);
        result := result + controlType + '(' +controlName+')'+#13#10;  
      end;
    end;
  except
    result := result + 'Exception'+#13#10;
  end;
end;

function TRTTI.loadFromFile(fileName: string): string;
var
   name, compType, parent, txt : string;
   doc: IXMLDOMDocument;
   nodes : IXMLDOMNodeList;
   node : IXMLDOMNode;
   i : integer;
   obj : TObject;
begin
   doc := CoDOMDocument30.Create;
   doc.async := false;
   if doc.load(fileName) then begin
      nodes := doc.selectNodes('//components/component');
      for i := 0 to nodes.length-1 do begin
         node := nodes[i];
         if node <> nil then begin
            name := node.attributes[0].nodevalue;
            try
              parent := node.attributes.getNamedItem('parent').nodeValue;
            except
              parent := owner.Name;
            end;
            compType := node.attributes[1].nodevalue;
            obj := owner.FindComponent(name);
            if obj = nil then // component new component
               obj := createComponent(compType, name, findParent(parent));
            if obj <> nil then
               txt := txt + name+ #13#10 + Process(node, name, obj, findParent(parent));
         end;
      end;
   end else begin
      txt := Format('errorCode: %d'+#13#10+'Reason: %s'+#13#10+'Line: %d'+#13#10+'LinePos: %d'+#13#10+'FilePos:%d',
             [doc.parseError.errorCode,
              doc.parseError.reason   ,
              doc.parseError.line     ,
              doc.parseError.linepos  ,
              doc.parseError.filepos]);
   end;
   result := txt;
end;

function TRTTI.createComponent(cType, cName: string; parent : TWinControl): TControl;
begin
 result := RegisteredClasses[cType].Create(owner);
 parent.insertControl(result);
 (result as TControl).Name := cName;
end;

procedure TRTTI.setControl(o : TWinControl; controlType, controlName:string);
var
  control:TControl;
begin
  control := TControl(owner.FindComponent(controlName));
  if Assigned(control) then // this control exists, so set its parent...
    control.Parent := o
  else
    control := createComponent(controlType, controlName, o);
end;

function TRTTI.findParent(name: string): TWinControl;
var
  root : TComponent;
  function getName(comp:TComponent; name:string): TComponent;
  var
    i : integer;
  begin
    result := nil;
    for i := 0 to comp.ComponentCount - 1 do
      if (comp.Components[i].Name = name) and (comp.Components[i] is TWinControl) then begin
        result := comp.Components[i];
        break;
      end else if comp.Components[i].ComponentCount > 0 then
        result := getName(comp.Components[i], name);
  end; // getName
begin
  root := self;
  while Assigned(root.Owner)  do
    root := root.Owner;
  result := TWinControl(getName(root, name));
end; // findControl

end.

