unit CHCheckBox;

{ ####################################################################################
  TCHCheckBox

  Version   :   1.0.1
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TCustomControl

  History:
  1.0.0 - 21.07.2002    - First Release
  1.0.1 - 31.08.2002    - BUG: Speicherleck wenn "LoadBitmap", wieder freigeben

  #################################################################################### }

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi,

  _CHClassGraphic, _CHClassControls, _CHClassFunctions, _CHFunction, _CHTypes;

type
  TCHCustomCheckBox = class(TCustomControl)
  private
    FCaptionLayout : TCHCaptionLayout;
    FBox : TCHBoxCB;
    FGlyph : TCHGlyph;
    FBorder : TCHBorder;
    FFill : TCHFillB;
    FFocus : TCHFocus;

    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnClick : TNotifyEvent;

    FClientRect : TRect;
    FWorkRect : TRect;
    FCaptionRect : TRect;
    FColor: TColor;
    FHook : Boolean;
    FCross : Boolean;
    FSolid : Boolean;
    FNormal : Boolean;
    FFlat : Boolean;
    FChecked: Boolean;
    FDefault : Boolean;
    FActive : Boolean;
    FClicksDisabled : Boolean;
    FAllowGrayed: Boolean;
    FTmpColor : TBoxFillColor;
    FStatus : string;
    FState : TCheckBoxState;
    FCIdx : Integer;
    FError : TError;

    procedure SetTextStyle(const Value : TTextStyle);
    procedure SetColor(const Value : TColor);
    procedure SetChecked(const Value : Boolean);
    procedure SetAllowGrayed(const Value : Boolean);
    procedure SetState(const Value : TCheckBoxState);

    procedure GetRectSize;
    procedure MakeBorder(ButtonState : Boolean);
    procedure GetBoxBmp(Style : TBoxStyle; Fill : TCheckBoxFill; Color : TBoxFillColor);
    procedure DoState;
    procedure DrawFocus;

    procedure CMMouseEnter(var Message : TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message : TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message : TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Msg : TMessage); message CM_TEXTCHANGED;
    procedure CMDialogChar(var Message : TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMEnabledChanged(var Message :TMessage); message CM_ENABLEDCHANGED;
    procedure WMEraseBkgnd(var Message : TWMEraseBkgnd); message WM_EraseBkgnd;
    procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure UpdateChanges(Sender : TObject);
  protected
    procedure Click; override;
    procedure Paint; override;
    procedure WndProc(var Message: TMessage); override;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
  published
    property OnMouseEnter : TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave : TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnClick : TNotifyEvent read FOnClick write FOnClick;

    property AllowGrayed : Boolean read FAllowGrayed Write SetAllowGrayed;
    property Box : TCHBoxCB read FBox Write FBox;
    property Border : TCHBorder read FBorder write FBorder;
    property CaptionLayout : TCHCaptionLayout read FCaptionLayout write FCaptionLayout;
    property Checked : Boolean read FChecked Write SetChecked;
    property Color: TColor read FColor write SetColor;
    property Fill : TCHFillB read FFill write FFill;
    property Focus : TCHFocus read FFocus write FFocus;
    property Glyph : TCHGlyph read FGlyph Write FGlyph;
    property State : TCheckBoxState read FState write SetState;
  end;

  TCHCheckBox = class(TCHCustomCheckBox)
  published
    property Action;
    property Anchors;
    property BiDiMode;
    property Caption;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabStop default True;
    property TabOrder;
    property Visible;

    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDockDrop;
    property OnDockOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

procedure Register;

implementation

{$R Checkbox.res}

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHCheckBox]);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHCustomCheckBox.Create(aOwner: TComponent);
begin
  inherited Create(AOwner);

  FBox := TCHBoxCB.Create;
  FBox.OnChange := UpdateChanges;
  FBorder := TCHBorder.Create;
  FBorder.OnChange := UpdateChanges;
  FGlyph := TCHGlyph.Create;
  FGlyph.OnChange := UpdateChanges;
  FCaptionLayout := TCHCaptionLayout.Create;
  FCaptionLayout.OnChange := UpdateChanges;
  FFill := TCHFillB.Create;
  FFill.OnChange := UpdateChanges;
  FFocus := TCHFocus.Create;
  FFocus.OnChange := UpdateChanges;

  Width := 110;
  Height := 20;

  FCaptionLayout.Alignment := tgAuto;

  FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_BLANK');
  FGlyph.AlignMode := gmControl;

  FBorder.HighlightColor := clWhite;
  FBorder.ShadowColor := clGray;
  FBorder.Width := 0;
  FBorder.SingleWidth := 1;

  FCIdx := 4;
  FColor := clBtnFace;
  FStatus := 'none';
  FChecked := False;
  FAllowGrayed := False;
  FTmpColor := bcBlack;
  TabStop := True;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHCustomCheckBox.Destroy;
begin
  DeleteObject(FGlyph.Glyph.Handle);
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.Paint;
begin
  inherited Paint;
  FError := errNone;

  Canvas.Font.Assign(Self.Font);

  // Checkbox Bitmap
  GetBoxBmp(FBox.Style, FBox.Fill, FBox.Color);

  // TextStyle
  SetTextStyle(FCaptionLayout.TextStyle);

  // Rect berechnen
  GetRectSize;

  // Caption
  DrawCaption(Self, Self.Canvas, FCaptionRect, FWorkRect, FClientRect, FCaptionLayout.Angle, FCaptionLayout.HighlightDepth,
      FCaptionLayout.ShadowDepth, FCaptionLayout.HighlightDirection, FCaptionLayout.ShadowDirection,
      FGlyph.TransparentMode, FCaptionLayout.HighlightColor, FCaptionLayout.ShadowColor,
      Self.Color, Self.Font.Color, FGlyph.TransparentColor, Self.Caption, FCaptionLayout.TextStyle,
      FCaptionLayout.Alignment, fsNormal, FGlyph.AlignMode, FGlyph.Alignment, FGlyph.Glyph,
      FGlyph.PosX, FGlyph.PosY, FGlyph.Space, FCaptionLayout.PosX, FCaptionLayout.PosY, FCIdx,
      FFill.Transparent, False, False, FCaptionLayout.Antialiasing, FError);

  // Border zeichnen
  MakeBorder(False);

  // Focus
  if FActive or FFocus.Show then
    DrawFocus;

end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.GetRectSize;
var
  nSingle, nBorderWidth : Integer;
begin
  nSingle := 0;
  nBorderWidth := 0;

  if (FBorder.Style = bsExtended) or (FBorder.Style = bsFlat) then
    nSingle := FBorder.SingleWidth;

  if (FBorder.Style = bsExtended) or (FBorder.Style = bsNormal) then
    nBorderWidth := FBorder.Width;

  // ClientRect
  FClientRect := GetClientRect;

  // WorkRect
  FWorkRect.Left := FClientRect.Left + (nBorderWidth + nSingle);
  FWorkRect.Top := FClientRect.Top + (nBorderWidth + nSingle);
  FWorkRect.Right := FClientRect.Right - (nBorderWidth + nSingle);
  FWorkRect.Bottom := FClientRect.Bottom - (nBorderWidth + nSingle);

  // TextRect
  FCaptionRect.Left := 0;
  FCaptionRect.Top := 0;
  FCaptionRect.Right := 0;
  FCaptionRect.Bottom := 0;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.MakeBorder(ButtonState : Boolean);
var
  nRow, nSingle, I : Integer;
begin
  nSingle := 0;

  with Canvas do
  begin
    Pen.Width := 1;
    Pen.Style := psSolid;
    Brush.Style := bsSolid;

    // Single-Border
    if (FBorder.Style = bsExtended) or (FBorder.Style = bsFlat) then
    begin
      Pen.Color := FBorder.Color;
      for I := 0 to FBorder.SingleWidth - 1 do
      begin
        MoveTo(I, I);
        LineTo(Width - 1 - I, I);
        LineTo(Width - 1 - I, Height - 1 - I);
        LineTo(I, Height - 1 - I);
        LineTo(I, I);
      end;
      nSingle := FBorder.SingleWidth;
    end;

    // Button-Border
    if (FBorder.Style = bsExtended) or (FBorder.Style = bsNormal) then
    begin
      // oben + links
      if (ButtonState = True) then
        Pen.Color := FBorder.ShadowColor
      else
        Pen.Color := FBorder.HighlightColor;

      for nRow := 0 to FBorder.Width - 1 do
      begin
        MoveTo(Width - nRow - 1 - nSingle, nRow + nSingle);
        LineTo(nRow + nSingle, nRow + nSingle);
        LineTo(nRow + nSingle, Height - nRow - 1 - nSingle);
      end;

      // rechts + unten
      if (ButtonState = True) then
        Pen.Color := FBorder.HighlightColor
      else
        Pen.Color := FBorder.ShadowColor;

      for nRow := 0 to FBorder.Width - 1 do
      begin
        MoveTo(Width - nRow - 1 - nSingle, nRow + nSingle);
        LineTo(Width - nRow - 1 - nSingle, Height - nRow - 1 - nSingle);
        LineTo(nRow + nSingle - 1, Height - nRow - 1 - nSingle);
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.DoState;
begin
  if FStatus = 'none' then
  begin
    FStatus := 'checked';
    FChecked := True;
    FState := cbChecked;
  end
  else if FStatus = 'checked' then
  begin
    if FAllowGrayed then
    begin
      FStatus := 'disable';
      FChecked := True;
      FState := cbGrayed;
    end
    else
    begin
      FStatus := 'none';
      FChecked := False;
      FState := cbUnchecked;
    end;
  end
  else if FStatus = 'disable' then
  begin
    FStatus := 'none';
    FChecked := False;
    FState := cbUnchecked;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.DrawFocus;
var
  nWidth, X, Y, PixStep, PosLeft, PosRight, PosTop, PosBottom : Integer;
begin
  // Step
  PixStep := FFocus.Step;

  // Style
  if FFocus.Style = csSolid then
    PixStep := 1;

  // Mode + Pos
  if FFocus.Mode = cmAuto then
  begin
    PosLeft := FCaptionRect.Left - 2;
    PosRight := FCaptionRect.Right + 1;
    PosTop := FCaptionRect.Top - 2;
    PosBottom := FCaptionRect.Bottom;
  end
  else
  begin
    PosLeft := FFocus.PosLeft;
    PosRight := FFocus.PosRight;
    PosTop := FFocus.PosTop;
    PosBottom := FFocus.PosBottom;
  end;

  with Self.Canvas do
  begin
    for nWidth := 0 to FFocus.Width - 1 do
    begin
      // top
      X := PosLeft;
      while X <= PosRight do
      begin
        Pixels[X, PosTop + nWidth] := FFocus.Color;
        Inc(X, PixStep);
      end;
      // bottom
      X := PosLeft;
      while X <= PosRight do
      begin
        Pixels[X, PosBottom - nWidth] := FFocus.Color;
        Inc(X, PixStep);
      end;
      // left
      Y := PosTop;
      while Y <= PosBottom do
      begin
        Pixels[PosLeft + nWidth, Y] := FFocus.Color;
        Inc(Y, PixStep);
      end;
      // right
      Y := PosTop;
      while Y <= PosBottom do
      begin
        Pixels[PosRight - nWidth, Y] := FFocus.Color;
        Inc(Y, PixStep);
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.GetBoxBmp(Style: TBoxStyle; Fill: TCheckBoxFill; Color: TBoxFillColor);
begin
  FNormal := False;
  FFlat := False;
  FHook := False;
  FCross := False;
  FSolid := False;

  FTmpColor := Color;

  if FStatus = 'disable' then
  begin
    FTmpColor := bcGray;
  end
  else if FStatus = 'none' then
  begin
    FTmpColor := bcNone;
  end;

  // NORMAL
  if Style = bxNormal then
  begin
    FNormal := True;
    if FTmpColor = bcNone then
      FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_BlANK')
    else
    begin
      if Fill = cfHook then
      begin
        FHook := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_HOOK_GRAY');
      end
      else if Fill = cfCross then
      begin
        FCross := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_CROSS_GRAY');
      end
      else if Fill = cfSolid then
      begin
        FSolid := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_N_FILL_GRAY');
      end;
    end;
  end
  // FLAT
  else if Style = bxFlat then
  begin
    FFlat := True;
    if FTmpColor = bcNone then
      FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_BlANK')
    else
    begin
      if Fill = cfHook then
      begin
        FHook := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_HOOK_GRAY');
      end
      else if Fill = cfCross then
      begin
        FCross := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_CROSS_GRAY');
      end
      else if Fill = cfSolid then
      begin
        FSolid := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'CB_F_FILL_GRAY');
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.Click;
begin
  inherited;
  DoState;

  if Assigned(FOnClick) then
    FOnClick(Self);

  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.SetChecked(const Value: Boolean);
begin
  if FChecked <> Value then
  begin
    FChecked := Value;
    DoState;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.SetState(const Value: TCheckBoxState);
begin
  if FState <> Value then
  begin
    FState := Value;
    DoState;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.UpdateChanges(Sender: TObject);
begin
  if csLoading in ComponentState then
  begin
    Exit;
  end;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMMouseEnter(var Message: TMessage);
begin
  if Assigned(FOnMouseEnter) then
    FOnMouseEnter(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
  begin
    if IsUnderlineChar(Caption, CharCode) then
    begin
      if Enabled and Assigned(FOnClick) then
        FOnClick(Self);
      Result := 1;
    end
    else
      inherited;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMTextChanged(var Msg: TMessage);
begin
  inherited;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.SetColor(const Value: TColor);
begin
  if FColor <> Value then
  begin
    FColor := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.SetAllowGrayed(const Value: Boolean);
begin
  if FAllowGrayed <> Value then
  begin
    FAllowGrayed := Value;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.SetTextStyle(const Value: TTextStyle);
begin
  with FCaptionLayout do
  begin
    { ssNone }
    if Value = tsNone then
    begin
      ShadowDirection := drNone;
      ShadowDepth := 0;
      ShadowColor := clBtnShadow;

      HighlightDirection := drNone;
      HighlightDepth := 0;
      HighlightColor := clBtnHighlight;
    end
    { ssRaised }
    else if Value = tsRaised then
    begin
      ShadowDirection := drNone;
      ShadowDepth := 0;
      ShadowColor := clBtnShadow;

      HighlightDirection := drUpLeft;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRaisedColor }
    else if Value = tsRaisedColor then
    begin
      ShadowDirection := drDownRight;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;
      Self.Font.Color := clBtnFace;

      HighlightDirection := drUpLeft;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRaisedShadow }
    else if Value = tsRaisedShadow then
    begin
      ShadowDirection := drDownRight;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;

      HighlightDirection := drUpLeft;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRecessed }
    else if Value = tsRecessed then
    begin
      ShadowDirection := drNone;
      ShadowDepth := 0;
      ShadowColor := clBtnShadow;

      HighlightDirection := drDownRight;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRecessedColor }
    else if Value = tsRecessedColor then
    begin
      ShadowDirection := drUpLeft;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;
      Self.Font.Color := clBtnFace;

      HighlightDirection := drDownRight;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRecessedShadow }
    else if Value = tsRecessedShadow then
    begin
      ShadowDirection := drUpLeft;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;

      HighlightDirection := drDownRight;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssShadow }
    else if Value = tsShadow then
    begin
      ShadowDirection := drDownRight;
      ShadowDepth := 2;
      ShadowColor := clBtnShadow;

      HighlightDirection := drNone;
      HighlightDepth := 0;
      HighlightColor := clBtnHighlight;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMFocusChanged(var Message: TCMFocusChanged);
begin
  with Message do
  begin
    if (Sender is TCHCheckBox) then
    begin
      if (FFocus.Active = True) then
      begin
        FActive := Sender = Self;
        Invalidate;
      end
      else
        FActive := FDefault;
    end;
  end;

  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.CMExit(var Message: TCMExit);
begin
  with Message do
  begin
    FActive := False;
    Invalidate;
  end;

  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomCheckBox.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_LBUTTONDOWN, WM_RBUTTONDOWN, WM_LBUTTONDBLCLK:
	    if not (csDesigning in ComponentState) and not Focused then
      begin
		    FClicksDisabled := True;
        Windows.SetFocus(Handle);
		    FClicksDisabled := False;
        if not Focused then
          Exit;
      end;
    CN_COMMAND:
      if FClicksDisabled then
        Exit;
  end;

  inherited WndProc(Message);
end;


end.
