unit CHEdit;

{ ####################################################################################
  TCHEdit

  Version   :   1.0.0
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TEdit

  History:
  1.0.0 - 21.07.2002    - First Release

  #################################################################################### }

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics,

  _CHClassControls, _CHTypes;

type
  TCHCustomEdit = class(TCustomEdit)
  private
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnEnterPress: TKeyEvent;
    FOnESCPress: TKeyEvent;
    FOnAltPress: TKeyEvent;
    FOnStrgPress: TKeyEvent;
    FOnShiftPress: TKeyEvent;

    FFocuscolor : TColor;
    FOrigColor : TColor;
    FAlignment : TAlignment;
    FTextTyp : TTextTyp;
    FSelectMode : TSelectMode;
    FMask : String;
    FDouble : Double;
    FUserKeyEvent : Boolean;
    FSelectPos: Word;
    FSelectLength : Word;
    FCaseMode : TCaseMode;
    FKeyPress : TCHKeyPress;
    FShowCursor: Boolean;
    FBeep : TEditBeep;
    FFocus: Boolean;

    procedure SetFocusColor(const Value: TColor);
    procedure SetAlignment(const Value: TAlignment);
    procedure SetTextTyp(const Value: TTextTyp);
    procedure SetSelectMode(const Value: TSelectMode);
    procedure SetSelectPos(const Value: Word);
    procedure SetSelectLength(const Value: Word);
    procedure SetCaseMode(const Value: TCaseMode);
    procedure SetShowCursor(const Value: Boolean);

    function GetDouble(const Value : string) : Double;

    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure UpdateChanges(Sender: TObject);
  protected
    procedure CreateParams(var params : TCreateParams); override;
    procedure KeyPress(var Key : Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Change; override;
  public
    constructor Create(AOwner : TComponent); override;
  published
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnKeyEnterPress : TKeyEvent read FOnEnterPress write FOnEnterPress;
    property OnKeyESCPress : TKeyEvent read FOnESCPress write FOnESCPress;
    property OnKeyAltPress : TKeyEvent read FOnAltPress write FOnAltPress;
    property OnKeyStrgPress : TKeyEvent read FOnStrgPress write FOnStrgPress;
    property OnKeyShiftPress : TKeyEvent read FOnShiftPress write FOnShiftPress;

    property Alignment : TAlignment read FAlignment Write SetAlignment;
    property BeepOption : TEditBeep read FBeep Write FBeep;
    property CaseMode : TCaseMode read FCaseMode Write SetCaseMode;
    property Focus : Boolean read FFocus Write FFocus;
    property FocusColor : TColor read FFocuscolor Write SetFocusColor;
    property Mask : string read FMask Write FMask;
    property SelectMode : TSelectMode read FSelectMode Write SetSelectMode;
    property SelectLength : Word read FSelectLength Write SetSelectLength;
    property SelectPos : Word read FSelectPos Write SetSelectPos;
    property ShowCursor : Boolean read FShowCursor Write SetShowCursor;
    property TextTyp : TTextTyp read FTextTyp Write SetTextTyp;
    property Keys : TCHKeyPress read FKeyPress Write FKeyPress;
  end;

  TCHEdit = class(TCHCustomEdit)
  protected
    property CharCase;
  published
    property Alignment;
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property BeepOption;
    property BevelEdges;
    property BevelInner;
    property BevelKind;
    property BevelOuter;
    property BiDiMode;
    property BorderStyle;
    property CaseMode;
    property Color;
    property Constraints;
    property Ctl3D;
    property Cursor;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Focuscolor;
    property Font;
    property Height;
    property HelpContext;
    //property HelpKeyword;
    //property HelpType;
    property HideSelection;
    property Hint;
    property ImeMode;
    property ImeName;
    property Keys;
    property Left;
    property Mask;
    property MaxLength;
    property Name;
    property OEMConvert;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property SelectLength;
    property SelectMode;
    property SelectPos;
    property ShowCursor;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Tag;
    property Text;
    property TextTyp;
    property Top;
    property Visible;
    property Width;

    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHEdit]);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHCustomEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FKeyPress := TCHKeyPress.Create;
  FKeyPress.OnChange := UpdateChanges;

  FShowCursor := True;
  FFocusColor := clInfoBk;
  FFocus := False;
  FBeep := ebStandard;
  FAlignment := taLeftJustify;
  Mask := '#,##0.00'
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.UpdateChanges(Sender: TObject);
begin
  if csLoading in ComponentState then
  begin
    Exit;
  end;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.CreateParams(var params: TCreateParams);
const
  Alignments: array[TAlignment] of Cardinal = (ES_LEFT,ES_RIGHT,ES_CENTER);
begin
  inherited CreateParams(Params);
    Params.Style := Params.Style or Alignments[FAlignment];
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.CMEnter(var Message: TCMEnter);
begin
  inherited;
  FOrigColor := Color;
  if FFocus then
  begin
    Color := FFocuscolor;
  end;

  if FSelectMode = smAll then
    SelectAll
  else if FSelectMode = smFirst then
  begin
    SelStart := 0;
    SelLength := 0;
  end
  else if FSelectMode = smLast then
  begin
    SelStart := Length(Text);
    SelLength := 0;
  end
  else if FSelectMode = smCustom then
  begin
    SelStart := FSelectPos;
    SelLength := FSelectLength;
  end;

  if FShowCursor = False then
    HideCaret(Self.Handle)
  else
    ShowCaret(Self.Handle);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.CMExit(var Message: TCMExit);
begin
  Color := FOrigColor;

  if FTextTyp = ttDouble then
  begin
    FDouble := GetDouble(Text);
    Text := FormatFloat(Mask, FDouble);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.CMMouseEnter(var Message: TMessage);
begin
    if Assigned(FOnMouseEnter) then
    FOnMouseEnter(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetFocusColor(const Value: TColor);
begin
  if FFocuscolor <> Value then
  begin
    FFocuscolor := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetAlignment(const Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    RecreateWnd;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetTextTyp(const Value: TTextTyp);
begin
  if FTextTyp <> Value then
  begin
    FTextTyp := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetSelectMode(const Value: TSelectMode);
begin
  if FSelectMode <> Value then
  begin
    FSelectMode := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetSelectPos(const Value: Word);
begin
  if FSelectPos <> Value then
  begin
    FSelectPos := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetSelectLength(const Value: Word);
begin
  if FSelectLength <> Value then
  begin
    FSelectLength := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetCaseMode(const Value: TCaseMode);
begin
  if FCaseMode <> Value then
  begin
    FCaseMode := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.SetShowCursor(const Value: Boolean);
begin
  if FShowCursor <> Value then
  begin
    FShowCursor := Value;
    if FShowCursor = False then
      HideCaret(Self.Handle)
    else
      ShowCaret(Self.Handle);
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHCustomEdit.GetDouble(const Value: string): Double;
var
  nChar : Integer;
  sDouble : string;
begin
  Result := 0;
  sDouble := '';
  for nChar := 1 to Length(Value) do
  begin
    case Value[nChar] of
      '0'..'9' : sDouble := sDouble + Value[nChar];
      '+','-'  : if nChar = 1 then
                   sDouble := sDouble + Value[nChar];
    end;
    if Value[nChar] = DecimalSeparator then
      sDouble := sDouble + Value[nChar];
  end;

  if sDouble <> '' then
    Result := StrToFloat(sDouble);
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.KeyPress(var Key: Char);
var
  bOk : Boolean;
  nNull : Integer;
begin
  bOk := False;
  nNull := GetSelStart;

  if not FUserKeyEvent then
  begin
    if FTextTyp = ttNormal then
      bOk := True;

    // Integer
    if FTextTyp = ttInteger then
    begin
      case Key of
        '0'..'9' : bOk := True;
        '+','-'  : if nNull = 0 then
                     bOk := True;
      end;

      if (Ord(Key) = VK_BACK) then
        bOk := True;

      if not bOk then
      begin
        Beep;
        Abort;
      end;
    end
    // Double
    else if FTextTyp = ttDouble then
    begin
      case Key of
        '0'..'9' : bOk := True;
        '.',','  : bOk := True;
        '+','-'  : if nNull = 0 then
                     bOk := True;
      end;

      if (Key = DecimalSeparator) or (Ord(Key) = VK_BACK) then
        bOk := True;
    end;

    // Eingabe nicht zulassen
    if not bOk then
    begin
      Beep;
      Abort;
    end;
  end;

  // Beepoption
  if FBeep = ebNoEnterBeep then
  begin
    if Key = #13 then
      Key := #0;
  end
  else if FBeep = ebAlwaysBeep then
  begin
    Beep;
  end;

  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  FUserKeyEvent := False;

  // Alt Key
  if (Key = VK_MENU	) and (FKeyPress.Alt) then
  begin
    FUserKeyEvent := True;
    if Assigned(FOnAltPress) then
      FOnAltPress(Self, Key, Shift)
    else
      Key := vk_Clear;
    Exit;
  end;

  // Enter Key
  if (Key = vk_Return) and (FKeyPress.Enter) then
  begin
    FUserKeyEvent := True;
    if Assigned(FOnEnterPress) then
      FOnEnterPress(Self, Key, Shift)
    else
      Key := vk_Clear;
    Exit;
  end;

  // ESC Key
  if (Key = vk_Escape) and (FKeyPress.ESC) then
  begin
    FUserKeyEvent := True;
    if Assigned(FOnESCPress) then
      FOnESCPress(Self, Key, Shift)
    else
      Key := vk_Clear;
    Exit;
  end;

  // Strg Key
  if (Key = VK_CONTROL) and (FKeyPress.Strg) then
  begin
    FUserKeyEvent := True;
    if Assigned(FOnStrgPress) then
      FOnStrgPress(Self, Key, Shift)
    else
      Key := vk_Clear;
    Exit;
  end;

  // Shift Key
  if (Key = VK_SHIFT	) and (FKeyPress.Shift) then
  begin
    FUserKeyEvent := True;
    if Assigned(FOnShiftPress) then
      FOnShiftPress(Self, Key, Shift)
    else
      Key := vk_Clear;
    Exit;
  end;

  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomEdit.Change;
var
  nCaretPos : Word;
  sFirst, sText : String;
begin

  if not FUserKeyEvent then
  begin
    if FTextTyp = ttNormal then
    begin
      if FCaseMode = cmNormal then
      begin
        Self.CharCase := ecNormal;
      end
      else if FCaseMode = cmUppercase then
      begin
        Self.CharCase := ecUpperCase;
      end
      else if FCaseMode = cmLowercase then
      begin
        Self.CharCase := ecLowerCase;
      end
      else if FCaseMode = cmFirstUpper then
      begin
        if Text <> '' then
        begin
          nCaretPos := Self.SelStart;
          sText := Text;
          sFirst := AnsiUpperCase(Text[1]);
          Delete(sText, 1, 1);
          Text := sFirst + sText;
          Self.SelStart := nCaretPos;
        end;
      end
    end;
  end;

  inherited Change;

end;

end.
