unit CHImage;

{ ####################################################################################
  TCHImage (Beta)

  Version   :   0.8.0 (Beta)
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TGraphicControl

  History:
  0.8.0 - 21.07.2002    - First Release

  #################################################################################### }

interface

uses
  Forms, Windows, Messages, SysUtils, Classes, Controls, ExtCtrls, Graphics, ShellApi,
  Buttons, jpeg,

  _CHClassGraphic, _CHClassControls, _CHClassFunctions, _CHFunction, _CHTypes;

type
  TCHCustomImage = class;

  TCHFrame = class(TPersistent)
  private
    FOwner : TCHCustomImage;
    FFrameActive : Boolean;
    FFrameColor : TColor;
    FFrameStyle : TFrameStyle;
    FFrameWidth :  Integer;

    procedure SetFrameActive(const Value : Boolean);
    procedure SetFrameColor(const Value : TColor);
    procedure SetFrameStyle(const Value : TFrameStyle);
    procedure SetFrameWidth(const Value : Integer);
  protected

  public
    constructor Create(AOwner : TCHCustomImage); virtual;
  published
    property Active : Boolean read FFrameActive write SetFrameActive;
    property Color : TColor read FFrameColor write SetFrameColor;
    property Style : TFrameStyle read FFrameStyle write SetFrameStyle;
    property Width : Integer read FFrameWidth write SetFrameWidth;
  end;

  //***********************************************************************************************************

  TCHPicture = class(TPersistent)
  private
    FOwner : TCHCustomImage;
    FPicture : TPicture;
    FGraphicMode : TBitmapMode;
    FTransColor: TColor;
    FTransMode: TTransMode;

    procedure SetPicture(const Value : TPicture);
    procedure SetGraphicMode(const Value : TBitmapMode);
    procedure SetTransColor(const Value: TColor);
    procedure SetTransMode(const Value: TTransMode);
  protected

  public
    constructor Create(AOwner : TCHCustomImage); virtual;
  published
    property Graphic : TPicture read FPicture write SetPicture;
    property Mode : TBitmapMode read FGraphicMode write SetGraphicMode;
    property TransparentColor : TColor read FTransColor Write SetTransColor;
    property TransparentMode : TTransMode read FTransMode Write SetTransMode;
  end;

  //***********************************************************************************************************

  TCHRotation = class(TPersistent)
  private
    FOwner : TCHCustomImage;
    FAngle: TAngle;
    FBackcolor: TColor;
    FxAxis: Word;
    FyAxis: Word;
    procedure SetAngle(const Value: TAngle);
    procedure SetBackcolor(const Value: TColor);
    procedure SetxAxis(const Value: Word);
    procedure SetyAxis(const Value: Word);
  protected

  public
    constructor Create(AOwner : TCHCustomImage); virtual;
  published
    property Angle : TAngle read FAngle Write SetAngle;
    property xAxis : Word read FxAxis Write SetxAxis;
    property yAxis : Word read FyAxis Write SetyAxis;
    property Backcolor : TColor read FBackcolor Write SetBackcolor;
  end;

  //***********************************************************************************************************

  TCHFade = class(TPersistent)
  private
    FOwner : TCHCustomImage;
    FFadeActive : Boolean;
    FFadeInterval : Word;
    FFadeMode : TFadeMode;
    FFadeLoop : Word;
    procedure SetFadeInterval(const Value: Word);
    procedure SetFadeLoop(const Value: Word);
    procedure SetFadeMode(const Value: TFadeMode);
    procedure SetFadeActive(const Value: Boolean);
  protected

  public
    constructor Create(AOwner : TCHCustomImage); virtual;
  published
    property Active : Boolean read FFadeActive Write SetFadeActive;
    property Interval : Word read FFadeInterval Write SetFadeInterval;
    property Mode : TFadeMode read FFadeMode Write SetFadeMode;
    property Loops : Word read FFadeLoop Write SetFadeLoop;
  end;

  //***********************************************************************************************************

  TCHCustomImage = class(TGraphicControl)
  private
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FPic : TCHPicture;
    FFrame : TCHFrame;
    FFade : TCHFade;
    FRotation : TCHRotation;
    FOnProgress: TProgressEvent;
    FShowDesignFrame : Boolean;
    FIncrementalDisplay: Boolean;
    FTransparent: Boolean;
    FDrawing: Boolean;
    FProportional: Boolean;
    FAutosize : Boolean;
    FImageRect : TRect;
    FClientRect : TRect;
    FBitmap : TBitmap;
    FRotateBmp : TBitmap;
    FOriginalBmp : TBitmap;
    FTmpBmp : TBitmap;

    procedure DrawFrame;
    procedure DrawFrameDesignTime;
    procedure GetImageRect;
    procedure DoAutosize(const Value: Boolean);

    function GetCanvas: TCanvas;
    procedure PictureChanged(Sender: TObject);
    procedure SetTransparent(const Value: Boolean);
    procedure SetProportional(Value: Boolean);

    procedure CMMouseEnter(var Message : TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message : TMessage); message CM_MOUSELEAVE;
    procedure SetShowDesignFrame(const Value: Boolean);
  protected
    function CanAutoSize(var NewWidth, NewHeight: Integer): Boolean; override;
    function DestRect: TRect;
    function DoPaletteChange: Boolean;
    function GetPalette: HPALETTE; override;
    procedure Paint; override;
    procedure Resize; override;
    procedure Progress(Sender: TObject; Stage: TProgressStage;
      PercentDone: Byte; RedrawNow: Boolean; const R: TRect; const Msg: string); dynamic;

    procedure FadeIn(Bitmap : TBitmap; Interval : Word);
    procedure FadeOut(Bitmap : TBitmap; Interval : Word);
    procedure RotateBitmap(var OriginalBmp, RotatedBmp : TBitmap; Angle : Double;
      xAxis, yAxis : Integer; BackColor : TColor);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Canvas: TCanvas read GetCanvas;
  published
    property OnMouseEnter : TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave : TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;

    property Autosize : Boolean read FAutosize Write DoAutosize;
    property Fade : TCHFade read FFade Write FFade;
    property Frame : TCHFrame read FFrame Write FFrame;
    property Picture : TCHPicture read FPic Write FPic;
    property Rotation : TCHRotation read FRotation Write FRotation;
    property IncrementalDisplay: Boolean read FIncrementalDisplay write FIncrementalDisplay;
    property Proportional: Boolean read FProportional write SetProportional;
    property ShowDesignFrame : Boolean read FShowDesignFrame write SetShowDesignFrame;
    property Transparent: Boolean read FTransparent write SetTransparent;
  end;

  //***********************************************************************************************************

  TCHImage = class(TCHCustomImage)
  published
    property Align;
    property Anchors;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;

    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

uses consts;

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHImage]);
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHCustomImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FPic := TCHPicture.Create(Self);
  FFrame := TCHFrame.Create(Self);
  FRotation := TCHRotation.Create(Self);
  FFade := TCHFade.Create(self);
  FBitmap := TBitmap.Create;

  ControlStyle := ControlStyle + [csReplicatable];
  FPic.FPicture.OnChange := PictureChanged;
  FPic.FPicture.OnProgress := Progress;
  FShowDesignFrame := True;
  Height := 105;
  Width := 105;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHCustomImage.Destroy;
begin
  FBitmap.Free;
  FPic.FPicture.Free;
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHCustomImage.GetPalette: HPALETTE;
begin
  Result := 0;
  if FPic.FPicture.Graphic <> nil then
	  Result := FPic.FPicture.Graphic.Palette;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHCustomImage.DestRect: TRect;
var
  w, h, cw, ch: Integer;
  xyaspect: Double;
begin
  w := FPic.FPicture.Width;
  h := FPic.FPicture.Height;
  cw := ClientWidth;
  ch := ClientHeight;
  if (FPic.Mode = bmStretch) or (Proportional and ((w > cw) or (h > ch))) then
  begin
	if Proportional and (w > 0) and (h > 0) then
	begin
      xyaspect := w / h;
      if w > h then
      begin
        w := cw;
        h := Trunc(cw / xyaspect);
        if h > ch then  // woops, too big
        begin
          h := ch;
          w := Trunc(ch * xyaspect);
        end;
      end
      else
      begin
        h := ch;
        w := Trunc(ch * xyaspect);
        if w > cw then  // woops, too big
        begin
          w := cw;
          h := Trunc(cw / xyaspect);
        end;
      end;
    end
    else
    begin
      w := cw;
      h := ch;
    end;
  end;

  with Result do
  begin
    Left := 0;
    Top := 0;
    Right := w;
    Bottom := h;
  end;

  if FPic.Mode = bmCenter then
	  OffsetRect(Result, (cw - w) div 2, (ch - h) div 2);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.Paint;
var
  Save: Boolean;
  X, Y, LCount : Integer;
begin
  if FFrame.Active then
  begin
    X := FFrame.FFrameWidth;
    Y := FFrame.FFrameWidth;
  end
  else
  begin
    X := 0;
    Y := 0;
  end;

  // Imageflche
  GetImageRect;

  // Rahmen DesigneTime
  if csDesigning in ComponentState then
    DrawFrameDesignTime;

  // Image zuweisen
  if not (FPic.FPicture.Graphic = nil) then
  begin
    FBitmap.Assign(FPic.FPicture.Graphic);

    // Rotation
    if FRotation.Angle > 0 then
    begin
      try
        FOriginalBmp := TBitmap.Create;
        FRotateBmp := TBitmap.Create;
        FTmpBmp := TBitmap.Create;

        FOriginalBmp.Assign(FBitmap);
        RotateBitmap(FOriginalBmp, FRotateBmp, FRotation.FAngle,
          FRotation.FxAxis, FRotation.FyAxis, FRotation.FBackcolor);
        FBitmap.Assign(FRotateBmp);
      finally
        FRotateBmp.Free;
        FOriginalBmp.Free;
      end;
    end;

    // Transparent
    with FBitmap do
    begin
      if FTransparent then
      begin
        if FPic.FTransMode = tmCustom then
          TransparentColor := FPic.FTransColor
        else
          TransparentColor := FBitmap.Canvas.Pixels[0, FBitmap.Height - 1];
        Transparent := True;
      end
      else
        Transparent := False;
    end;

    Save := FDrawing;
    FDrawing := True;
    with inherited Canvas do
    begin
      // Normal
      if FPic.Mode = bmNormal then
        Draw(X, Y, FBitmap)
      // Stretch
      else if FPic.Mode = bmStretch then
      begin
        StretchDraw(FImageRect, FBitmap);
      end
      // Tile
      else if FPic.Mode = bmTile then
      begin
        Brush.Bitmap := FBitmap;
        FillRect(FImageRect);
      end
      // Center
      else if FPic.Mode = bmCenter then
      begin
        X := ((FImageRect.Right - FImageRect.Left) div 2) - (FBitmap.Width div 2) + FFrame.FFrameWidth;
        Y := ((FImageRect.Bottom - FImageRect.Top) div 2) - (FBitmap.Height div 2) + FFrame.FFrameWidth;
        Draw(X, Y, FBitmap)
      end
    end;
    FDrawing := Save;
  end;

  // Frame
  DrawFrame;

  // Fade
  if FFade.Active then
  begin
    for LCount := 1 to FFade.Loops do
    begin
      Application.ProcessMessages;
      if FFade.Mode = fmIn then
      begin
        FBitmap.Assign(FPic.FPicture);
        FadeIn(FBitmap, FFade.Interval);
      end
      else if FFade.Mode = fmOut then
      begin
        FBitmap.Assign(FPic.FPicture);
        FadeOut(FBitmap, FFade.Interval);
      end
      else if FFade.Mode = fmInOut then
      begin
        FBitmap.Assign(FPic.FPicture);
        FadeIn(FBitmap, FFade.Interval);
        FBitmap.Assign(FPic.FPicture);
        FadeOut(FBitmap, FFade.Interval);
      end
      else if FFade.Mode = fmOutIn then
      begin
        FBitmap.Assign(FPic.FPicture);
        FadeOut(FBitmap, FFade.Interval);
        FBitmap.Assign(FPic.FPicture);
        FadeIn(FBitmap, FFade.Interval);
      end;
    end;
    FFade.Active := False;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHCustomImage.DoPaletteChange: Boolean;
var
  ParentForm: TCustomForm;
  Tmp: TGraphic;
begin
  Result := False;
  Tmp := FPic.FPicture.Graphic;
  if Visible and (not (csLoading in ComponentState)) and (Tmp <> nil) and
	(Tmp.PaletteModified) then
  begin
	if (Tmp.Palette = 0) then
	  Tmp.PaletteModified := False
	else
	begin
	  ParentForm := GetParentForm(Self);
	  if Assigned(ParentForm) and ParentForm.Active and Parentform.HandleAllocated then
	  begin
		if FDrawing then
		  ParentForm.Perform(wm_QueryNewPalette, 0, 0)
		else
		  PostMessage(ParentForm.Handle, wm_QueryNewPalette, 0, 0);
		Result := True;
		Tmp.PaletteModified := False;
	  end;
	end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.Progress(Sender: TObject; Stage: TProgressStage;
  PercentDone: Byte; RedrawNow: Boolean; const R: TRect; const Msg: string);
begin
  if FIncrementalDisplay and RedrawNow then
  begin
	if DoPaletteChange then
    Update
	else
    Paint;
  end;
  if Assigned(FOnProgress) then
    FOnProgress(Sender, Stage, PercentDone, RedrawNow, R, Msg);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHCustomImage.GetCanvas: TCanvas;
var
  Bitmap: TBitmap;
begin
  if FPic.FPicture.Graphic = nil then
  begin
	Bitmap := TBitmap.Create;
	try
	  Bitmap.Width := Width;
	  Bitmap.Height := Height;
	  FPic.FPicture.Graphic := Bitmap;
	finally
	  Bitmap.Free;
	end;
  end;
  if FPic.FPicture.Graphic is TBitmap then
	  Result := TBitmap(FPic.FPicture.Graphic).Canvas
  else
	raise EInvalidOperation.Create(SImageCanvasNeedsBitmap);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.SetTransparent(const Value: Boolean);
begin
  if Value <> FTransparent then
  begin
	  FTransparent := Value;
	  PictureChanged(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.SetProportional(Value: Boolean);
begin
  if FProportional <> Value then
  begin
	  FProportional := Value;
	  PictureChanged(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.PictureChanged(Sender: TObject);
var
  G: TGraphic;
begin
  if AutoSize and (FPic.FPicture.Width > 0) and (FPic.FPicture.Height > 0) then
	  SetBounds(Left, Top, FPic.FPicture.Width, FPic.FPicture.Height);
  G := FPic.FPicture.Graphic;
  if G <> nil then
  begin
	if not ((G is TMetaFile) or (G is TIcon)) then
	  G.Transparent := FTransparent;
	if (not G.Transparent) and (FPic.Mode = bmStretch) and not Proportional then
	  ControlStyle := ControlStyle + [csOpaque]
	else  // picture might not cover entire clientrect
	  ControlStyle := ControlStyle - [csOpaque];
	if DoPaletteChange and FDrawing then
    Update;
  end
  else
    ControlStyle := ControlStyle - [csOpaque];
  if not FDrawing then
    Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHCustomImage.CanAutoSize(var NewWidth, NewHeight: Integer): Boolean;
begin
  Result := True;
  if not (csDesigning in ComponentState) or (FPic.FPicture.Width > 0) and
    (FPic.FPicture.Height > 0) then
  begin
    if Align in [alNone, alLeft, alRight] then
      NewWidth := FPic.FPicture.Width;
    if Align in [alNone, alTop, alBottom] then
      NewHeight := FPic.FPicture.Height;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.GetImageRect;
begin
  FClientRect := GetClientRect;
  if FFrame.Active then
  begin
    try
      FImageRect.Left := FClientRect.Left + FFrame.FFrameWidth;
      FImageRect.Top :=  FClientRect.Top + Frame.FFrameWidth;
      FImageRect.Right := FClientRect.Right - FFrame.FFrameWidth;
      FImageRect.Bottom := FClientRect.Bottom - FFrame.FFrameWidth;
    except
      FImageRect := ClientRect;
    end;
  end
  else
  begin
    FImageRect := GetClientRect;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.RotateBitmap(var OriginalBmp, RotatedBmp : TBitmap; Angle : Double;
  xAxis, yAxis : Integer; BackColor : TColor);
const
    MaxPixelCount = 32768;
    Black : RGBTriple = (rgbtBlue:0; rgbtGreen:0; rgbtRed:0);

type
  pRGBArray = ^TRGBArray;
  TRGBArray = array[0..MaxPixelCount - 1] of TRGBTriple;

var
  BW, BH, i, j, x, y : Word;
  cosTheta, sinTheta, Theta, jPrimeSinTheta, jPrimeCosTheta :  DOUBLE;
  iOriginal, iPrime, iPrimeRotated, jOriginal, jPrime, jPrimeRotated :  INTEGER;
  ScanlineBytes, iRot, jRot : Integer;
  RowOriginal, RowRotated : pRGBArray;
  POriginalStart, POriginal : Pointer;
  TmpBmp : TBitmap;
  BRect : TRect;
begin
  BRect := Bounds(0, 0, Self.Width, Self.Height);

  TmpBmp := TBitmap.Create;
  TmpBmp.Width := Self.Width;
  TmpBmp.Height := Self.Height;

  TmpBmp.Canvas.Brush.Color := BackColor;
  TmpBmp.Canvas.Brush.Style := bsSolid;
  TmpBmp.Canvas.FillRect(BRect);
  X := ((FImageRect.Right - FImageRect.Left) div 2) - (FBitmap.Width div 2) + FFrame.FFrameWidth;
  Y := ((FImageRect.Bottom - FImageRect.Top) div 2) - (FBitmap.Height div 2) + FFrame.FFrameWidth;
  TmpBmp.Canvas.Draw(X, Y, OriginalBmp);

  OriginalBmp.Assign(TmpBmp);
  TmpBmp.Free;

  if OriginalBmp.PixelFormat <> pf24bit then
    OriginalBmp.PixelFormat := pf24bit;

  RotatedBmp.Width  := Self.Width;
  RotatedBmp.Height := Self.Height;
  RotatedBmp.Canvas.Brush.Color := BackColor;
  RotatedBmp.Canvas.Brush.Style := bsSolid;
  RotatedBmp.Canvas.FillRect(BRect);

  //xAxis := Self.Width div 2;
  //yAxis := Self.Height div 2;

  RotatedBmp.PixelFormat := OriginalBmp.PixelFormat;

  Theta := -Angle * PI / 180;
  sinTheta := SIN(Theta);
  cosTheta := COS(Theta);

  ScanlineBytes := Integer(OriginalBmp.Scanline[1]) - Integer(OriginalBmp.Scanline[0]);

  BW := OriginalBmp.Width  - 1;
  BH := OriginalBmp.Height - 1;
  iRot := (2 * xAxis) + 1;
  jRot := (2 * YAxis) + 1;

  RowRotated := RotatedBmp.ScanLine[BH];
  POriginalStart := OriginalBmp.ScanLine[0];

  for j := BH downto 0 do
  begin

    jPrime := (2 * j) - jRot;
    jPrimeSinTheta := jPrime * sinTheta;
    jPrimeCosTheta := jPrime * cosTheta;
    POriginal := POriginalStart;
    for i := BW downto 0 do
    begin

      iPrime := (2 * i) - iRot;
      iPrimeRotated := ROUND(iPrime * cosTheta - jPrimeSinTheta);
      iOriginal := (iPrimeRotated - 1) div 2 + xAxis;
      if (iOriginal >= 0) and (iOriginal <= BW) then
      begin
        jPrimeRotated := ROUND(iPrime * sinTheta + jPrimeCosTheta);
        jOriginal := (jPrimeRotated - 1) div 2 + yAxis;
        if (jOriginal >= 0) and (jOriginal <= BH) then
        begin
          RowOriginal   := Pointer(Integer(POriginal) + (jOriginal * ScanLineBytes));
          RowRotated[i] := RowOriginal[iOriginal];
        end
        else
//          RowRotated[i].rgbtBlue  := GetBValue(BackColor);
//          RowRotated[i].rgbtGreen := GetGValue(BackColor);
//          RowRotated[i].rgbtRed   := GetRValue(BackColor);

          //RowRotated[i] := Black;
      end
      else
//        RowRotated[i].rgbtBlue  := GetBValue(BackColor);
//        RowRotated[i].rgbtGreen := GetGValue(BackColor);
//        RowRotated[i].rgbtRed   := GetRValue(BackColor);

        //RowRotated[i] := Black;
    end;
    Dec(Integer(RowRotated), ScanLineBytes);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.DrawFrameDesignTime;
var
  X, Y, PixStep : Integer;
begin
  PixStep := 2;
  if FShowDesignFrame then
  begin
    with inherited Canvas do
    begin
      // top
      X := 0;
      while X < Self.Width -1 do
      begin
        Pixels[X, 0] := clGray;
        Inc(X, PixStep);
      end;
      // bottom
      X := 0;
      while X < Self.Width -1 do
      begin
        Pixels[X, (Self.Height -1)] := FFrame.Color;
        Inc(X, PixStep);
      end;
      // left
      Y := 0;
      while Y < Self.Height -1 do
      begin
        Pixels[0, Y] := FFrame.Color;
        Inc(Y, PixStep);
      end;
      // right
      Y := 0;
      while Y < Self.Height -1 do
      begin
        Pixels[(Self.Width -1), Y] := FFrame.Color;
        Inc(Y, PixStep);
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.DrawFrame;
var
  X, Y, PixStep, nWidth : Integer;
begin
  PixStep := 1;
  if FFrame.Active then
  begin
    if FFrame.Style = fsDot then
      PixStep := 2
    else if FFrame.Style = fsSolid then
      PixStep := 1;

    with inherited Canvas do
    begin
      for nWidth := 0 to FFrame.Width - 1 do
      begin
        // top
        X := 0;
        while X < Self.Width -1 do
        begin
          Pixels[X, nWidth] := FFrame.Color;
          Inc(X, PixStep);
        end;
        // bottom
        X := 0;
        while X < Self.Width -1 do
        begin
          Pixels[X, (Self.Height -1) - nWidth] := FFrame.Color;
          Inc(X, PixStep);
        end;
        // left
        Y := 0;
        while Y < Self.Height -1 do
        begin
          Pixels[nWidth, Y] := FFrame.Color;
          Inc(Y, PixStep);
        end;
        // right
        Y := 0;
        while Y < Self.Height -1 do
        begin
          Pixels[(Self.Width -1) - nWidth, Y] := FFrame.Color;
          Inc(Y, PixStep);
        end;
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.FadeIn(Bitmap : TBitmap; Interval : Word);
type
  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array[0..32767] of TRGBTriple;
var
  OriginalBmp : TBitmap;
  I, y, x : Word;
  Line, OrigLine : PRGBTripleArray;
begin
  OriginalBmp := TBitmap.Create;
  try
    OriginalBmp.PixelFormat := pf32bit;
    OriginalBmp.Assign(Bitmap);
    // Start Fading
    for I := 0 to 32 do
    begin
      for y := 0 to (Bitmap.Height - 1) do
      begin
        OrigLine := OriginalBmp.ScanLine[y];
        Line := Bitmap.ScanLine[y];
        for x := 0 to (Bitmap.Width - 1) do
        begin
          Line[x].rgbtRed := (I * origLine[x].rgbtRed) shr 5;
          Line[x].rgbtGreen := (I * origLine[x].rgbtGreen) shr 5;
          Line[x].rgbtBlue := (I * origLine[x].rgbtBlue) shr 5;
        end;
      end;
      Sleep(Interval);
      Self.Canvas.Draw(1, 1, Bitmap);
      Application.ProcessMessages;
    end;
  finally
    OriginalBmp.Free;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.FadeOut(Bitmap : TBitmap; Interval : Word);
type
  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array[0..32767] of TRGBTriple;
var
  OriginalBmp : TBitmap;
  I, y, x : Word;
  Line, OrigLine : PRGBTripleArray;
begin
  OriginalBmp := TBitmap.Create;
  try
    OriginalBmp.PixelFormat := pf32bit;
    OriginalBmp.Assign(Bitmap);
    // Start Fading
    for I := 32 downto 0 do
    begin
      for y := 0 to (Bitmap.Height - 1) do
      begin
        OrigLine := OriginalBmp.ScanLine[y];
        Line := Bitmap.ScanLine[y];
        for x := 0 to (Bitmap.Width - 1) do
        begin
          Line[x].rgbtRed := (I * origLine[x].rgbtRed) shr 5;
          Line[x].rgbtGreen := (I * origLine[x].rgbtGreen) shr 5;
          Line[x].rgbtBlue := (I * origLine[x].rgbtBlue) shr 5;
        end;
      end;
      Sleep(Interval);
      Self.Canvas.Draw(1, 1, Bitmap);
      Application.ProcessMessages;
    end;
  finally
    OriginalBmp.Free;

  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.SetShowDesignFrame(const Value: Boolean);
begin
  if FShowDesignFrame <> Value then
  begin
    FShowDesignFrame := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.DoAutosize(const Value: Boolean);
begin
  FAutosize := Value;
  if FAutosize and (FPic.FPicture.Width > 0) and (FPic.FPicture.Height > 0) then
  begin
    SetBounds(Left, Top, FPic.FPicture.Width, FPic.FPicture.Height);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.Resize;
begin
  inherited Resize;
  FRotation.FxAxis := Width div 2;
  FRotation.FyAxis := Height div 2;
  DoAutosize(FAutosize);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
//procedure TCHCustomImage.SetTransparent(const Value: Boolean);
//begin
//  if FTransparent <> Value then
//  begin
//    FTransparent := Value;
//    Invalidate;
//  end;
//end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.CMMouseEnter(var Message: TMessage);
begin
  if Assigned(FOnMouseEnter) then
    FOnMouseEnter(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomImage.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(self);
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ ---  CHFrame  ---}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

constructor TCHFrame.Create(AOwner: TCHCustomImage);
begin
  inherited Create;
  FOwner := AOwner;
  FFrameActive := False;
  FFrameColor := clBlack;
  FFrameWidth := 1;
  FFrameStyle := fsSolid;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFrame.SetFrameActive(const Value: Boolean);
begin
  if FFrameActive <> Value then
  begin
    FFrameActive := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFrame.SetFrameColor(const Value: TColor);
begin
  if FFrameColor <> Value then
  begin
    FFrameColor := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFrame.SetFrameStyle(const Value: TFrameStyle);
begin
  if FFrameStyle <> Value then
  begin
    FFrameStyle := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFrame.SetFrameWidth(const Value: Integer);
begin
  if FFrameWidth <> Value then
  begin
    FFrameWidth := Value;
    FOwner.Invalidate;
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ ---  CHPicture  ---}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

constructor TCHPicture.Create(AOwner: TCHCustomImage);
begin
  inherited Create;
  FOwner := AOwner;
  FPicture := TPicture.Create;
  FGraphicMode := bmNormal;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPicture.SetPicture(const Value: TPicture);
begin
  FPicture.Assign(Value);
  FOwner.FRotation.FxAxis := FOwner.Width div 2;
  FOwner.FRotation.FyAxis := FOwner.Height div 2;
  FOwner.Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPicture.SetGraphicMode(const Value: TBitmapMode);
begin
  if FGraphicMode <> Value then
  begin
    FGraphicMode := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPicture.SetTransColor(const Value: TColor);
begin
  if FTransColor <> Value then
  begin
    FTransColor := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPicture.SetTransMode(const Value: TTransMode);
begin
  if FTransMode <> Value then
  begin
    FTransMode := Value;
    FOwner.Invalidate;
  end;
end;




{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ ---  CHRotation  ---}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHRotation.Create(AOwner: TCHCustomImage);
begin
  inherited Create;
  FOwner := AOwner;
  FAngle := 0;
  FBackcolor := FOwner.Color;
  FxAxis := FOwner.Width div 2;
  FyAxis := FOwner.Height div 2;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRotation.SetAngle(const Value: TAngle);
begin
  if FAngle <> Value then
  begin
    FAngle := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRotation.SetBackcolor(const Value: TColor);
begin
  if FBackcolor <> Value then
  begin
    FBackcolor := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRotation.SetxAxis(const Value: Word);
begin
  if FxAxis <> Value then
  begin
    FxAxis := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRotation.SetyAxis(const Value: Word);
begin
  if FyAxis <> Value then
  begin
    FyAxis := Value;
    FOwner.Invalidate;
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ ---  CHFade  ---}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHFade.Create(AOwner: TCHCustomImage);
begin
  inherited Create;
  FOwner := AOwner;
  FFadeActive := False;
  FFadeInterval := 5;
  FFadeMode := fmOutIn;
  FFadeLoop := 1;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFade.SetFadeActive(const Value: Boolean);
begin
  if FFadeActive <> Value then
  begin
    FFadeActive := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFade.SetFadeInterval(const Value: Word);
begin
  if FFadeInterval <> Value then
  begin
    FFadeInterval := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFade.SetFadeLoop(const Value: Word);
begin
  if FFadeLoop <> Value then
  begin
    FFadeLoop := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFade.SetFadeMode(const Value: TFadeMode);
begin
  if FFadeMode <> Value then
  begin
    FFadeMode := Value;
    FOwner.Invalidate;
  end;
end;




end.
