unit CHPanel;

{ ####################################################################################
  TCHPanel

  Version   :   1.1.0
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TPanel

  History:
  1.0.0 - 21.07.2002    - First Release
  1.0.1 - 20.08.2002    - BUG: Transparents funktionierte nicht
  1.1.0 . 29.08.2002    - NEW: OnMouseEnter, OnMouseLeave
                        - NEW: Layout berarbeitet, Resize, Collapse

  #################################################################################### }

interface

uses
  Windows, Forms, Messages, SysUtils, Classes, Controls, ExtCtrls, Graphics, ShellApi, Buttons, jpeg,
  _CHClassGraphic, _CHClassControls, _CHClassFunctions, _CHFunction, _CHTypes,
  CHButton;

type
  TCHPanel = class;

  TCHPanelLayout = class(TPersistent)
  private
    FOwner : TCHPanel;
    FShowControl: Boolean;

    FControlText: TCaption;
    FControlColor: TColor;
    FControlStyle: TPanelBevel;
    FShowMoveDown: Boolean;
    FShowMoveRight: Boolean;
    FShowMoveUp: Boolean;
    FShowMoveLeft: Boolean;
    FSpace : Integer;
    FAllButtonWidth : Integer;
    FControlY: Word;
    FControlX: Word;
    FFont: TFont;
    FControlTextAlignment: TAlignment;

    procedure SetControlColor(const Value: TColor);
    procedure SetControlText(const Value: TCaption);
    procedure SetShowControl(const Value: Boolean);
    procedure SetShowMoveUp(const Value: Boolean);
    procedure SetShowMoveDown(const Value: Boolean);
    procedure SetShowMoveLeft(const Value: Boolean);
    procedure SetShowMoveRight(const Value: Boolean);
    procedure SetSpace(const Value: integer);
    procedure SetControlX(const Value: Word);
    procedure SetControlY(const Value: Word);
    procedure SetControlStyle(const Value: TPanelBevel);
    procedure SetFont(const Value: TFont);
    procedure SetControlTextAlignment(const Value: TAlignment);
    procedure SetMoveButton;
    function GetFont: TFont;
    function GetControlText: TCaption;
  protected

  public
    constructor Create(AOwner : TCHPanel); virtual;
    destructor Destroy; override;
  published
    property ControlColor : TColor read FControlColor Write SetControlColor;
    property ControlFont : TFont read GetFont write SetFont;
    property ControlText : TCaption read GetControlText Write SetControlText;
    property ControlTextAlignment : TAlignment read FControlTextAlignment Write SetControlTextAlignment;
    property ControlStyle : TPanelBevel read FControlStyle Write SetControlStyle;
    property ControlX : Word read FControlX Write SetControlX;
    property ControlY : Word read FControlY Write SetControlY;
    property ShowControl : Boolean read FShowControl Write SetShowControl;
    property ShowMoveUp : Boolean read FShowMoveUp Write SetShowMoveUp;
    property ShowMoveDown : Boolean read FShowMoveDown Write SetShowMoveDown;
    property ShowMoveLeft : Boolean read FShowMoveLeft Write SetShowMoveLeft;
    property ShowMoveRight : Boolean read FShowMoveRight Write SetShowMoveRight;
    property Space : Integer read FSpace Write SetSpace;
  end;

  //***********************************************************************************************************

  TCHPanelMover = class(TPersistent)
  private
    FOwner : TCHPanel;
    FCanCollapse: Boolean;
    FCanResize: Boolean;
    FAutosizer : Boolean;
    FMoveSpeed: TMoveSpeed;
    FMinHeight : Integer;
    FMaxHeight : Integer;
    FMinWidth : Integer;
    FMaxWidth : Integer;
    FMoveStep : Integer;
    FMoveMode : TMoveMode;
    FMoveTimer : TTimer;
    FMoveDirection : TCollapseDirection;

    procedure SetCanCollapse(const Value: Boolean);
    procedure SetCanResize(const Value: Boolean);
    procedure SetMoveSpeed(const Value: TMoveSpeed);
    procedure SetAutosizer(const Value: Boolean);
    procedure SetMoveMode(const Value: TMoveMode);
    procedure DoAutosizer;
    procedure DoCollapse(Sender: TObject);
  protected

  public
    constructor Create(AOwner : TCHPanel); virtual;
    destructor Destroy; override;
  published
    property Autosizer : Boolean read FAutosizer Write SetAutosizer;
    property CanCollapse : Boolean read FCanCollapse Write SetCanCollapse;
    property CanResize : Boolean read FCanResize Write SetCanResize;
    property CollapseSpeed : TMoveSpeed read FMoveSpeed Write SetMoveSpeed;
    property CollapseStep : Integer read FMoveStep Write FMoveStep;
    property MinHeight : Integer read FMinHeight Write FMinHeight;
    property MaxHeight : Integer read FMaxHeight Write FMaxHeight;
    property MinWidth : Integer read FMinWidth Write FMinWidth;
    property MaxWidth : Integer read FMaxWidth Write FMaxWidth;
    property MoveMode : TMoveMode read FMoveMode Write SetMoveMode;
  end;

  //***********************************************************************************************************

  TCHPanel = class(TPanel)
  private
    FOnMouseDown : TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;

    FGradient : TCHGradient;
    FPanelMover : TCHPanelMover;
    FPanelLayout: TCHPanelLayout;
    FFill : TCHFill;
    FBitmap : TCHBitmap;

    FControlPanel : TPanel;
    FClientRect : TRect;
    FPanelRect : TRect;

    FMoverUp : TCHButton;
    FMoverDown : TCHButton;
    FMoverLeft : TCHButton;
    FMoverRight : TCHButton;
    FBuffered: Boolean;
    FBackgroundBmp : TBitmap;
    FError : TError;
    FResize: Boolean;
    FResizeStartPos: TPoint;

    procedure SetGradient(const Value: TCHGradient);
    procedure SetBuffered(const Value: Boolean);
    procedure MakeForeground;
    procedure CMMouseEnter(var Message : TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message : TMessage); message CM_MOUSELEAVE;
    procedure MoverUpClick(Sender: TObject);
    procedure MoverDownClick(Sender: TObject);
    procedure MoverLeftClick(Sender: TObject);
    procedure MoverRightClick(Sender: TObject);
    procedure ControlClick(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure GetRectSize;
    procedure UpdateChanges(Sender: TObject);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure Resize; override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    procedure Collapse(Direction : TCollapseDirection);
  published
    property OnMouseDown: TNotifyEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseEnter : TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave : TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property Bitmap : TCHBitmap read FBitmap Write FBitmap;
    property Buffered : Boolean read FBuffered write SetBuffered;
    property Fill : TCHFill read FFill write FFill;
    property Gradient : TCHGradient read FGradient write SetGradient;
    property Layout : TCHPanelLayout read FPanelLayout Write FPanelLayout;
    property Mover : TCHPanelMover read FPanelMover Write FPanelMover;
  end;

procedure Register;

implementation

{$R Panel.res}

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHPanel]);
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FGradient := TCHGradient.Create;
  FGradient.OnChange := UpdateChanges;
  FFill := TCHFill.Create;
  FFill.OnChange := UpdateChanges;
  FBitmap := TCHBitmap.Create;
  FBitmap.OnChange := UpdateChanges;

  Width := 200;
  Height := 130;
  ControlStyle := ControlStyle + [csAcceptsControls, csOpaque];

  FBackgroundBmp := TBitmap.Create;
  FPanelLayout := TCHPanelLayout.Create(Self);
  FPanelMover := TCHPanelMover.Create(Self);
  FPanelMover.MoveMode := mmNone;

  // Control Area
  FControlPanel := TPanel.Create(Self);
  with FControlPanel do
  begin
    Parent := Self;
    ParentFont := False;
    Alignment := taRightJustify;
    DoubleBuffered := True;
    Caption := '';
    Color := FPanelLayout.ControlColor;
    BevelOuter := FPanelLayout.FControlStyle;

    Height := 26;
    Width := Self.Width - (FPanelLayout.ControlX * 2);
    Left := FPanelLayout.ControlX;
    Top := -Height;
    Visible := False;

    OnMouseDown := ControlClick;
  end;

  // Collapse Up
  FMoverUp := TCHButton.Create(FControlPanel);
  with FMoverUp do
  begin
    Parent := FControlPanel;
    Width := 20;
    Height := 20;

    Border.Width := 1;
    Caption := '';
    Color := FPanelLayout.ControlColor;
    Focus.Active := False;

    with Glyph do
    begin
      Glyph.LoadFromResourceName(hInstance, 'PN_MOVE_UP');
      Alignment := gaCenter;
      AlignMode := gmControl;
      Space := 0;
    end;
    OnClick := MoverUpClick;
  end;

  // Collapse Down
  FMoverDown := TCHButton.Create(FControlPanel);
  with FMoverDown do
  begin
    Parent := FControlPanel;
    Width := 20;
    Height := 20;

    Border.Width := 1;
    Caption := '';
    Color := FPanelLayout.ControlColor;
    Focus.Active := False;

    with Glyph do
    begin
      Glyph.LoadFromResourceName(hInstance, 'PN_MOVE_DOWN');
      Alignment := gaCenter;
      AlignMode := gmControl;
      Space := 0;
    end;
    OnClick := MoverDownClick;
  end;

  // Collapse Left
  FMoverLeft := TCHButton.Create(FControlPanel);
  with FMoverLeft do
  begin
    Parent := FControlPanel;
    Width := 20;
    Height := 20;

    Border.Width := 1;
    Caption := '';
    Color := FPanelLayout.ControlColor;
    Focus.Active := False;

    with Glyph do
    begin
      Glyph.LoadFromResourceName(hInstance, 'PN_MOVE_LEFT');
      Alignment := gaCenter;
      AlignMode := gmControl;
      Space := 0;
    end;
    OnClick := MoverLeftClick;
  end;

  // Collapse Right
  FMoverRight := TCHButton.Create(FControlPanel);
  with FMoverRight do
  begin
    Parent := FControlPanel;
    Width := 20;
    Height := 20;

    Border.Width := 1;
    Caption := '';
    Color := FPanelLayout.ControlColor;
    Focus.Active := False;

    with Glyph do
    begin
      Glyph.LoadFromResourceName(hInstance, 'PN_MOVE_RIGHT');
      Alignment := gaCenter;
      AlignMode := gmControl;
      Space := 0;
    end;
    OnClick := MoverRightClick;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHPanel.Destroy;
begin
  FBackgroundBmp.Free;
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHPanelMover.Create(AOwner: TCHPanel);
begin
  inherited Create;
  FOwner := AOwner;

  FAutosizer := True;
  FMoveSpeed := 20;
  FMoveStep := 10;
  FMoveTimer := TTimer.Create(AOwner);
  FMoveTimer.Enabled := False;
  FMoveTimer.Interval := FMoveSpeed;
  FMoveTimer.OnTimer := DoCollapse;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.UpdateChanges(Sender: TObject);
begin
  if csLoading in ComponentState then
    Exit;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.CMMouseEnter(var Message: TMessage);
begin
  if Assigned(FOnMouseEnter) then
    FOnMouseEnter(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.Resize;
begin
  inherited;
  FControlPanel.Width := Self.Width - (FPanelLayout.ControlX * 2);
  FPanelMover.DoAutosizer;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.Paint;
var
  nX_BackgroundBmp, nY_BackgroundBmp, nX_Resize, nY_Resize : Integer;
  TmpFont : TFont;
begin
  inherited Paint;

  FError := errNone;
  Canvas.Font.Assign(Self.Font);

  // Control Panel
  FControlPanel.Width := Self.Width - (FPanelLayout.ControlX * 2);
  FControlPanel.Font := FPanelLayout.FFont;

  // Rect berechnen
  GetRectSize;


  // GRADIENT
  if Fill.Style = fsGradient then
  begin
    if Length(FGradient.FGradientColorArray) > 1 then
    begin
      DrawGradient_C(Self, Self.Canvas, FClientRect, FGradient.FGradientColorArray,
        FGradient.Style, FGradient.Rotation);
    end
    // nur eine Farbe = True --> einfarbig
    else if (Length(FGradient.FGradientColorArray) = 1) then
    begin
      Canvas.Brush.Color := FGradient.FGradientColorArray[0];
      Canvas.Brush.Style := bsSolid;
      Canvas.FillRect(FClientRect);
    end
    // alle Farben = False --> ohne Fllung
    else
    begin
      Canvas.Brush.Color := Color;
      Canvas.Brush.Style := bsSolid;
      Canvas.FillRect(FClientRect);
    end;
  end;

  // NORMAL
  if (FFill.Style = fsNormal) then
  begin
    //
  end;

  // BITMAP
  if (FFill.Style = fsBitmap) then
  begin
    // Bitmap prfen
    if not FBitmap.Bitmap.Empty then
    begin
      // Bitmap erstellen
      MakeForeground;

      // Bitmap zeichnen
      with Canvas do
      begin

        if FBitmap.Mode = bmNormal then
        begin
          FBackgroundBmp.Canvas.Draw(0, 0, FBitmap.Bitmap);
          nX_BackgroundBmp := FPanelRect.Left;
          nY_BackgroundBmp := FPanelRect.Top;
          Draw(nX_BackgroundBmp, nY_BackgroundBmp, FBackgroundBmp);
        end
        else if FBitmap.Mode = bmStretch then
        begin
          FBackgroundBmp.Canvas.Draw(0, 0, FBitmap.Bitmap);
          StretchDraw(FPanelRect, FBackgroundBmp);
        end
        else if FBitmap.Mode = bmTile then
        begin
          FBackgroundBmp.Canvas.Draw(0, 0, FBitmap.Bitmap);
          Brush.Bitmap := FBackgroundBmp;
          FillRect(FPanelRect);
        end
        else if FBitmap.Mode = bmCenter then
        begin
          // X
          if (FPanelRect.Right - FPanelRect.Left) > FBitmap.Bitmap.Width then
            nX_BackgroundBmp := (FPanelRect.Right div 2) - (FBitmap.Bitmap.Width div 2)
          else
            nX_BackgroundBmp := -(FBitmap.Bitmap.Width - FPanelRect.Right) div 2;

          // Y
          if (FPanelRect.Bottom - FPanelRect.Top) > FBitmap.Bitmap.Height then
            nY_BackgroundBmp := (FPanelRect.Bottom div 2) - (FBitmap.Bitmap.Height div 2)
          else
            nY_BackgroundBmp := -(FBitmap.Bitmap.Height - FPanelRect.Bottom) div 2;

          FBackgroundBmp.Canvas.Draw(nX_BackgroundBmp, nY_BackgroundBmp, FBitmap.Bitmap);
          Draw(0, 0, FBackgroundBmp);
        end;
      end;
    end
    else
      FError := errNoBitmap;
  end;

  // Transparent
  if (FFill.Transparent) and (FFill.Style = fsNormal) then
  begin
    DrawTransparentBmp(Self, Canvas);
    Canvas.CopyRect(FClientRect, canvas, FClientRect);
    Brush.Style := bsClear;
  end;

  // Resize
  if FPanelMover.CanResize then
  begin
    TmpFont := Canvas.Font;
    Canvas.Font.Name := 'Marlett';
    Canvas.Font.Size := 12;
    nX_Resize := ClientWidth - Canvas.TextWidth('o');
    nY_Resize := ClientHeight - Canvas.TextHeight('o');
    Canvas.Brush.Style := bsClear;
    Canvas.Pen.Color := clBlack;
    Canvas.TextOut(nX_Resize, nY_Resize, 'o');
    Canvas.Font := TmpFont;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.GetRectSize;
begin
  // Mae Rechteck zuweisen
  FClientRect := GetClientRect;

  FPanelRect.Left := FClientRect.Left + BevelWidth + BorderWidth;
  FPanelRect.Top := FClientRect.Top + BevelWidth + BorderWidth;
  FPanelRect.Right := FClientRect.Right - BevelWidth + BorderWidth;
  FPanelRect.Bottom := FClientRect.Bottom - BevelWidth + BorderWidth;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.Collapse(Direction: TCollapseDirection);
begin
  FPanelMover.FMoveDirection := Direction;
  FPanelMover.FMoveTimer.Enabled := True;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.SetBuffered(const Value: Boolean);
begin
  if FBuffered <> Value then
  begin
    FBuffered := Value;
    if (FBuffered = True) then
      Self.DoubleBuffered := True
    else
      Self.DoubleBuffered := False;

    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MakeForeground;
var
  BRect : TRect;
begin
  BRect := Bounds(0, 0, Self.Width, Self.Height);
  FBackgroundBmp.Canvas.Brush.Color := clWhite;
  FBackgroundBmp.Canvas.Brush.Style := bsSolid;
  FBackgroundBmp.Canvas.FillRect(BRect);

  if (FBitmap.Mode = bmNormal) or (FBitmap.Mode = bmCenter) then
  begin
    FBackgroundBmp.Width := FPanelRect.Right - FPanelRect.Left;
    FBackgroundBmp.Height := FPanelRect.Bottom - FPanelRect.Top;
  end
  else if (FBitmap.Mode = bmStretch) or (FBitmap.Mode = bmTile) then
  begin
    FBackgroundBmp.Width := FBitmap.Bitmap.Width;
    FBackgroundBmp.Height := FBitmap.Bitmap.Height;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.SetGradient(const Value: TCHGradient);
begin
  FGradient.Assign(Value);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MoverDownClick(Sender: TObject);
begin
  FPanelMover.FMoveDirection := cdDown;
  FPanelMover.FMoveTimer.Enabled := True;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MoverLeftClick(Sender: TObject);
begin
  FPanelMover.FMoveDirection := cdLeft;
  FPanelMover.FMoveTimer.Enabled := True;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MoverRightClick(Sender: TObject);
begin
  FPanelMover.FMoveDirection := cdRight;
  FPanelMover.FMoveTimer.Enabled := True;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MoverUpClick(Sender: TObject);
begin
  FPanelMover.FMoveDirection := cdUp;
  FPanelMover.FMoveTimer.Enabled := True;
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.ControlClick(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
const
  SC_DRAGMOVE = $F012;
begin
  if (FPanelMover.MoveMode = mmControl) or (FPanelMover.MoveMode = mmAll) then
  begin
    ReleaseCapture;
    Self.Perform(WM_SYSCOMMAND, SC_DRAGMOVE, 0);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
const
  SC_DRAGMOVE = $F012;
begin
  if Assigned(FOnMouseDown) then
    FOnMouseDown(Self);

  // Resize
  if FPanelMover.CanResize then
  begin
    if (Button = mbLeft) and ((Width - x) < 10) and ((Height - y) < 10) then
    begin
      FResize := True;
      MouseCapture := True;
      FResizeStartPos := Point(x, y);
      Screen.Cursor := crSizeNWSE;
    end;
  end
  // Move
  else if (FPanelMover.MoveMode = mmAll) then
  begin
    ReleaseCapture;
    Self.Perform(WM_SYSCOMMAND, SC_DRAGMOVE, 0);
  end
  else
    inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewRect : TRect;
begin
  if FResize then
  begin
    NewRect := BoundsRect;
    SetBounds(NewRect.left, NewRect.Top, NewRect.Right - NewRect.Left +
      X - FResizeStartPos.X, NewRect.Bottom - NewRect.Top + Y - FResizeStartPos.Y);
    FResizeStartPos := Point(x, y);

    // MinWidth
    if (Width < FPanelMover.FMinWidth) then
    begin
      NewRect.Right := FPanelMover.FMinWidth;
      SetBounds(NewRect.left, NewRect.Top, NewRect.Right, NewRect.Bottom - NewRect.Top + Y - FResizeStartPos.Y);
    end;
    // MaxWidth
    if (Width > FPanelMover.FMaxWidth) then
    begin
      NewRect.Right := FPanelMover.FMaxWidth;
      SetBounds(NewRect.left, NewRect.Top, NewRect.Right, NewRect.Bottom - NewRect.Top + Y - FResizeStartPos.Y);
    end;

    // MinHeight
    if  (Height < FPanelMover.FMinHeight) then
    begin
      NewRect.Bottom := FPanelMover.FMinHeight;
      SetBounds(NewRect.left, NewRect.Top, NewRect.Right - NewRect.Left + X - FResizeStartPos.X, NewRect.Bottom);
    end;
    // MaxHeight
    if  (Height > FPanelMover.FMaxHeight) then
    begin
      NewRect.Bottom := FPanelMover.FMaxHeight;
      SetBounds(NewRect.left, NewRect.Top, NewRect.Right - NewRect.Left + X - FResizeStartPos.X, NewRect.Bottom);
    end;
  end
  else
  begin
    inherited;
    if FPanelMover.CanResize and ((Width - x) < 10) and ((Height - y) < 10) then
      Cursor := crSizeNWSE
    else
      Cursor := crDefault;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanel.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  if FResize then
  begin
    FResize := False;
    MouseCapture := false;
    Screen.Cursor := crDefault;
  end
  else
    inherited;

end;



{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ ---  CHPanelLayout  ---}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }


constructor TCHPanelLayout.Create(AOwner: TCHPanel);
begin
  inherited Create;
  FOwner := AOwner;

  FFont := TFont.Create;
  FFont.Name := 'Arial';
  FFont.Size := 10;
  FFont.Style := [fsBold];
  FFont.Color := clWhite;

  FShowControl := False;
  FShowMoveDown := True;
  FShowMoveUp := True;
  FShowMoveRight := True;
  FShowMoveLeft := True;
  FControlColor := clActiveCaption;
  FControlStyle := bvLowered;
  FControlTextAlignment := taRightJustify;
  FControlX := 3;
  FControlY := 3;
  FSpace := 0;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHPanelLayout.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetControlColor(const Value: TColor);
begin
  if FControlColor <> Value then
  begin
    FControlColor := Value;
    FOwner.FControlPanel.Color := Value;
    FOwner.FMoverUp.Color := Value;
    FOwner.FMoverDown.Color := Value;
    FOwner.FMoverLeft.Color := Value;
    FOwner.FMoverRight.Color := Value;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetControlText(const Value: TCaption);
begin
  if FControlText <> Value then
  begin
    FControlText := Value;
    FOwner.FControlPanel.Caption := Value;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetMoveButton;
var
  nNextLeft, nTop, nButtonWidth : Integer;
begin
  nNextLeft := 3;
  nTop := 3;
  nButtonWidth := FOwner.FMoverUp.Width;

  // Designtime
  if csDesigning in FOwner.ComponentState then
  begin
    FOwner.FMoverUp.Top := -FOwner.FMoverUp.Height;
    FOwner.FMoverDown.Top := -FOwner.FMoverDown.Height;
    FOwner.FMoverLeft.Top := -FOwner.FMoverLeft.Height;
    FOwner.FMoverRight.Top := -FOwner.FMoverRight.Height ;
  end
  // Runtime
  else
  begin
    FOwner.FMoverUp.Visible := False;
    FOwner.FMoverDown.Visible := False;
    FOwner.FMoverLeft.Visible := False;
    FOwner.FMoverRight.Visible := False;
  end;

  // Up
  if FShowMoveUp then
  begin
    FOwner.FMoverUp.Visible := True;
    FOwner.FMoverUp.Left := nNextLeft;
    FOwner.FMoverUp.Top := nTop;
    nNextLeft := nNextLeft + nButtonWidth + FSpace;
  end;
  // Down
  if FShowMoveDown then
  begin
    FOwner.FMoverDown.Visible := True;
    FOwner.FMoverDown.Left := nNextLeft;
    FOwner.FMoverDown.Top := nTop;
    nNextLeft := nNextLeft + nButtonWidth + FSpace;
  end;
  // Left
  if FShowMoveLeft then
  begin
    FOwner.FMoverLeft.Visible := True;
    FOwner.FMoverLeft.Left := nNextLeft;
    FOwner.FMoverLeft.Top := nTop;
    nNextLeft := nNextLeft + nButtonWidth + FSpace;
  end;
  // Right
  if FShowMoveRight then
  begin
    FOwner.FMoverRight.Visible := True;
    FOwner.FMoverRight.Left := nNextLeft;
    FOwner.FMoverRight.Top := nTop;
    nNextLeft := nNextLeft + nButtonWidth + FSpace;
  end;

  FAllButtonWidth := nNextLeft;
  FOwner.Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetSpace(const Value: integer);
begin
  if FSpace <> Value then
  begin
    if Value < 0 then
      FSpace := 0
    else
      FSpace := Value;
  end;
  SetMoveButton;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetShowControl(const Value: Boolean);
begin
  if FShowControl <> Value then
  begin
    FShowControl := Value;
    if FShowControl then
    begin
      FOwner.FControlPanel.Top := FControlY;
      FOwner.FControlPanel.Left := FControlX;
      FOwner.FControlPanel.Visible := True;
      SetMoveButton;
    end
    else
    begin
      FOwner.FControlPanel.Top := -FOwner.FControlPanel.Height;
      FOwner.FControlPanel.Visible := False;
    end;

  end;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetShowMoveDown(const Value: Boolean);
begin
  if FShowMoveDown <> Value then
  begin
    FShowMoveDown := Value;
  end;
  SetMoveButton;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetShowMoveLeft(const Value: Boolean);
begin
  if FShowMoveLeft <> Value then
  begin
    FShowMoveLeft := Value;
  end;
  SetMoveButton;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetShowMoveRight(const Value: Boolean);
begin
  if FShowMoveRight <> Value then
  begin
    FShowMoveRight := Value;
  end;
  SetMoveButton;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetShowMoveUp(const Value: Boolean);
begin
  if FShowMoveUp <> Value then
  begin
    FShowMoveUp := Value;
  end;
  SetMoveButton;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetControlX(const Value: Word);
begin
  if FControlX <> Value then
  begin
    FControlX := Value;
    FOwner.FControlPanel.Left := FControlX;
  end;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetControlY(const Value: Word);
begin
  if FControlY <> Value then
  begin
    FControlY := Value;
    FOwner.FControlPanel.Top := FControlY;
  end;
  FOwner.FPanelMover.DoAutosizer;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetControlStyle(const Value: TPanelBevel);
begin
  if FControlStyle <> Value then
  begin
    FControlStyle := Value;
    FOwner.FControlPanel.BevelOuter := FControlStyle;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetFont(const Value: TFont);
begin
  FOwner.FControlPanel.Font := Value;
  FOwner.Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelLayout.SetControlTextAlignment(const Value: TAlignment);
begin
  if FControlTextAlignment <> Value then
  begin
    FControlTextAlignment := Value;
    FOwner.FControlPanel.Alignment := FControlTextAlignment;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHPanelLayout.GetControlText: TCaption;
begin
  Result := FOwner.FControlPanel.Caption;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
function TCHPanelLayout.GetFont: TFont;
begin
  Result := FOwner.FControlPanel.Font;
end;





{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ ---  CHPanelMover  ---}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

destructor TCHPanelMover.Destroy;
begin
  FMoveTimer.Free;
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.DoAutosizer;
begin
  if FAutosizer then
  begin
    FMaxHeight := FOwner.Height;
    FMaxWidth := FOwner.Width;
    if FOwner.FPanelLayout.ShowControl then
    begin
      FMinHeight := FOwner.FControlPanel.Top + FOwner.FControlPanel.Height +
        FOwner.FPanelLayout.FControlY;
      FMinWidth := FOwner.FControlPanel.Left + FOwner.FPanelLayout.FAllButtonWidth +
        FOwner.FPanelLayout.FControlX + 3;
    end
    else
    begin
      FMinHeight := 0;
      FMinWidth := 0;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.DoCollapse;
begin
  if FCanCollapse then
  begin
    Autosizer := False;
    FMoveTimer.Interval := FMoveSpeed;

    // Down
    if FMoveDirection = cdDown then
    begin
      if (FOwner.Height <= FMaxHeight) and not (FMoveSpeed = 0) then
      begin
        FOwner.Height := FOwner.Height + FMoveStep;
        FOwner.Repaint;
      end
      else
      begin
        FOwner.Height := FMaxHeight;
        FOwner.Repaint;
        FMoveTimer.Enabled := False;
      end;
    end
    // Up
    else if FMoveDirection = cdUp then
    begin
      if (FOwner.Height >= FMinHeight) and not (FMoveSpeed = 0) then
      begin
        FOwner.Height := FOwner.Height - FMoveStep;
        FOwner.Repaint;
      end
      else
      begin
        FOwner.Height := FMinHeight + 1;
        FOwner.Repaint;
        FMoveTimer.Enabled := False;
      end;
    end
    // Left
    else if FMoveDirection = cdLeft then
    begin
      if FMinWidth < FOwner.FPanelLayout.FAllButtonWidth + FOwner.FPanelLayout.FControlX then
        FMinWidth := FOwner.FPanelLayout.FAllButtonWidth + FOwner.FPanelLayout.FControlX;

      if (FOwner.Width >= FMinWidth) and not (FMoveSpeed = 0) then
      begin
        FOwner.Width := FOwner.Width - FMoveStep;
        FOwner.Repaint;
      end
      else
      begin
        FOwner.Width := FMinWidth;
        FOwner.Repaint;
        FMoveTimer.Enabled := False;
      end;
    end
    // Right
    else if FMoveDirection = cdRight then
    begin
      if (FOwner.Width <= FMaxWidth) and not (FMoveSpeed = 0) then
      begin
        FOwner.Width := FOwner.Width + FMoveStep;
        FOwner.Repaint;
      end
      else
      begin
        FOwner.Width := FMaxWidth;
        FOwner.Repaint;
        FMoveTimer.Enabled := False;
      end;
    end
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.SetAutosizer(const Value: Boolean);
begin
  if FAutosizer <> Value then
  begin
    FAutosizer := Value;
    DoAutosizer;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.SetCanCollapse(const Value: Boolean);
begin
  if FCanCollapse <> Value then
  begin
    FCanCollapse := Value;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.SetCanResize(const Value: Boolean);
begin
  if FCanResize <> Value then
  begin
    FCanResize := Value;
    FOwner.Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.SetMoveMode(const Value: TMoveMode);
begin
  if FMoveMode <> Value then
  begin
    FMoveMode := Value;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHPanelMover.SetMoveSpeed(const Value: TMoveSpeed);
begin
  if FMoveSpeed <> Value then
  begin
    FMoveSpeed := Value;
  end;
end;




end.
