unit CHRadioButton;

{ ####################################################################################
  TCHRadioButton

  Version   :   1.0.1
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TCustomControl

  History:
  1.0.0 - 21.07.2002    - First Release
  1.0.1 - 31.08.2002    - BUG: Speicherleck wenn "LoadBitmap", wieder freigeben

  #################################################################################### }

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi, Forms,

  _CHClassGraphic, _CHClassControls, _CHClassFunctions, _CHFunction, _CHTypes;

type
  TCHCustomRadioButton = class(TCustomControl)
  private
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnClick : TNotifyEvent;

    FCaptionLayout : TCHCaptionLayout;
    FBox : TCHBoxRB;
    FGlyph : TCHGlyph;
    FBorder : TCHBorder;
    FFill : TCHFillB;
    FFocus : TCHFocus;

    FClientRect : TRect;
    FWorkRect : TRect;
    FCaptionRect : TRect;
    FColor: TColor;
    FGroupIndex : Integer;
    FDot : Boolean;
    FCross : Boolean;
    FSolid : Boolean;
    FHLine : Boolean;
    FNormal : Boolean;
    FFlat : Boolean;
    FChecked: Boolean;
    FActive : Boolean;
    FDefault : Boolean;
    FClicksDisabled : Boolean;
    FTmpColor : TBoxFillColor;
    FCIdx : Integer;
    FError : TError;

    procedure SetTextStyle(const Value: TTextStyle);
    procedure SetColor(const Value: TColor);
    procedure SetChecked(const Value: Boolean);

    procedure GetRectSize;
    procedure MakeBorder(ButtonState : Boolean);
    procedure GetBoxBmp(Style : TBoxStyle; Fill : TRadioButtonFill; Color : TBoxFillColor);
    procedure DoGroup;
    procedure DrawFocus;

    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message : TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Msg : TMessage); message CM_TEXTCHANGED;
    procedure CMDialogChar(var Message : TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMEnabledChanged(var Message :TMessage); message CM_ENABLEDCHANGED;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_EraseBkgnd;
    procedure CMFocusChanged(var Message: TCMFocusChanged); message CM_FOCUSCHANGED;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure UpdateChanges(Sender: TObject);
  protected
    procedure Paint; override;
    procedure Click; override;
    procedure WndProc(var Message: TMessage); override;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
  published
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;

    property Box : TCHBoxRB read FBox Write FBox;
    property Border : TCHBorder read FBorder write FBorder;
    property CaptionLayout : TCHCaptionLayout read FCaptionLayout write FCaptionLayout;
    property Checked : Boolean read FChecked Write SetChecked;
    property Color: TColor read FColor write SetColor;
    property Fill : TCHFillB read FFill write FFill;
    property Focus : TCHFocus read FFocus write FFocus;
    property Glyph : TCHGlyph read FGlyph Write FGlyph;
    property GroupIndex: Integer read FGroupIndex write FGroupIndex;
  end;

  TCHRadioButton = class(TCHCustomRadioButton)
  published
    property Action;
    property Anchors;
    property BiDiMode;
    property Caption;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentBiDiMode;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabStop default True;
    property TabOrder;
    property Visible;

    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDockDrop;
    property OnDockOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

procedure Register;

implementation

{$R Radiobutton.res}

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHRadioButton]);
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHCustomRadioButton.Create(aOwner: TComponent);
begin
  inherited Create(AOwner);

  FBox := TCHBoxRB.Create;
  FBox.OnChange := UpdateChanges;
  FBorder := TCHBorder.Create;
  FBorder.OnChange := UpdateChanges;
  FGlyph := TCHGlyph.Create;
  FGlyph.OnChange := UpdateChanges;
  FCaptionLayout := TCHCaptionLayout.Create;
  FCaptionLayout.OnChange := UpdateChanges;
  FFill := TCHFillB.Create;
  FFill.OnChange := UpdateChanges;
  FFocus := TCHFocus.Create;
  FFocus.OnChange := UpdateChanges;

  Width := 120;
  Height := 20;

  FCaptionLayout.Alignment := tgAuto;

  FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_BLANK');
  FGlyph.AlignMode := gmControl;

  FBorder.HighlightColor := clWhite;
  FBorder.ShadowColor := clGray;
  FBorder.Width := 0;
  FBorder.SingleWidth := 1;

  FCIdx := 5;
  FColor := clBtnFace;
  FChecked := False;
  FTmpColor := bcBlack;
  FGroupIndex := 0;
  TabStop := True;

  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHCustomRadioButton.Destroy;
begin
  DeleteObject(FGlyph.Glyph.Handle);
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.Click;
begin
  inherited;
  FChecked := True;

  if Assigned(FOnClick) then
    FOnClick(Self);

  DoGroup;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.GetBoxBmp(Style : TBoxStyle; Fill : TRadioButtonFill; Color : TBoxFillColor);
begin
  FNormal := False;
  FFlat := False;

  FDot := False;
  FCross := False;
  FSolid := False;
  FHLine := False;

  if FChecked = False then
    FTmpColor := bcNone
  else
    FTmpColor := Color;

  // NORMAL
  if Style = bxNormal then
  begin
    FNormal := True;
    if FTmpColor = bcNone then
      FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_BlANK')
    else
    begin
      if Fill = rfDot then
      begin
        FDot := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_DOT_GRAY');
      end
      else if Fill = rfCross then
      begin
        FCross := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_CROSS_GRAY');
      end
      else if Fill = rfSolid then
      begin
        FSolid := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_FILL_GRAY');
      end
      else if Fill = rfHLine then
      begin
        FHLine := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_N_HLINE_GRAY');
      end;
    end;
  end
  // FLAT
  else if Style = bxFlat then
  begin
    FFlat := True;
    if FTmpColor = bcNone then
      FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_BlANK')
    else
    begin
      if Fill = rfDot then
      begin
        FDot := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_DOT_GRAY');
      end
      else if Fill = rfCross then
      begin
        FCross := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_CROSS_GRAY');
      end
      else if Fill = rfSolid then
      begin
        FSolid := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_FILL_GRAY');
      end
      else if Fill = rfHLine then
      begin
        FHLine := True;
        if FTmpColor = bcBlack then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_BLACK')
        else if FTmpColor = bcBlue then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_BLUE')
        else if FTmpColor = bcGreen then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_GREEN')
        else if FTmpColor = bcLime then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_LIME')
        else if FTmpColor = bcRed then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_RED')
        else if FTmpColor = bcYellow then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_YELLOW')
        else if FTmpColor = bcGray then
          FGlyph.Glyph.Handle := LoadBitmap(hInstance, 'RB_F_HLINE_GRAY');
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.MakeBorder(ButtonState: Boolean);
var
  nRow, nSingle, I : Integer;
begin
  nSingle := 0;

  with Canvas do
  begin
    Pen.Width := 1;
    Pen.Style := psSolid;
    Brush.Style := bsSolid;

    // Single-Border
    if (FBorder.Style = bsExtended) or (FBorder.Style = bsFlat) then
    begin
      Pen.Color := FBorder.Color;
      for I := 0 to FBorder.SingleWidth - 1 do
      begin
        MoveTo(I, I);
        LineTo(Width - 1 - I, I);
        LineTo(Width - 1 - I, Height - 1 - I);
        LineTo(I, Height - 1 - I);
        LineTo(I, I);
      end;
      nSingle := FBorder.SingleWidth;
    end;

    // Button-Border
    if (FBorder.Style = bsExtended) or (FBorder.Style = bsNormal) then
    begin
      // oben + links
      if (ButtonState = True) then
        Pen.Color := FBorder.ShadowColor
      else
        Pen.Color := FBorder.HighlightColor;

      for nRow := 0 to FBorder.Width - 1 do
      begin
        MoveTo(Width - nRow - 1 - nSingle, nRow + nSingle);
        LineTo(nRow + nSingle, nRow + nSingle);
        LineTo(nRow + nSingle, Height - nRow - 1 - nSingle);
      end;

      // rechts + unten
      if (ButtonState = True) then
        Pen.Color := FBorder.HighlightColor
      else
        Pen.Color := FBorder.ShadowColor;

      for nRow := 0 to FBorder.Width - 1 do
      begin
        MoveTo(Width - nRow - 1 - nSingle, nRow + nSingle);
        LineTo(Width - nRow - 1 - nSingle, Height - nRow - 1 - nSingle);
        LineTo(nRow + nSingle - 1, Height - nRow - 1 - nSingle);
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.Paint;
begin
  inherited Paint;
  FError := errNone;

  Canvas.Font.Assign(Self.Font);

  // Radio Bitmap
  GetBoxBmp(FBox.Style, FBox.Fill, FBox.Color);

  // TextStyle
  SetTextStyle(FCaptionLayout.TextStyle);

  // Rect berechnen
  GetRectSize;

  // Caption
  DrawCaption(Self, Self.Canvas, FCaptionRect, FWorkRect, FClientRect, FCaptionLayout.Angle, FCaptionLayout.HighlightDepth,
      FCaptionLayout.ShadowDepth, FCaptionLayout.HighlightDirection, FCaptionLayout.ShadowDirection,
      FGlyph.TransparentMode, FCaptionLayout.HighlightColor, FCaptionLayout.ShadowColor,
      Self.Color, Self.Font.Color, FGlyph.TransparentColor, Self.Caption, FCaptionLayout.TextStyle,
      FCaptionLayout.Alignment, fsNormal, FGlyph.AlignMode, FGlyph.Alignment, FGlyph.Glyph,
      FGlyph.PosX, FGlyph.PosY, FGlyph.Space, FCaptionLayout.PosX, FCaptionLayout.PosY, FCIdx,
      FFill.Transparent, False, False, FCaptionLayout.Antialiasing, FError);

  // Border zeichnen
  MakeBorder(False);

  // Focus
  if FActive or FFocus.Show then
    DrawFocus;

end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.GetRectSize;
var
  nSingle, nBorderWidth : Integer;
begin
  nSingle := 0;
  nBorderWidth := 0;

  if (FBorder.Style = bsExtended) or (FBorder.Style = bsFlat) then
    nSingle := FBorder.SingleWidth;

  if (FBorder.Style = bsExtended) or (FBorder.Style = bsNormal) then
    nBorderWidth := FBorder.Width;

  // ClientRect
  FClientRect := GetClientRect;

  // WorkRect
  FWorkRect.Left := FClientRect.Left + (nBorderWidth + nSingle);
  FWorkRect.Top := FClientRect.Top + (nBorderWidth + nSingle);
  FWorkRect.Right := FClientRect.Right - (nBorderWidth + nSingle);
  FWorkRect.Bottom := FClientRect.Bottom - (nBorderWidth + nSingle);

  // TextRect
  FCaptionRect.Left := 0;
  FCaptionRect.Top := 0;
  FCaptionRect.Right := 0;
  FCaptionRect.Bottom := 0;
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.DoGroup;
var
  nCtl : Integer;
  SelfRadio : TCHRadioButton;
begin
    if FChecked then
    begin
      if Parent <> nil then
      begin
        for nCtl := 0 to Parent.ControlCount - 1 do
        begin
          if Parent.Controls[nCtl] is TCHRadioButton then
          begin
            SelfRadio := TCHRadioButton(Parent.Controls[nCtl]);
            if (SelfRadio <> Self) and (SelfRadio.GroupIndex = GroupIndex) then
              SelfRadio.SetChecked(False);
          end;
        end;
      end;
      if csDesigning in ComponentState then
      begin
        if (GetParentForm(self) <> nil) and (GetParentForm(self).Designer <> nil) then
          GetParentForm(self).Designer.Modified;
      end;
    end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
  begin
    if IsUnderlineChar(Caption, CharCode) then
    begin
      if Enabled and Assigned(FOnClick) then
        FOnClick(Self);
      Result := 1;
    end
    else
      inherited;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMMouseEnter(var Message: TMessage);
begin
    if Assigned(FOnMouseEnter) then
    FOnMouseEnter(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMTextChanged(var Msg: TMessage);
begin
  inherited;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.SetChecked(const Value: Boolean);
begin
  if FChecked <> Value then
  begin
    FChecked := Value;

    DoGroup;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.SetColor(const Value: TColor);
begin
  if FColor <> Value then
  begin
    FColor := Value;
    Invalidate;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.SetTextStyle(const Value: TTextStyle);
begin
  with FCaptionLayout do
  begin
    { ssNone }
    if Value = tsNone then
    begin
      ShadowDirection := drNone;
      ShadowDepth := 0;
      ShadowColor := clBtnShadow;

      HighlightDirection := drNone;
      HighlightDepth := 0;
      HighlightColor := clBtnHighlight;
    end
    { ssRaised }
    else if Value = tsRaised then
    begin
      ShadowDirection := drNone;
      ShadowDepth := 0;
      ShadowColor := clBtnShadow;

      HighlightDirection := drUpLeft;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRaisedColor }
    else if Value = tsRaisedColor then
    begin
      ShadowDirection := drDownRight;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;
      Self.Font.Color := clBtnFace;

      HighlightDirection := drUpLeft;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRaisedShadow }
    else if Value = tsRaisedShadow then
    begin
      ShadowDirection := drDownRight;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;

      HighlightDirection := drUpLeft;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRecessed }
    else if Value = tsRecessed then
    begin
      ShadowDirection := drNone;
      ShadowDepth := 0;
      ShadowColor := clBtnShadow;

      HighlightDirection := drDownRight;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRecessedColor }
    else if Value = tsRecessedColor then
    begin
      ShadowDirection := drUpLeft;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;
      Self.Font.Color := clBtnFace;

      HighlightDirection := drDownRight;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssRecessedShadow }
    else if Value = tsRecessedShadow then
    begin
      ShadowDirection := drUpLeft;
      ShadowDepth := 1;
      ShadowColor := clBtnShadow;

      HighlightDirection := drDownRight;
      HighlightDepth := 1;
      HighlightColor := clBtnHighlight;
    end
    { ssShadow }
    else if Value = tsShadow then
    begin
      ShadowDirection := drDownRight;
      ShadowDepth := 2;
      ShadowColor := clBtnShadow;

      HighlightDirection := drNone;
      HighlightDepth := 0;
      HighlightColor := clBtnHighlight;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.UpdateChanges(Sender: TObject);
begin
  if csLoading in ComponentState then
  begin
    Exit;
  end;
  Invalidate;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.DrawFocus;
var
  nWidth, X, Y, PixStep, PosLeft, PosRight, PosTop, PosBottom : Integer;
begin
  // Step
  PixStep := FFocus.Step;

  // Style
  if FFocus.Style = csSolid then
    PixStep := 1;

  // Mode + Pos
  if FFocus.Mode = cmAuto then
  begin
    PosLeft := FCaptionRect.Left - 2;
    PosRight := FCaptionRect.Right + 1;
    PosTop := FCaptionRect.Top - 2;
    PosBottom := FCaptionRect.Bottom;
  end
  else
  begin
    PosLeft := FFocus.PosLeft;
    PosRight := FFocus.PosRight;
    PosTop := FFocus.PosTop;
    PosBottom := FFocus.PosBottom;
  end;

  with Self.Canvas do
  begin
    for nWidth := 0 to FFocus.Width - 1 do
    begin
      // top
      X := PosLeft;
      while X <= PosRight do
      begin
        Pixels[X, PosTop + nWidth] := FFocus.Color;
        Inc(X, PixStep);
      end;
      // bottom
      X := PosLeft;
      while X <= PosRight do
      begin
        Pixels[X, PosBottom - nWidth] := FFocus.Color;
        Inc(X, PixStep);
      end;
      // left
      Y := PosTop;
      while Y <= PosBottom do
      begin
        Pixels[PosLeft + nWidth, Y] := FFocus.Color;
        Inc(Y, PixStep);
      end;
      // right
      Y := PosTop;
      while Y <= PosBottom do
      begin
        Pixels[PosRight - nWidth, Y] := FFocus.Color;
        Inc(Y, PixStep);
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMFocusChanged(var Message: TCMFocusChanged);
begin
  with Message do
  begin
    if (Sender is TCHRadioButton) then
    begin
      if (FFocus.Active = True) then
      begin
        FActive := Sender = Self;
        Invalidate;
      end
      else
        FActive := FDefault;
    end;
  end;

  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.CMExit(var Message: TCMExit);
begin
  with Message do
  begin
    FActive := False;
    Invalidate;
  end;

  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCustomRadioButton.WndProc(var Message: TMessage);
begin
  case Message.Msg of
	  WM_LBUTTONDOWN, WM_RBUTTONDOWN, WM_LBUTTONDBLCLK:
	    if not (csDesigning in ComponentState) and not Focused then
      begin
		    FClicksDisabled := True;
        Windows.SetFocus(Handle);
		    FClicksDisabled := False;
        if not Focused then
          Exit;
      end;
    CN_COMMAND:
      if FClicksDisabled then
        Exit;
  end;

  inherited WndProc(Message);
end;






end.
