unit CHRichedit;

{ ####################################################################################
  TCHRichedit

  Version   :   1.0.0
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TRichedit

  History:
  1.0.0 - 21.07.2002    - First Release

  #################################################################################### }


interface

uses
  Forms, Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi, ComCtrls, Richedit,

  _CHClassGraphic, _CHClassControls, _CHClassFunctions, _CHFunction, _CHTypes;

type
  TCHRichedit = class(TRichedit)
  private
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;

    FMargin : TCHMargin;
    FMarginLeft : TControlCanvas;
    FMarginRight : TControlCanvas;
    FMarginTop : TControlCanvas;
    FMarginBottom : TControlCanvas;

    procedure SetMargin(Left, Right, Top, Bottom : Word);
    //procedure DrawOutput(Memo : TRichedit; var Bmp : TBitmap);

    procedure UpdateChanges(Sender: TObject);
  protected
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure WMPaint(var Message : TWMPaint); message WM_PAINT;

    procedure CreateWnd; override;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;

    property Margin : TCHMargin read FMargin Write FMargin;

  end;

procedure Register;

implementation


procedure Register;
begin
  RegisterComponents('CH Pack', [TCHRichedit]);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.CMMouseEnter(var Message: TMessage);
begin
    if Assigned(FOnMouseEnter) then
    FOnMouseEnter(self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(self);
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHRichedit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csAcceptsControls];
  FMargin := TCHMargin.Create;
  FMargin.OnChange := UpdateChanges;

  FMarginLeft := TControlCanvas.Create;
  FMarginLeft.Control := Self;
  FMarginRight := TControlCanvas.Create;
  FMarginRight.Control := Self;
  FMarginTop := TControlCanvas.Create;
  FMarginTop.Control := Self;
  FMarginBottom := TControlCanvas.Create;
  FMarginBottom.Control := Self;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHRichedit.Destroy;
begin
   FMarginLeft.Free;
   FMarginRight.Free;
   FMarginTop.Free;
   FMarginBottom.Free;
  inherited;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.WMPaint(var Message: TWMPaint);
var
  LeftRect, TopRect, RightRect, BottomRect : TRect;
  LineCount : Integer;
begin
  inherited;
  // Left
  if FMargin.MLeft > 0 then
  begin
    LeftRect := Rect(-1, -1, FMargin.MLeft, Self.Height);
    with FMarginLeft do
    begin
      Brush.Color := FMargin.Color;
      FrameRect(LeftRect);
      InflateRect(LeftRect, -1, -1);
      FrameRect(LeftRect);
      FillRect(LeftRect);
    end;
  end;

  // Top
  if FMargin.MTop > 0 then
  begin
    TopRect := Rect(-1, -1, Self.Width, FMargin.MTop);
    with FMarginTop do
    begin
      Brush.Color := FMargin.Color;
      FrameRect(TopRect);
      InflateRect(TopRect, -1, -1);
      FrameRect(TopRect);
      FillRect(TopRect);
    end;
  end;

  // Right
  if FMargin.MRight > 0 then
  begin
    RightRect := Rect(Self.Width - FMargin.MRight - 4, -1, Self.Width, Self.Height);
    with FMarginRight do
    begin
      Brush.Color := FMargin.Color;
      FrameRect(RightRect);
      InflateRect(RightRect, -1, -1);
      FrameRect(RightRect);
      FillRect(RightRect);
    end;
  end;

  // Bottom
  if FMargin.MBottom > 0 then
  begin
    BottomRect := Rect(-1, Self.Height - FMargin.MBottom - 4, Self.Width, Self.Height);
    with FMarginBottom do
    begin
      Brush.Color := FMargin.Color;
      FrameRect(BottomRect);
      InflateRect(BottomRect, -1, -1);
      FrameRect(BottomRect);
      FillRect(BottomRect);
    end;
  end;

  // Line Margin
  if FMargin.LineWidth > 0 then
  begin
    for LineCount := 0 to FMargin.LineWidth  - 1 do
    begin
      if FMargin.MLeft > 0 then
      begin
        FMarginLeft.Pen.Color := FMargin.LineColor;
        FMarginLeft.MoveTo(LeftRect.Right - LineCount, LeftRect.Top);
        FMarginLeft.LineTo(LeftRect.Right - LineCount, LeftRect.Bottom);
      end;

      if FMargin.MTop > 0 then
      begin
        FMarginTop.Pen.Color := FMargin.LineColor;
        FMarginTop.MoveTo(TopRect.Left, TopRect.Bottom - LineCount);
        FMarginTop.LineTo(TopRect.Right, TopRect.Bottom - LineCount);
      end;

      if FMargin.MRight > 0 then
      begin
        FMarginRight.Pen.Color := FMargin.LineColor;
        FMarginRight.MoveTo(RightRect.Left - LineCount, RightRect.Top);
        FMarginRight.LineTo(RightRect.Left - LineCount, RightRect.Bottom);
      end;

      if FMargin.MBottom > 0 then
      begin
        FMarginBottom.Pen.Color := FMargin.LineColor;
        FMarginBottom.MoveTo(BottomRect.Left, BottomRect.Top - LineCount);
        FMarginBottom.LineTo(BottomRect.Right, BottomRect.Top - LineCount);
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.SetMargin(Left, Right, Top, Bottom: Word);
var
  TextRect : TRect;
begin
  // Pos
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@TextRect));
  TextRect.Left := Left + 1;
  TextRect.Right := TextRect.Right - Right;
  TextRect.Top := Top;
  TextRect.Bottom := TextRect.Bottom - Bottom;
  SendMessage(Handle, EM_SETRECT, 0, LongInt(@TextRect));
  Refresh;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
//procedure TCHRichedit.DrawOutput(Memo: TRichedit; var Bmp: TBitmap);
//var
//  TextRect : TRect;
//  Range : TFormatRange;
//begin
//  Bmp.Width := Memo.Width;
//  Bmp.Height := Memo.Height;
//  Bmp.Canvas.Brush.Color := Memo.Color;
//  Bmp.Canvas.Brush.Style := bsClear;
//
//  TextRect := Rect(0,0,Memo.Width * Screen.PixelsPerInch, Memo.Height * Screen.PixelsPerInch);
//
//  Range.hdc := Bmp.Canvas.Handle;
//  Range.hdcTarget := Bmp.Canvas.Handle;
//  Range.rc := TextRect;
//  Range.rcPage := TextRect;
//  Range.chrg.cpMin := 0;
//  Range.chrg.cpMax := -1;
//  SendMessage(Memo.Handle, EM_FORMATRANGE, 1, Longint(@range));
//  SendMessage(Memo.Handle, EM_FORMATRANGE, 0, 0);
//
//end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.UpdateChanges(Sender: TObject);
begin
  if csLoading in ComponentState then
  begin
    Exit;
  end;
  RecreateWnd;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or WS_CLIPCHILDREN;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.CreateWnd;
begin
  inherited CreateWnd;
  SetMargin(FMargin.MLeft, FMargin.MRight, FMargin.MTop, FMargin.MBottom);

end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHRichedit.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  inherited;
  SetMargin(FMargin.MLeft, FMargin.MRight, FMargin.MTop, FMargin.MBottom);
end;





end.
