unit _CHClassControls;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi,

  _CHFunction, _CHTypes;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ CaptionLayout }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHCaptionLayout = class(TPersistent)
  private
    FAngle : TAngle;
    FCosAngle : Double;
    FSinAngle : Double;
    FTextStyle : TTextStyle;
    FTextAlign : TTextAlign;

    FChangingShadow : Boolean;
    FShadowColor : TColor;
    FShadowDepth : TShadowDepth;
    FShadowDirection : TDirection;

    FChangingHighlight : Boolean;
    FHighlightColor : TColor;
    FHighlightDepth : Integer;
    FHighlightDirection : TDirection;

    FOnChange: TNotifyEvent;
    FAntialiasing: Boolean;
    FPosY: Integer;
    FPosX: Integer;
    procedure SetAntialiasing(const Value: Boolean);
    procedure SetPosX(const Value: Integer);
    procedure SetPosY(const Value: Integer);
  protected
    procedure SetAngle(const Value: TAngle);
    procedure SetShadowDirection(const Value : TDirection);
    procedure SetShadowDepth(const Value : TShadowDepth);
    procedure SetShadowColor(const Value : TColor);
    procedure SetHighlightDirection(const Value : TDirection);
    procedure SetHighlightDepth(const Value : Integer);
    procedure SetHighlightColor(const Value : TColor);
    procedure SetTextStyle(const Value : TTextStyle);
    procedure SetTextAlign(const Value : TTextAlign);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Alignment : TTextAlign read FTextAlign write SetTextAlign;
    property Angle : TAngle read FAngle write SetAngle;
    property Antialiasing : Boolean read FAntialiasing write SetAntialiasing;
    property HighlightDirection : TDirection read FHighlightDirection write SetHighlightDirection;
    property HighlightDepth : Integer read FHighlightDepth write SetHighlightDepth;
    property HighlightColor : TColor read FHighlightColor write SetHighlightColor;
    property PosX : Integer read FPosX Write SetPosX;
    property PosY : Integer read FPosY Write SetPosY;
    property ShadowDirection : TDirection read FShadowDirection write SetShadowDirection;
    property ShadowDepth : TShadowDepth read FShadowDepth write SetShadowDepth;
    property ShadowColor : TColor read FShadowColor write SetShadowColor;
    property TextStyle : TTextStyle read FTextStyle write SetTextStyle;
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ CaptionLayout }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHBorder = class(TPersistent)
  private
    FBorderColor : TColor;
    FBorderHighlight : TColor;
    FBorderShadow : TColor;
    FBorderWidth : Word;
    FBorderStyle : TBorderStyle;
    FBorderSingle : Word;

    FOnChange: TNotifyEvent;
  protected
    procedure SetBorderColor(const Value: TColor);
    procedure SetBorderHighlight(const Value: TColor);
    procedure SetBorderShadow(const Value: TColor);
    procedure SetBorderWidth(const Value: Word);
    procedure SetBorderSingle(const Value: Word);
    procedure SetBorderStyle(const Value : TBorderStyle);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Color : TColor read FBorderColor write SetBorderColor;
    property HighlightColor : TColor read FBorderHighlight write SetBorderHighlight;
    property ShadowColor : TColor read FBorderShadow write SetBorderShadow;
    property Width : Word read FBorderWidth write SetBorderWidth;
    property SingleWidth : Word read FBorderSingle write SetBorderSingle;
    property Style : TBorderStyle read FBorderStyle write SetBorderStyle;
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ KeyPress }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHKeyPress = class(TPersistent)
  private
    FvkEnter, FvkESC, FvkStrg, FvkAlt, FvkShift : Boolean;

    FOnChange: TNotifyEvent;
  protected
    procedure SetKeyPress(Index: Integer; Value : Boolean);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Alt : Boolean Index 0 read FvkAlt Write SetKeyPress;
    property Enter : Boolean Index 1 read FvkEnter Write SetKeyPress;
    property ESC : Boolean Index 2 read FvkESC Write SetKeyPress;
    property Strg : Boolean Index 3 read FvkStrg Write SetKeyPress;
    property Shift : Boolean Index 4 read FvkShift Write SetKeyPress;
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Box (Checkbox) }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHBoxCB = class(TPersistent)
  private
    FBoxStyle : TBoxStyle;
    FBoxColor : TBoxFillColor;
    FBoxFill : TCheckBoxFill;
    FBoxFocusColor : TColor;

    FOnChange: TNotifyEvent;
    procedure SetBoxColor(const Value: TBoxFillColor);
    procedure SetBoxFill(const Value: TCheckBoxFill);
    //procedure SetBoxFocuscolor(const Value: TColor);
    procedure SetBoxStyle(const Value: TBoxStyle);
  protected

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Style : TBoxStyle read FBoxStyle Write SetBoxStyle;
    property Color : TBoxFillColor read FBoxColor Write SetBoxColor;
    property Fill : TCheckBoxFill  read FBoxFill Write SetBoxFill;
    //property Focuscolor : TColor read FBoxFocusColor Write SetBoxFocuscolor;
  end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Box (Radiobutton) }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHBoxRB = class(TPersistent)
  private
    FBoxStyle : TBoxStyle;
    FBoxColor : TBoxFillColor;
    FBoxFill : TRadioButtonFill;
    FBoxFocusColor : TColor;

    FOnChange: TNotifyEvent;
    procedure SetBoxColor(const Value: TBoxFillColor);
    procedure SetBoxFill(const Value: TRadioButtonFill);
    //procedure SetBoxFocuscolor(const Value: TColor);
    procedure SetBoxStyle(const Value: TBoxStyle);
  protected

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Style : TBoxStyle read FBoxStyle Write SetBoxStyle;
    property Color : TBoxFillColor read FBoxColor Write SetBoxColor;
    property Fill : TRadioButtonFill  read FBoxFill Write SetBoxFill;
    //property Focuscolor : TColor read FBoxFocusColor Write SetBoxFocuscolor;
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Focus-Control }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHFocus = class(TPersistent)
  private
    FFocusShow : Boolean;
    FFocusActive : Boolean;
    FFocusStyle : TFocusStyle;
    FFocusColor : TColor;
    FFocusMode : TFocusMode;

    FOnChange: TNotifyEvent;
    FFocusWidth: Integer;
    FFocusPosTop: Integer;
    FFocusPosLeft: Integer;
    FFocusPosRight: Integer;
    FFocusPosBottom: Integer;
    FFocusStep: Integer;
    procedure SetFocusColor(const Value: TColor);
    procedure SetFocusShow(const Value: Boolean);
    procedure SetFocusStyle(const Value: TFocusStyle);
    procedure SetFocusWidth(const Value: Integer);
    procedure SetFocusPosBottom(const Value: Integer);
    procedure SetFocusPosLeft(const Value: Integer);
    procedure SetFocusPosRight(const Value: Integer);
    procedure SetFocusPosTop(const Value: Integer);
    procedure SetFocusActive(const Value: Boolean);
    procedure SetFocusMode(const Value: TFocusMode);
    procedure SetFocusStep(const Value: Integer);
  protected

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Active : Boolean read FFocusActive Write SetFocusActive;
    property Color : TColor read FFocusColor Write SetFocusColor;
    property Mode : TFocusMode read FFocusMode Write SetFocusMode;
    property Show : Boolean read FFocusShow Write SetFocusShow;
    property Step : Integer read FFocusStep Write SetFocusStep;
    property Style : TFocusStyle read FFocusStyle Write SetFocusStyle;
    property PosTop : Integer read FFocusPosTop Write SetFocusPosTop;
    property PosBottom : Integer read FFocusPosBottom Write SetFocusPosBottom;
    property PosLeft : Integer read FFocusPosLeft Write SetFocusPosLeft;
    property PosRight : Integer read FFocusPosRight Write SetFocusPosRight;
    property Width : Integer read FFocusWidth Write SetFocusWidth;
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{Margin }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHMargin = class(TPersistent)
  private
    FOnChange: TNotifyEvent;
    FMarginBottom: Word;
    FMarginTop: Word;
    FMarginLeft: Word;
    FMarginRight: Word;
    FMarginColor: TColor;
    FMarginLineColor: TColor;
    FMarginLineWidth: Word;
    procedure SetMarginBottom(const Value: Word);
    procedure SetMarginLeft(const Value: Word);
    procedure SetMarginRight(const Value: Word);
    procedure SetMarginTop(const Value: Word);
    procedure SetMarginColor(const Value: TColor);
    procedure SetMarginLineColor(const Value: TColor);
    procedure SetMarginLineWidth(const Value: Word);
  protected
    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property MBottom : Word read FMarginBottom Write SetMarginBottom;
    property MLeft : Word read FMarginLeft Write SetMarginLeft;
    property MRight : Word read FMarginRight Write SetMarginRight;
    property MTop : Word read FMarginTop Write SetMarginTop;
    property Color : TColor read FMarginColor Write SetMarginColor;
    property LineColor : TColor read FMarginLineColor Write SetMarginLineColor;
    property LineWidth : Word read FMarginLineWidth Write SetMarginLineWidth;
  end;


implementation


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ C a p t i o n L a y o u t }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHCaptionLayout.Create;
begin
  inherited Create;

  FTextAlign := tgCenter;
  FAngle := 0;
  FHighlightDirection := drNone;
  FHighlightDepth := 0;
  FHighlightColor := clBtnHighlight;
  FShadowDirection := drNone;
  FShadowDepth := 0;
  FShadowColor := clBtnShadow;
  FTextStyle := tsNone;
  FAntialiasing := True;

  FChangingShadow := False;
  FChangingHighlight := False;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetAngle(const Value: TAngle);
begin
  if FAngle <> Value then
  begin
    FAngle := Value;
    FCosAngle := Cos(Value * Pi / 180);
    FSinAngle := Sin(Value * Pi / 180);
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetHighlightColor(const Value: TColor);
begin
  if FHighlightColor <> Value then
  begin
    FHighlightColor := Value;
    FTextStyle := tsCustom;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetHighlightDepth(const Value: Integer);
begin
  if FHighlightDepth <> Value then
  begin
    FHighlightDepth := Value;
    FTextStyle := tsCustom;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetHighlightDirection(const Value: TDirection);
begin
  if FHighlightDirection <> Value then
  begin
    FHighlightDirection := Value;
    FTextStyle := tsCustom;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetShadowColor(const Value: TColor);
begin
  if FShadowColor <> Value then
  begin
    FShadowColor := Value;
    FTextStyle := tsCustom;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetShadowDepth(const Value: TShadowDepth);
begin
  if FShadowDepth <> Value then
  begin
    FShadowDepth := Value;
    FTextStyle := tsCustom;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetShadowDirection(const Value: TDirection);
begin
  if FShadowDirection <> Value then
  begin
    FShadowDirection := Value;
    FTextStyle := tsCustom;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetTextAlign(const Value: TTextAlign);
begin
  if FTextAlign <> Value then
  begin
    FTextAlign := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetTextStyle(const Value: TTextStyle);
begin
  if FTextStyle <> Value then
  begin
    FTextStyle := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetAntialiasing(const Value: Boolean);
begin
  if FAntialiasing <> Value then
  begin
    FAntialiasing := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetPosX(const Value: Integer);
begin
  if FPosX <> Value then
  begin
    FPosX := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHCaptionLayout.SetPosY(const Value: Integer);
begin
  if FPosY <> Value then
  begin
    FPosY := Value;
    DoChange(self);
  end;
end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ B o r d e r }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

procedure TCHBorder.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHBorder.Create;
begin
  inherited Create;

  FBorderColor := clBlack;
  FBorderHighlight := clWhite;
  FBorderShadow := clGray;
  FBorderWidth := 2;
  FBorderSingle := 1;
  FBorderStyle := bsNormal;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.SetBorderColor(const Value: TColor);
begin
  if FBorderColor <> Value then
  begin
    FBorderColor := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.SetBorderHighlight(const Value: TColor);
begin
  if FBorderHighlight <> Value then
  begin
    FBorderHighlight := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.SetBorderShadow(const Value: TColor);
begin
  if FBorderShadow <> Value then
  begin
    FBorderShadow := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.SetBorderSingle(const Value: Word);
begin
  if FBorderSingle <> Value then
  begin
    FBorderSingle := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.SetBorderStyle(const Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBorder.SetBorderWidth(const Value: Word);
begin
  if FBorderWidth <> Value then
  begin
    FBorderWidth := Value;
    DoChange(Self);
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ K e y P r e s s }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHKeyPress.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHKeyPress.Create;
begin
  inherited Create;

  FvkEnter := True;
  FvkESC := True;
  FvkStrg := True;
  FvkAlt := True;
  FvkShift := True;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHKeyPress.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHKeyPress.SetKeyPress(Index: Integer; Value: Boolean);
begin
  case Index of
    0: FvkAlt := Value;
    1: FvkEnter := Value;
    2: FvkESC := Value;
    3: FvkStrg := Value;
    4: FvkShift := Value;
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ C h e c k B o x }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxCB.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHBoxCB.Create;
begin
  inherited Create;

  FBoxColor := bcBlack;
  FBoxFocusColor := clInfoBk;
  FBoxStyle := bxNormal;
  FBoxFill := cfHook;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxCB.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxCB.SetBoxColor(const Value: TBoxFillColor);
begin
  if FBoxColor <> Value then
  begin
    FBoxColor := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxCB.SetBoxFill(const Value: TCheckBoxFill);
begin
  if FBoxFill <> Value then
  begin
    FBoxFill := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
//procedure TCHBoxCB.SetBoxFocuscolor(const Value: TColor);
//begin
//  if FBoxFocusColor <> Value then
//  begin
//    FBoxFocusColor := Value;
//    DoChange(Self);
//  end;
//end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxCB.SetBoxStyle(const Value: TBoxStyle);
begin
  if FBoxStyle <> Value then
  begin
    FBoxStyle := Value;
    DoChange(Self);
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ R a d i o b u t t o n }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

procedure TCHBoxRB.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHBoxRB.Create;
begin
  inherited Create;

  FBoxColor := bcBlack;
  FBoxFocusColor := clInfoBk;
  FBoxStyle := bxNormal;
  FBoxFill := rfDot;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxRB.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxRB.SetBoxColor(const Value: TBoxFillColor);
begin
  if FBoxColor <> Value then
  begin
    FBoxColor := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxRB.SetBoxFill(const Value: TRadioButtonFill);
begin
  if FBoxFill <> Value then
  begin
    FBoxFill := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
//procedure TCHBoxRB.SetBoxFocuscolor(const Value: TColor);
//begin
//  if FBoxFocusColor <> Value then
//  begin
//    FBoxFocusColor := Value;
//    DoChange(Self);
//  end;
//end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBoxRB.SetBoxStyle(const Value: TBoxStyle);
begin
  if FBoxStyle <> Value then
  begin
    FBoxStyle := Value;
    DoChange(Self);
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ F o c u s }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

procedure TCHFocus.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHFocus.Create;
begin
  inherited Create;

  FFocusActive := True;
  FFocusShow := False;
  FFocusColor := clBlack;
  FFocusStyle := csDot;
  FFocusWidth := 1;
  FFocusMode := cmAuto;
  FFocusStep := 2;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusActive(const Value: Boolean);
begin
  if FFocusActive <> Value then
  begin
    FFocusActive := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusColor(const Value: TColor);
begin
  if FFocusColor <> Value then
  begin
    FFocusColor := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusMode(const Value: TFocusMode);
begin
  if FFocusMode <> Value then
  begin
    FFocusMode := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusPosBottom(const Value: Integer);
begin
  FFocusPosBottom := Value;
  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusPosLeft(const Value: Integer);
begin
  FFocusPosLeft := Value;
  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusPosRight(const Value: Integer);
begin
  FFocusPosRight := Value;
  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusPosTop(const Value: Integer);
begin
  FFocusPosTop := Value;
  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusShow(const Value: Boolean);
begin
  if FFocusShow <> Value then
  begin
    FFocusShow := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusStep(const Value: Integer);
begin
  if FFocusStep <> Value then
  begin
    FFocusStep := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusStyle(const Value: TFocusStyle);
begin
  if FFocusStyle <> Value then
  begin
    FFocusStyle := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFocus.SetFocusWidth(const Value: Integer);
begin
  if FFocusWidth <> Value then
  begin
    FFocusWidth := Value;
    DoChange(Self);
  end;
end;





{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ M a r g i n }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

procedure TCHMargin.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHMargin.Create;
begin
  inherited Create;

  FMarginBottom := 0;
  FMarginTop := 0;
  FMarginLeft := 0;
  FMarginRight := 0;
  FMarginColor := clInfoBk;
  FMarginLineColor := clBlack;
  FMarginLineWidth := 0;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginBottom(const Value: Word);
begin
  if FMarginBottom <> Value then
  begin
    FMarginBottom := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginColor(const Value: TColor);
begin
  if FMarginColor <> Value then
  begin
    FMarginColor := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginLeft(const Value: Word);
begin
  if FMarginLeft <> Value then
  begin
    FMarginLeft := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginLineColor(const Value: TColor);
begin
  if FMarginLineColor <> Value then
  begin
    FMarginLineColor := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginLineWidth(const Value: Word);
begin
  if FMarginLineWidth <> Value then
  begin
    FMarginLineWidth := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginRight(const Value: Word);
begin
  if FMarginRight <> Value then
  begin
    FMarginRight := Value;
    DoChange(self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHMargin.SetMarginTop(const Value: Word);
begin
  if FMarginTop <> Value then
  begin
    FMarginTop := Value;
    DoChange(self);
  end;
end;

end.
