unit _CHClassFunctions;

interface

uses Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi, Math,

  _CHClassGraphic, _CHClassControls, _CHFunction, _CHTypes;


procedure DrawGradient_G(Ctl : TGraphicControl; Canvas : TCanvas; Rect: TRect; aColor: array of TColor;
  GradinetStyle : TGradientStyle; GradientRotation : TGradientStep);
procedure DrawGradient_C(Ctl : TCustomControl; Canvas : TCanvas; Rect: TRect; aColor: array of TColor;
  GradinetStyle : TGradientStyle; GradientRotation : TGradientStep);
procedure DrawCaption(Ctl : TCustomControl; Canvas : TCanvas; var CaptionRect, WorkRect, ClientRect : TRect;
  Angle : TAngle; HDepth, SDepth : TShadowDepth; HDirection, SDirection : TDirection; TransMode : TTransMode;
  HColor, SColor, CtlColor, CtlFontColor, TransColor : TColor;
  Text : string; TextStyle : TTextStyle; Alignment : TTextAlign; FillStyle : TFillStyle;
  GlyphAlignMode : TGlyphAlignMode; GlyphAlignment : TGlyphAlignment; Glyph : TBitmap;
  GlyphXPos, GlyphYPos, GlyphSpace, TextXPos, TextYPos, CIdx : Integer;
  Transparent, State, Autosize, Antialiasing : Boolean; Error : TError);
procedure DrawTransparentBmp(Control: TControl; Dest: TCanvas);
procedure TextAlignment(var CaptionRect : TRect; WorkRect : TRect; Alignment : TTextAlign; Size : TSize;
  TextXPos, TextYPos : Integer);
procedure AlignPos(var nGlyph_X, nGlyph_Y : Integer; const XPos, YPos, Space : Integer;
  ButtonRect, ClientRect : TRect;
  Size : TSize; Glyph : TBitmap; GlyphAlignMode : TGlyphAlignMode; GlyphAlignment : TGlyphAlignment);
procedure AutosizeCorrect(var x, y, xTmp, yTmp : Integer; SDirection, HDirection : TDirection;
  SDepth, HDepth : Integer);

implementation


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure DrawGradient_G(Ctl : TGraphicControl; Canvas : TCanvas; Rect: TRect; aColor: array of TColor;
  GradinetStyle : TGradientStyle; GradientRotation : TGradientStep);
var
  SubRow, ColorCount, ColorPart, ColPos, MaxColor, ColorEnd, ColorDivider, PixCount : integer;
  SavPenWidth, nBitCount,  nGradEnd, nHBmpWidth, nVBmpHeight  : Integer;
  I, nX, nDivY, nDivX, nStart, nStep : Integer;
  nFaktor, nGradStep, nStepSum, ColFaktor : Double;

  SavPenStyle : TPenStyle;
  SavPenColor : TColor;

  aSubRGB : TRGBArray;
  aRGBArray : array of TRGBArray;
  HBitmap, VBitmap, TmpHBitmap, TmpVBitmap : TBitmap;
begin
  nBitCount := 0;
  nX := 0;
  nStepSum := 0;
  MaxColor := High(aColor);
  if MaxColor > 0 then
  begin
    with Canvas do
    begin
      Pen.Width := 1;
      Pen.Style := psSolid;
      SavPenWidth := Pen.Width;
      SavPenStyle := Pen.Style;
      SavPenColor := Pen.Color;

      HBitmap := TBitmap.Create;
      HBitmap.Width := Ctl.Width * 4;
      HBitmap.Height := 1;

      VBitmap := TBitmap.Create;
      VBitmap.Width := 1;
      VBitmap.Height := Ctl.Height * 4;

      Setlength(aRGBArray, MaxColor + 1);

      for ColorCount := 0 to MaxColor do
      begin
        aColor[ColorCount] := colortorgb(aColor[ColorCount]);
        aRGBArray[ColorCount][0] := GetRValue(aColor[ColorCount]);
        aRGBArray[ColorCount][1] := GetGValue(aColor[ColorCount]);
        aRGBArray[ColorCount][2] := GetBValue(aColor[ColorCount]);
      end;

      // ******************* Gradient Horizontal **************************
      PixCount := (Rect.Right - Rect.Left);

      ColorDivider := Round(PixCount / MaxColor);
      for ColorCount := 0 to MaxColor - 1 do
      begin
        if ColorCount = MaxColor - 1 then
          ColorEnd := PixCount - ColorCount * ColorDivider - 1
        else
          ColorEnd := ColorDivider;

        for SubRow := 0 to ColorEnd do
        begin
          ColPos := SubRow + ColorCount * ColorDivider;
          ColFaktor := SubRow / ColorEnd;

          for ColorPart := 0 to 3 do
            aSubRGB[ColorPart] := trunc(aRGBArray[ColorCount][ColorPart] +
              ((aRGBArray[ColorCount + 1][ColorPart] - aRGBArray[ColorCount][ColorPart]) * ColFaktor));

          Pen.Color := RGB(aSubRGB[0], aSubRGB[1], aSubRGB[2]);

          // horizontales Bitmap zeichnen
          HBitmap.Canvas.Pixels[ColPos, 0] := Pen.Color;
          HBitmap.Canvas.Pixels[((Ctl.Width * 2) - 1) - ColPos, 0] := Pen.Color;
          HBitmap.Canvas.Pixels[(Ctl.Width * 2) + ColPos, 0] := Pen.Color;
          HBitmap.Canvas.Pixels[(HBitmap.Width - 1) - ColPos, 0] := Pen.Color;
        end;
      end;

      // ******************* Gradient Vertikal **************************
      PixCount := Rect.Bottom - Rect.Top;

      ColorDivider := Round(PixCount / MaxColor);
      for ColorCount := 0 to MaxColor - 1 do
      begin
        if ColorCount = MaxColor - 1 then
          ColorEnd := PixCount - ColorCount * ColorDivider - 1
        else
          ColorEnd := ColorDivider;

        for SubRow := 0 to ColorEnd do
        begin
          ColPos := SubRow + ColorCount * ColorDivider;
          ColFaktor := SubRow / ColorEnd;

          for ColorPart := 0 to 3 do
            aSubRGB[ColorPart] := trunc(aRGBArray[ColorCount][ColorPart] +
              ((aRGBArray[ColorCount + 1][ColorPart] - aRGBArray[ColorCount][ColorPart]) * ColFaktor));

          Pen.Color := RGB(aSubRGB[0], aSubRGB[1], aSubRGB[2]);

          // veticales Bitmap zeichnen
          VBitmap.Canvas.Pixels[0, ColPos] := Pen.Color;
          VBitmap.Canvas.Pixels[0, ((Ctl.Height * 2) - 1) - ColPos] := Pen.Color;
          VBitmap.Canvas.Pixels[0, (Ctl.Height * 2) + ColPos] := Pen.Color;
          VBitmap.Canvas.Pixels[0, (VBitmap.Height - 1) - ColPos] := Pen.Color;

        end;
      end;

      // Gre horizontales Bitmap
      while (Ctl.Height * 2) > HBitmap.Width do
      begin
        nGradEnd := HBitmap.Width;
        HBitmap.Width := HBitmap.Width * 2;
        HBitmap.Canvas.Draw(nGradend, 0, HBitmap);
      end;
      // Gre vertikales Bitmap
      while (Ctl.Width * 2) > VBitmap.Height do
      begin
        nGradEnd := VBitmap.Height;
        VBitmap.Height := VBitmap.Height * 2;
        VBitmap.Canvas.Draw(0, nGradend, VBitmap);
      end;


      nHBmpWidth := Ctl.Width;
      nVBmpHeight := Ctl.Height;

      if (GradinetStyle = gsHorizontal_L) or
        (GradinetStyle = gsHorizontal_R) or
        (GradinetStyle = gsArrow_L) or
        (GradinetStyle = gsArrow_R) or
        (GradinetStyle = gsDiagonal_R) or
        (GradinetStyle = gsDiagonal_L) then
      begin
        nBitCount := Rect.Bottom - Rect.Top;
      end
      else if (GradinetStyle = gsVertical_T) or
        (GradinetStyle = gsVertical_B) or
        (GradinetStyle = gsArrow_T) or
        (GradinetStyle = gsArrow_B) then
      begin
        nBitCount := Rect.Right - Rect.Left;
      end;


      // Control mit Gradientbitmap fllen
      if GradinetStyle <> gsRotation then
      begin
        for I := 0 to nBitCount do
        begin
          // gsHorizontal_L
          if GradinetStyle = gsHorizontal_L then
            Draw(-nHBmpWidth, I, HBitmap)
          // gsHorizontal_R
          else if GradinetStyle = gsHorizontal_R then
            Draw(0, I, HBitmap)
          // gsVertical_T
          else if GradinetStyle = gsVertical_T then
            Draw(I, -nVBmpHeight, VBitmap)
          // gsVertical_B
          else if GradinetStyle = gsVertical_B then
            Draw(I, 0, VBitmap)
          // gsArrow_R
          else if GradinetStyle = gsArrow_R then
          begin
            nDivY := Ctl.Height div 2;
            if I < (nBitCount div 2) then
              Draw(I - nDivY, I, HBitmap)
            else
              Draw(-I + nDivY, I, HBitmap);
          end
          // gsArrow_L
          else if GradinetStyle = gsArrow_L then
          begin
            if I < (nBitCount div 2) then
              Draw(-I, I, HBitmap)
            else
            begin
              Inc(nX);
              Draw(-I + (nX * 2), I, HBitmap);
            end;
          end
          // gsArrow_B
          else if GradinetStyle = gsArrow_B then
          begin
            nDivX := Ctl.Width div 2;
            if I < (nBitCount div 2) then
              Draw(I, I - nDivX, VBitmap)
            else
              Draw(I, -I + nDivX, VBitmap);
          end
          // gsArrow_T
          else if GradinetStyle = gsArrow_T then
          begin
            if I < (nBitCount div 2) then
              Draw(I, -I, VBitmap)
            else
            begin
              Inc(nX);
              Draw(I, -I + (nX * 2), VBitmap);
            end;
          end
          // gsDiagonal_L (Left)
          else if GradinetStyle = gsDiagonal_L then
          begin
            nStep := 0;
            nFaktor := (Ctl.Width / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Width / Ctl.Height) - ((nStep / 100) * (Ctl.Width / Ctl.Height));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Width * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(nX, I, HBitmap)
          end
          // gsDiagonal_R (Right)
          else if GradinetStyle = gsDiagonal_R then
          begin
            nStep := 200;
            nFaktor := (Ctl.Width / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Width / Ctl.Height) - ((nStep / 100) * (Ctl.Width / Ctl.Height));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Width * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(nX, I, HBitmap)
          end
        end;
      end
      // gsCustom
      else if GradinetStyle = gsRotation then
      begin
        if GradientRotation >= 0 then
          nBitCount := Rect.Bottom - Rect.Top
        else
          nBitCount := Rect.Right - Rect.Left;

        for I := 0 to nBitCount do
        begin
          if GradientRotation >= 0 then
          begin
            nStep := GradientRotation * 2;
            nFaktor := (Ctl.Width / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Width / Ctl.Height) - ((nStep / 100) * (Ctl.Width / Ctl.Height));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Width * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(nX, I, HBitmap);
          end
          else
          begin
            nStep := (GradientRotation * 2) * (-1);
            nFaktor := (Ctl.Height / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Height / Ctl.Width) - ((nStep / 100) * (Ctl.Height / Ctl.Width));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Height * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(I, nX, VBitmap);
          end;
        end;
      end;

      VBitmap.Free;
      HBitmap.Free;

      aRGBArray := nil;
      Pen.Width := SavPenWidth;
      Pen.Style := SavPenStyle;
      Pen.Color := SavPenColor;
    end;
  end;
end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure DrawGradient_C(Ctl : TCustomControl; Canvas : TCanvas; Rect: TRect; aColor: array of TColor;
  GradinetStyle : TGradientStyle; GradientRotation : TGradientStep);
var
  SubRow, ColorCount, ColorPart, ColPos, MaxColor, ColorEnd, ColorDivider, PixCount : integer;
  SavPenWidth, nBitCount,  nGradEnd, nHBmpWidth, nVBmpHeight  : Integer;
  I, nX, nDivY, nDivX, nStart, nStep : Integer;
  nFaktor, nGradStep, nStepSum, ColFaktor : Double;

  SavPenStyle : TPenStyle;
  SavPenColor : TColor;

  aSubRGB : TRGBArray;
  aRGBArray : array of TRGBArray;
  HBitmap, VBitmap, TmpHBitmap, TmpVBitmap : TBitmap;
begin
  nBitCount := 0;
  nX := 0;
  nStepSum := 0;
  MaxColor := High(aColor);
  if MaxColor > 0 then
  begin
    with Canvas do
    begin
      Pen.Width := 1;
      Pen.Style := psSolid;
      SavPenWidth := Pen.Width;
      SavPenStyle := Pen.Style;
      SavPenColor := Pen.Color;

      HBitmap := TBitmap.Create;
      HBitmap.Width := Ctl.Width * 4;
      HBitmap.Height := 1;

      VBitmap := TBitmap.Create;
      VBitmap.Width := 1;
      VBitmap.Height := Ctl.Height * 4;

      Setlength(aRGBArray, MaxColor + 1);

      for ColorCount := 0 to MaxColor do
      begin
        aColor[ColorCount] := colortorgb(aColor[ColorCount]);
        aRGBArray[ColorCount][0] := GetRValue(aColor[ColorCount]);
        aRGBArray[ColorCount][1] := GetGValue(aColor[ColorCount]);
        aRGBArray[ColorCount][2] := GetBValue(aColor[ColorCount]);
      end;

      // ******************* Gradient Horizontal **************************
      PixCount := (Rect.Right - Rect.Left);

      ColorDivider := Round(PixCount / MaxColor);
      for ColorCount := 0 to MaxColor - 1 do
      begin
        if ColorCount = MaxColor - 1 then
          ColorEnd := PixCount - ColorCount * ColorDivider - 1
        else
          ColorEnd := ColorDivider;

        for SubRow := 0 to ColorEnd do
        begin
          ColPos := SubRow + ColorCount * ColorDivider;
          ColFaktor := SubRow / ColorEnd;

          for ColorPart := 0 to 3 do
            aSubRGB[ColorPart] := trunc(aRGBArray[ColorCount][ColorPart] +
              ((aRGBArray[ColorCount + 1][ColorPart] - aRGBArray[ColorCount][ColorPart]) * ColFaktor));

          Pen.Color := RGB(aSubRGB[0], aSubRGB[1], aSubRGB[2]);

          // horizontales Bitmap zeichnen
          HBitmap.Canvas.Pixels[ColPos, 0] := Pen.Color;
          HBitmap.Canvas.Pixels[((Ctl.Width * 2) - 1) - ColPos, 0] := Pen.Color;
          HBitmap.Canvas.Pixels[(Ctl.Width * 2) + ColPos, 0] := Pen.Color;
          HBitmap.Canvas.Pixels[(HBitmap.Width - 1) - ColPos, 0] := Pen.Color;
        end;
      end;

      // ******************* Gradient Vertikal **************************
      PixCount := Rect.Bottom - Rect.Top;

      ColorDivider := Round(PixCount / MaxColor);
      for ColorCount := 0 to MaxColor - 1 do
      begin
        if ColorCount = MaxColor - 1 then
          ColorEnd := PixCount - ColorCount * ColorDivider - 1
        else
          ColorEnd := ColorDivider;

        for SubRow := 0 to ColorEnd do
        begin
          ColPos := SubRow + ColorCount * ColorDivider;
          ColFaktor := SubRow / ColorEnd;

          for ColorPart := 0 to 3 do
            aSubRGB[ColorPart] := trunc(aRGBArray[ColorCount][ColorPart] +
              ((aRGBArray[ColorCount + 1][ColorPart] - aRGBArray[ColorCount][ColorPart]) * ColFaktor));

          Pen.Color := RGB(aSubRGB[0], aSubRGB[1], aSubRGB[2]);

          // veticales Bitmap zeichnen
          VBitmap.Canvas.Pixels[0, ColPos] := Pen.Color;
          VBitmap.Canvas.Pixels[0, ((Ctl.Height * 2) - 1) - ColPos] := Pen.Color;
          VBitmap.Canvas.Pixels[0, (Ctl.Height * 2) + ColPos] := Pen.Color;
          VBitmap.Canvas.Pixels[0, (VBitmap.Height - 1) - ColPos] := Pen.Color;

        end;
      end;

      // Gre horizontales Bitmap
      while (Ctl.Height * 2) > HBitmap.Width do
      begin
        nGradEnd := HBitmap.Width;
        HBitmap.Width := HBitmap.Width * 2;
        HBitmap.Canvas.Draw(nGradend, 0, HBitmap);
      end;
      // Gre vertikales Bitmap
      while (Ctl.Width * 2) > VBitmap.Height do
      begin
        nGradEnd := VBitmap.Height;
        VBitmap.Height := VBitmap.Height * 2;
        VBitmap.Canvas.Draw(0, nGradend, VBitmap);
      end;


      nHBmpWidth := Ctl.Width;
      nVBmpHeight := Ctl.Height;

      if (GradinetStyle = gsHorizontal_L) or
        (GradinetStyle = gsHorizontal_R) or
        (GradinetStyle = gsArrow_L) or
        (GradinetStyle = gsArrow_R) or
        (GradinetStyle = gsDiagonal_R) or
        (GradinetStyle = gsDiagonal_L) then
      begin
        nBitCount := Rect.Bottom - Rect.Top;
      end
      else if (GradinetStyle = gsVertical_T) or
        (GradinetStyle = gsVertical_B) or
        (GradinetStyle = gsArrow_T) or
        (GradinetStyle = gsArrow_B) then
      begin
        nBitCount := Rect.Right - Rect.Left;
      end;


      // Control mit Gradientbitmap fllen
      if GradinetStyle <> gsRotation then
      begin
        for I := 0 to nBitCount do
        begin
          // gsHorizontal_L
          if GradinetStyle = gsHorizontal_L then
            Draw(-nHBmpWidth, I, HBitmap)
          // gsHorizontal_R
          else if GradinetStyle = gsHorizontal_R then
            Draw(0, I, HBitmap)
          // gsVertical_T
          else if GradinetStyle = gsVertical_T then
            Draw(I, -nVBmpHeight, VBitmap)
          // gsVertical_B
          else if GradinetStyle = gsVertical_B then
            Draw(I, 0, VBitmap)
          // gsArrow_R
          else if GradinetStyle = gsArrow_R then
          begin
            nDivY := Ctl.Height div 2;
            if I < (nBitCount div 2) then
              Draw(I - nDivY, I, HBitmap)
            else
              Draw(-I + nDivY, I, HBitmap);
          end
          // gsArrow_L
          else if GradinetStyle = gsArrow_L then
          begin
            if I < (nBitCount div 2) then
              Draw(-I, I, HBitmap)
            else
            begin
              Inc(nX);
              Draw(-I + (nX * 2), I, HBitmap);
            end;
          end
          // gsArrow_B
          else if GradinetStyle = gsArrow_B then
          begin
            nDivX := Ctl.Width div 2;
            if I < (nBitCount div 2) then
              Draw(I, I - nDivX, VBitmap)
            else
              Draw(I, -I + nDivX, VBitmap);
          end
          // gsArrow_T
          else if GradinetStyle = gsArrow_T then
          begin
            if I < (nBitCount div 2) then
              Draw(I, -I, VBitmap)
            else
            begin
              Inc(nX);
              Draw(I, -I + (nX * 2), VBitmap);
            end;
          end
          // gsDiagonal_L (Left)
          else if GradinetStyle = gsDiagonal_L then
          begin
            nStep := 0;
            nFaktor := (Ctl.Width / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Width / Ctl.Height) - ((nStep / 100) * (Ctl.Width / Ctl.Height));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Width * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(nX, I, HBitmap)
          end
          // gsDiagonal_R (Right)
          else if GradinetStyle = gsDiagonal_R then
          begin
            nStep := 200;
            nFaktor := (Ctl.Width / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Width / Ctl.Height) - ((nStep / 100) * (Ctl.Width / Ctl.Height));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Width * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(nX, I, HBitmap)
          end
        end;
      end
      // gsCustom
      else if GradinetStyle = gsRotation then
      begin
        if GradientRotation >= 0 then
          nBitCount := Rect.Bottom - Rect.Top
        else
          nBitCount := Rect.Right - Rect.Left;

        for I := 0 to nBitCount do
        begin
          if GradientRotation >= 0 then
          begin
            nStep := GradientRotation * 2;
            nFaktor := (Ctl.Width / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Width / Ctl.Height) - ((nStep / 100) * (Ctl.Width / Ctl.Height));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Width * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(nX, I, HBitmap);
          end
          else
          begin
            nStep := (GradientRotation * 2) * (-1);
            nFaktor := (Ctl.Height / 2) / 100;
            nStart := Round(nStep * nFaktor);
            nGradStep := (Ctl.Height / Ctl.Width) - ((nStep / 100) * (Ctl.Height / Ctl.Width));
            nStepSum := nStepSum + nGradStep;
            nX := (nStart + (Ctl.Height * 2)) * (-1);
            nX := Round(nX - nStepSum);
            Draw(I, nX, VBitmap);
          end;
        end;
      end;

      VBitmap.Free;
      HBitmap.Free;

      aRGBArray := nil;
      Pen.Width := SavPenWidth;
      Pen.Style := SavPenStyle;
      Pen.Color := SavPenColor;
    end;
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ zeichnet Caption }
procedure DrawCaption(Ctl : TCustomControl; Canvas : TCanvas; var CaptionRect, WorkRect, ClientRect : TRect;
  Angle : TAngle; HDepth, SDepth : TShadowDepth; HDirection, SDirection : TDirection; TransMode : TTransMode;
  HColor, SColor, CtlColor, CtlFontColor, TransColor : TColor;
  Text : string; TextStyle : TTextStyle; Alignment : TTextAlign; FillStyle : TFillStyle;
  GlyphAlignMode : TGlyphAlignMode; GlyphAlignment : TGlyphAlignment; Glyph : TBitmap;
  GlyphXPos, GlyphYPos, GlyphSpace, TextXPos, TextYPos, CIdx : Integer;
  Transparent, State, Autosize, Antialiasing : Boolean; Error : TError);
type
  TOffsetDirection = (Off_X, Off_Y);

const
  Offset_Values : array[TDirection, TOffsetDirection] of -1..1 =
  ((0, 0), (0, -1), (0, 1), (-1, 0), (1, 0), (-1, -1), (1, -1), (-1, 1), (1, 1));

var
  nGlyph_X, nGlyph_Y, nAngle, xTmp, yTmp, x, y : Integer;
  xTextLength, yTextLength, xTextHeight, yTextHeight : Integer;
  TempRect : TRect;
  SavFontColor : TColor;
  hSavFont : HFont;
  DC : HDC;
  Size : TSize;
  LogFont : TLogFont;
  PLeftTop, PCap : TPoint;
  Grad : Double;
begin
  SavFontColor := CtlFontColor;
  nGlyph_X := 0;
  nGlyph_Y := 0;
  xTextLength := 0;
  yTextLength := 0;
  xTextHeight := 0;
  yTextHeight := 0;
  PLeftTop.X := 0;
  PLeftTop.Y := 0;
  PCap.X := 0;
  PCap.Y := 0;
  nAngle := 0;
  x := 0;
  y := 0;

  with Canvas do
  begin
    //*****************************
    // +++ Transparent +++
    //*****************************
    if (FillStyle = fsNormal) or (Error = errNoBitmap) then
    begin
      if Transparent then
      begin
        DrawTransparentBmp(Ctl, Canvas);
        CopyRect(WorkRect, canvas, WorkRect);
      end
      else
      begin
        Brush.Color := CtlColor;
        Brush.Style := bsSolid;
        FillRect(WorkRect);
      end;
    end;
    Brush.Style := bsClear;


    DC := GetDC(0);
    hSavFont := SelectObject(DC, Font.Handle);
    GetTextExtentPoint32(DC, PChar(Text), Length(Text), Size);
    SelectObject(DC, hSavFont);
    ReleaseDC(0, DC);

    // TextRect
    CaptionRect.Left := WorkRect.Left;
    CaptionRect.Top := WorkRect.Top;
    CaptionRect.Right := WorkRect.Left + Size.cx;
    CaptionRect.Bottom := WorkRect.Top + Size.cy;


    // RadioButton, CheckBox
    if (CIdx = 4) or (CIdx = 5) then
    begin
      //*****************************
      // +++ Glyph +++
      //*****************************
      if Assigned(Glyph) then
      begin
        // AlignPos ermitteln
        AlignPos(nGlyph_X, nGlyph_Y, GlyphXPos, GlyphYPos, GlyphSpace, CaptionRect, ClientRect, Size,
          Glyph, GlyphAlignMode, GlyphAlignment);

        with Glyph do
        begin
          if TransMode = tmCustom then
            TransparentColor := TransColor
          else
            TransparentColor := Glyph.Canvas.Pixels[0, Glyph.Height - 1];

          Transparent := True;
        end;
        // Glyph zeichnen
        Draw(nGlyph_X, nGlyph_Y, Glyph);
      end;

      //*****************************
      // +++ Aligment +++
      //*****************************
      if Alignment = tgAuto then
      begin
        // Checkbox, RadioButton
        if (CIdx = 4) or (CIdx = 5) then
        begin
          TextXPos := nGlyph_X + Glyph.Width + 4;
          if CaptionRect.Bottom > Glyph.Height then
            TextYPos := nGlyph_Y - ((CaptionRect.Bottom - Glyph.Height) div 2)
          else
            TextYPos := nGlyph_Y + ((Glyph.Height - CaptionRect.Bottom) div 2)
        end;
        // Button
        if (CIdx = 2) then
        begin
          Alignment := tgCenter;
        end;
      end;

      if Autosize = False then
        TextAlignment(CaptionRect, WorkRect, Alignment, Size, TextXPos, TextYPos);
    end
    // Button
    else
    begin
      //*****************************
      // +++ Aligment +++
      //*****************************
      if Alignment = tgAuto then
      begin
        // Checkbox, RadioButton
        if (CIdx = 4) or (CIdx = 5) then
        begin
          TextXPos := nGlyph_X + Glyph.Width + 4;
          if CaptionRect.Bottom > Glyph.Height then
            TextYPos := nGlyph_Y - ((CaptionRect.Bottom - Glyph.Height) div 2)
          else
            TextYPos := nGlyph_Y + ((Glyph.Height - CaptionRect.Bottom) div 2)
        end;
        // BUTTON
        if (CIdx = 2) then
        begin
          Alignment := tgCenter;
        end;
      end;

      if Autosize = False then
        TextAlignment(CaptionRect, WorkRect, Alignment, Size, TextXPos, TextYPos);

      //*****************************
      // +++ Glyph +++
      //*****************************
      if Assigned(Glyph) then
      begin
        // AlignPos ermitteln
        AlignPos(nGlyph_X, nGlyph_Y, GlyphXPos, GlyphYPos, GlyphSpace, CaptionRect, ClientRect, Size,
          Glyph, GlyphAlignMode, GlyphAlignment);

        with Glyph do
        begin
          if TransMode = tmCustom then
            TransparentColor := TransColor
          else
            TransparentColor := Glyph.Canvas.Pixels[0, Glyph.Height - 1];

          Transparent := True;
        end;
        // Glyph zeichnen
        Draw(nGlyph_X, nGlyph_Y, Glyph);
      end;

    end;


    //*****************************
    // +++ Down +++
    //*****************************
    if State then
    begin
      CaptionRect.Left := CaptionRect.Left + 1;
      CaptionRect.Top := CaptionRect.Top + 1;
    end;


    //*****************************
    // +++ Angle +++
    //*****************************
    if Angle >= 360 then
      Angle := 0;

    if (Angle <> 0) then
    begin
      GetObject(Font.Handle, SizeOf(LogFont), Addr(LogFont));
      LogFont.lfEscapement := Angle * 10;
      LogFont.lfOutPrecision := OUT_TT_ONLY_PRECIS;
      if Antialiasing then
        LogFont.lfQuality := ANTIALIASED_QUALITY;

      Font.Handle := CreateFontIndirect(LogFont);

      if (Angle >= 1) and (Angle <= 90) then
      begin
        nAngle := Angle;
        Grad := (Pi * nAngle / 180);
        yTextLength := Round(sin(Grad) * (Size.cx div 2));
        xTextLength := -Round(cos(Grad) * (Size.cx div 2));
        xTextHeight := Round(sin(Grad) * (Size.cy div 2));
        yTextHeight := Round(cos(Grad) * (Size.cy div 2));
      end
      else if (Angle >= 91) and (Angle <= 180) then
      begin
        nAngle := Angle - 90;
        Grad := (Pi * nAngle / 180);
        yTextLength := Round(cos(Grad) * (Size.cx div 2));
        xTextLength := Round(sin(Grad) * (Size.cx div 2));
        xTextHeight := Round(cos(Grad) * (Size.cy div 2));
        yTextHeight := -Round(sin(Grad) * (Size.cy div 2));
      end
      else if (Angle >= 181) and (Angle <= 270) then
      begin
        nAngle := Angle - 180;
        Grad := (Pi * nAngle / 180);
        yTextLength := -Round(sin(Grad) * (Size.cx div 2));
        xTextLength := Round(cos(Grad) * (Size.cx div 2));
        xTextHeight := -Round(sin(Grad) * (Size.cy div 2));
        yTextHeight := -Round(cos(Grad) * (Size.cy div 2));
      end
      else if (Angle >= 271) and (Angle <= 359) then
      begin
        nAngle := Angle - 270;
        Grad := (Pi * nAngle / 180);
        yTextLength := -Round(cos(Grad) * (Size.cx div 2));
        xTextLength := -Round(sin(Grad) * (Size.cx div 2));
        xTextHeight := -Round(cos(Grad) * (Size.cy div 2));
        yTextHeight := Round(sin(Grad) * (Size.cy div 2));
      end;

      PLeftTop.X := CaptionRect.Left + (xTextLength - xTextHeight) + (Size.cx div 2);
      PLeftTop.Y := CaptionRect.Top + (yTextLength - yTextHeight) + (Size.cy div 2);
    end
    else
    begin
      GetObject(Font.Handle, SizeOf(LogFont), Addr(LogFont));
      LogFont.lfEscapement := 0;
      if Antialiasing then
        LogFont.lfQuality := ANTIALIASED_QUALITY;

      Font.Handle := CreateFontIndirect(LogFont);

      PLeftTop.X := CaptionRect.Left;
      PLeftTop.Y := CaptionRect.Top;
    end;

    // Autosize
    if Autosize then
    begin
      // Mae fr Zeichenbereich (Control) festlegen
      x := Abs(Trunc(Size.cx * cos(nAngle * Pi / 180))) +
        Abs(Trunc(Size.cy * sin(nAngle * Pi / 180)));
      y := Abs(Trunc(Size.cx * sin(nAngle * Pi / 180))) +
        Abs(Trunc(Size.cy * cos(nAngle * Pi / 180)));

      AutosizeCorrect(x, y, xTmp, yTmp, SDirection, HDirection, SDepth, HDepth);

      Ctl.Width := x;
      Ctl.Height := y;
      PLeftTop.X := xTmp;
      PLeftTop.Y := yTmp;
    end;


    //*****************************
    // +++ Enable +++
    //*****************************
    if Ctl.Enabled = False then
    begin
      PCap.X := PLeftTop.X + 1;
      PCap.Y := PLeftTop.Y + 1;
      Font.Color := clBtnHighlight;
      TextOut(PCap.X, PCap.Y, Text);

      PCap.X := PLeftTop.X - 1;
      PCap.Y := PLeftTop.Y - 1;
      Font.Color := clBtnShadow;
      TextOut(PCap.X, PCap.Y, Text);
    end
    else
    begin
      //***************
      // E F F E K T E
      //***************
      if TextStyle <> tsNone then
      begin
        // Pos. CaptionRect speichern
        TempRect := CaptionRect;

        // Pos. fr Rect-Shadow festlegen
        PCap.X := PLeftTop.X + (SDepth * Offset_Values[SDirection, Off_X]);
        PCap.Y := PLeftTop.Y + (SDepth * Offset_Values[SDirection, Off_Y]);
        // Font-Color fr Shadow setzen
        Font.Color := SColor;
        // Shadow-Text zeichnen
        TextOut(PCap.X, PCap.Y, Text);

        CaptionRect := TempRect;
        // Pos. fr Rect-Highlight festlegen
        PCap.X := PLeftTop.X + (HDepth * Offset_Values[HDirection, Off_X]);
        PCap.Y := PLeftTop.Y + (HDepth * Offset_Values[HDirection, Off_Y]);
        // Font-Color fr Highlight setzen
        Font.Color := HColor;
        // Highlight-Text zeichnen
        TextOut(PCap.X, PCap.Y, Text);

        // Pos. Captionrect restaurieren
        CaptionRect := TempRect;
      end;
      // Control-Font setzen
      Font.Color := SavFontColor;
      // Control-Text zeichen
      TextOut(PLeftTop.X, PLeftTop.Y, Text);
    end;
  end;
end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure DrawTransparentBmp(Control: TControl; Dest: TCanvas);
var
  SaveIndex: Integer;
  DC: HDC;
  Position: TPoint;
begin
  with Control do
  begin
    if Parent = nil then
      Exit;
    DC := Dest.Handle;
    SaveIndex := SaveDC(DC);
    GetViewportOrgEx(DC, Position);
    SetViewportOrgEx(DC, Position.X - Left, Position.Y - Top, nil);
    IntersectClipRect(DC, 0, 0, Parent.ClientWidth, Parent.ClientHeight);
    Parent.Perform(WM_ERASEBKGND, DC, 0);
    Parent.Perform(WM_PAINT, DC, 0);
    RestoreDC(DC, SaveIndex);
  end;
end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TextAlignment(var CaptionRect : TRect; WorkRect : TRect; Alignment : TTextAlign; Size : TSize;
  TextXPos, TextYPos : Integer);
var
  nWorkWidth, nWorkHeight, nCaptionWidth, nCaptionHeight : Integer;
begin
    nWorkWidth := WorkRect.Right - WorkRect.Left;
    nWorkHeight := WorkRect.Bottom - WorkRect.Top;
    nCaptionWidth := CaptionRect.Right - CaptionRect.Left;
    nCaptionHeight := CaptionRect.Bottom - CaptionRect.Top;

    if (Alignment = tgCustom) or (Alignment = tgAuto) then
    begin
      CaptionRect.Left := TextXPos;
      CaptionRect.Top := TextYPos;
    end
    else if Alignment = tgCenter then
    begin
      CaptionRect.Left := WorkRect.Left + Round((nWorkWidth / 2) - (nCaptionWidth / 2));
      CaptionRect.Top := WorkRect.Top +  Round((nWorkHeight / 2) - (nCaptionHeight / 2));
    end
    else if Alignment = tgCenterLeft then
    begin
      CaptionRect.Left := WorkRect.Left;
      CaptionRect.Top := WorkRect.Top + Round((nWorkHeight / 2) - (nCaptionHeight / 2));
    end
    else if Alignment = tgCenterRight then
    begin
      CaptionRect.Left := (WorkRect.Right - nCaptionWidth);
      CaptionRect.Top := WorkRect.Top + Round((nWorkHeight / 2) - (nCaptionHeight / 2));
    end
    else if Alignment = tgCenterTop then
    begin
      CaptionRect.Left := WorkRect.Left + Round((nWorkWidth / 2) - (nCaptionWidth / 2));
      CaptionRect.Top := WorkRect.Top;
    end
    else if Alignment = tgCenterBottom then
    begin
      CaptionRect.Left := WorkRect.Left + Round((nWorkWidth / 2) - (nCaptionWidth / 2));
      CaptionRect.Top := (WorkRect.Bottom - nCaptionHeight);
    end
    else if Alignment = tgTopLeft then
    begin
      CaptionRect.Left := WorkRect.Left;
      CaptionRect.Top := WorkRect.Top;
    end
    else if Alignment = tgTopRight then
    begin
      CaptionRect.Left := (WorkRect.Right - nCaptionWidth);
      CaptionRect.Top := WorkRect.Top;
    end
    else if Alignment = tgBottomLeft then
    begin
      CaptionRect.Left := WorkRect.Left;
      CaptionRect.Top := (WorkRect.Bottom - nCaptionHeight);
    end
    else if Alignment = tgBottomRight then
    begin
      CaptionRect.Left := (WorkRect.Right - nCaptionWidth);
      CaptionRect.Top := (WorkRect.Bottom - nCaptionHeight);
    end;

    CaptionRect.Right := CaptionRect.Left + nCaptionWidth;
    CaptionRect.Bottom := CaptionRect.Top + nCaptionHeight;

end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure AlignPos(var nGlyph_X, nGlyph_Y : Integer; const XPos, YPos, Space : Integer;
  ButtonRect, ClientRect : TRect;
  Size : TSize; Glyph : TBitmap; GlyphAlignMode : TGlyphAlignMode; GlyphAlignment : TGlyphAlignment);
begin
  // Align Mode "gmCaption"
    if GlyphAlignMode = gmCaption then
    begin
      if GlyphAlignment = gaLeft then
      begin
        nGlyph_X := ButtonRect.Left - Glyph.Width - Space;
        nGlyph_Y := ButtonRect.Top + Round((Size.cy / 2) - (Glyph.Height / 2));
      end
      else if GlyphAlignment = gaRight then
      begin
        nGlyph_X := ButtonRect.Left + Size.cx + Space;
        nGlyph_Y := ButtonRect.Top + Round((Size.cy / 2) - (Glyph.Height / 2));
      end
      else if GlyphAlignment = gaTop then
      begin
        nGlyph_X := ButtonRect.Left + Round((Size.cx / 2) - (Glyph.Width / 2));
        nGlyph_Y := ButtonRect.Top - Glyph.Height - Space;
      end
      else if GlyphAlignment = gaBottom then
      begin
        nGlyph_X := ButtonRect.Left + Round((Size.cx / 2) - (Glyph.Width / 2));
        nGlyph_Y := ButtonRect.Top + Size.cy + Space;
      end
      else if GlyphAlignment = gaCenter then
      begin
        nGlyph_X := ButtonRect.Left + Round((Size.cx / 2) - (Glyph.Width / 2));
        nGlyph_Y := ButtonRect.Top + Round((Size.cy / 2) - (Glyph.Height / 2));
      end;
    end
    // Align Mode "gmControl"
    else if GlyphAlignMode = gmControl then
    begin
      if GlyphAlignment = gaLeft then
      begin
        nGlyph_X := ClientRect.Left;
        nGlyph_Y := Round((ClientRect.Bottom / 2) - (Glyph.Height / 2));
      end
      else if GlyphAlignment = gaRight then
      begin
        nGlyph_X := ClientRect.Right - Glyph.Width;
        nGlyph_Y := Round((ClientRect.Bottom / 2) - (Glyph.Height / 2));
      end
      else if GlyphAlignment = gaTop then
      begin
        nGlyph_X := Round((ClientRect.Right / 2) - (Glyph.Width / 2));
        nGlyph_Y := ClientRect.Top;
      end
      else if GlyphAlignment = gaBottom then
      begin
        nGlyph_X := Round((ClientRect.Right / 2) - (Glyph.Width / 2));
        nGlyph_Y := ClientRect.Bottom - Glyph.Height;
      end
      else if GlyphAlignment = gaCenter then
      begin
        nGlyph_X := Round((ClientRect.Right / 2) - (Glyph.Width / 2));
        nGlyph_Y := Round((ClientRect.Bottom / 2) - (Glyph.Height / 2));
      end;
    end
    // Align Mode "gmCustom"
    else if GlyphAlignMode = gmCustom then
    begin
      nGlyph_X := XPos;
      nGlyph_Y := YPos;
    end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Korrigiert Autosize, damit auch Texteffekte angezeigt werden }
procedure AutosizeCorrect(var x, y, xTmp, yTmp : Integer; SDirection, HDirection : TDirection;
  SDepth, HDepth : Integer);
var
  nHighlight_Y, nShadow_Y, nHighlight_X, nShadow_X, nComp_X, nComp_Y : Integer;
begin
  // Mae von Autosize korrigieren, wenn Effekt
  if (SDirection <> drNone) or (HDirection <> drNone) then
  begin
    // B R E I T E
    nHighlight_Y := 0;
    nShadow_Y := 0;
    // Shadow Y wenn Up
    if (SDirection = drUp) or (SDirection = drUpLeft) or (SDirection = drUpRight) then
      nShadow_Y := SDepth;
    // Highlight Y wenn Up
    if (HDirection = drUp) or (HDirection = drUpLeft) or (HDirection = drUpRight) then
      nHighlight_Y := HDepth;
    // Y-Korrektur-Wert wenn xxUp
    nComp_Y := GetLargerInt(nShadow_Y, nHighlight_Y);
    yTmp := nComp_Y;

    nHighlight_Y := 0;
    nShadow_Y := 0;
    // Shadow Y wenn Down
    if (SDirection = drDown) or (SDirection = drDownLeft) or (SDirection = drDownRight) then
      nShadow_Y := SDepth;
    // Highlight Y wenn Down
    if (HDirection = drDown) or (HDirection = drDownLeft) or (HDirection = drDownRight) then
      nHighlight_Y := HDepth;

    // Y-Korrektur-Wert wenn xxDown
    nComp_Y := nComp_Y + GetLargerInt(nShadow_Y, nHighlight_Y);
    y := y + nComp_Y;


    // H  H E
    nHighlight_X := 0;
    nShadow_X := 0;
    // Shadow X wenn Left
    if (SDirection = drLeft) or (SDirection = drUpLeft) or (SDirection = drDownLeft) then
      nShadow_X := SDepth;
    // Highlight X wenn Left
    if (HDirection = drLeft) or (HDirection = drUpLeft) or (HDirection = drDownLeft) then
      nHighlight_X := HDepth;
    // X-Korrektur-Wert wenn xxLeft
    nComp_X := GetLargerInt(nShadow_X, nHighlight_X);
    xTmp:= nComp_X;

    nHighlight_X := 0;
    nShadow_X := 0;
    // Shadow X wenn Right
    if (SDirection = drRight) or (SDirection = drUpRight) or (SDirection = drDownRight) then
      nShadow_X := SDepth;
    // Highlight X wenn Right
    if (HDirection = drRight) or (HDirection = drUpRight) or (HDirection = drDownRight) then
      nHighlight_X := HDepth;

    // X-Korrektur-Wert wenn xxRight
    nComp_X := nComp_X + GetLargerInt(nShadow_X, nHighlight_X);
    x := x + nComp_X;
  end;
end;

end.
