unit _CHClassGraphic;

interface

uses Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi,

  _CHFunction, _CHTypes;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Gradient }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHGradient = class(TPersistent)
  private
    FColor0, FColor1, FColor2, FColor3, FColor4, FColor5, FColor6 : TColor;
    FColor7, FColor8, FColor9 : TColor;

    FActive0, FActive1, FActive2, FActive3, FActive4, FActive5, FActive6: Boolean;
    FActive7, FActive8, FActive9 : Boolean;

    FColorArray : array[0..9] of TColor;
    FActiveArray : array[0..9] of Boolean;

    FGradientStyle : TGradientStyle;
    FRotation : TGradientStep;
    FCustomDiagonal : Boolean;

    FOnChange: TNotifyEvent;

  protected
    procedure SetColor(Index: Integer; const Value: TColor);
    procedure SetActive(Index: Integer; Value : Boolean);
    procedure SetGradientStyle(const Value: TGradientStyle);
    procedure SetRotation(const Value: TGradientStep);
    procedure InitColorArray;
    procedure InitActiveArray;
    procedure SetGradientColorArray;
    procedure DoChange(Sender: TObject); dynamic;
  public
    FGradientColorArray : array of TColor;

    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Active0 : Boolean Index 0 read FActive0 Write SetActive;
    property Active1 : Boolean Index 1 read FActive1 Write SetActive;
    property Active2 : Boolean Index 2 read FActive2 Write SetActive;
    property Active3 : Boolean Index 3 read FActive3 Write SetActive;
    property Active4 : Boolean Index 4 read FActive4 Write SetActive;
    property Active5 : Boolean Index 5 read FActive5 Write SetActive;
    property Active6 : Boolean Index 6 read FActive6 Write SetActive;
    property Active7 : Boolean Index 7 read FActive7 Write SetActive;
    property Active8 : Boolean Index 8 read FActive8 Write SetActive;
    property Active9 : Boolean Index 9 read FActive9 Write SetActive;

    property Color0 : TColor Index 0 read FColor0 write SetColor;
    property Color1 : TColor Index 1 read FColor1 write SetColor;
    property Color2 : TColor Index 2 read FColor2 write SetColor;
    property Color3 : TColor Index 3 read FColor3 write SetColor;
    property Color4 : TColor Index 4 read FColor4 write SetColor;
    property Color5 : TColor Index 5 read FColor5 write SetColor;
    property Color6 : TColor Index 6 read FColor6 write SetColor;
    property Color7 : TColor Index 7 read FColor7 write SetColor;
    property Color8 : TColor Index 8 read FColor8 write SetColor;
    property Color9 : TColor Index 9 read FColor9 write SetColor;

    property Style : TGradientStyle read FGradientStyle write SetGradientStyle;
    property Rotation: TGradientStep read FRotation write SetRotation;
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Glyph }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHGlyph = class(TPersistent)
  private
    FPosX : Integer;
    FPosY : Integer;
    FGlyphSpace: Integer;
    FTransColor: TColor;
    FGlyphBmp: TBitmap;
    FTransMode: TTransMode;
    FGlyphAlignment: TGlyphAlignment;
    FGlyphAlignMode : TGlyphAlignMode;

    FOnChange: TNotifyEvent;
  protected
    procedure SetGlyphAlignment(const Value: TGlyphAlignment);
    procedure SetGlyphAlignMode(const Value: TGlyphAlignMode);
    procedure SetGlyph(const Value: TBitmap);
    procedure SetGlyphSpace(const Value : Integer);
    procedure SetTransColor(const Value : TColor);
    procedure SetTransMode(const Value : TTransMode);
    procedure SetPosX(const Value : Integer);
    procedure SetPosY(const Value : Integer);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Alignment : TGlyphAlignment read FGlyphAlignment write SetGlyphAlignment;
    property AlignMode : TGlyphAlignMode read FGlyphAlignMode Write SetGlyphAlignMode;
    property Glyph: TBitmap read FGlyphBmp write SetGlyph;
    property PosX : Integer read FPosX Write SetPosX;
    property PosY : Integer read FPosY Write SetPosY;
    property Space: Integer read FGlyphSpace write SetGlyphSpace;
    property TransparentColor : TColor read FTransColor Write SetTransColor;
    property TransparentMode : TTransMode read FTransMode Write SetTransMode;    
  end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Bitmap }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHBitmap = class(TPersistent)
  private
    FForegroundBmp: TBitmap;
    FBmpMode: TBitmapMode;

    FOnChange: TNotifyEvent;
  protected
    procedure SetForegroundBmp(const Value : TBitmap);
    procedure SetBmpMode(const Value : TBitmapMode);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Bitmap : TBitmap read FForegroundBmp Write SetForegroundBmp;
    property Mode : TBitmapMode read FBmpMode Write SetBmpMode;
  end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Fill }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHFill = class(TPersistent)
  private
     FFillStyle : TFillStyle;
     FTransparent : Boolean;

     FOnChange: TNotifyEvent;
  protected
    procedure SetFillStyle(const Value : TFillStyle);
    procedure SetTransparent(const Value : Boolean);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Style : TFillStyle read FFillStyle Write SetFillStyle;
    property Transparent : Boolean read FTransparent Write SetTransparent;
  end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Fill B}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
type
  TCHFillB = class(TPersistent)
  private
     FTransparent : Boolean;

     FOnChange: TNotifyEvent;
  protected
    procedure SetTransparent(const Value : Boolean);

    procedure DoChange(Sender: TObject); dynamic;
  public
    constructor Create; virtual;
    procedure Assign(Source: TPersistent); override;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Transparent : Boolean read FTransparent Write SetTransparent;
  end;

implementation


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ G r a d i e n t }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHGradient.Create;
begin
  inherited Create;

  InitActiveArray;
  InitColorArray;

  FColor0 := clGreen;
  FColor1 := clRed;
  FColor2 := clBlue;
  FColor3 := clYellow;
  FColor4 := clOlive;
  FColor5 := clWhite;
  FColor6 := clBlack;
  FColor7 := clGray;
  FColor8 := clLime;
  FColor9 := clPurple;
  FColorArray[0] := FColor0;
  FColorArray[1] := FColor1;
  FColorArray[2] := FColor2;
  FColorArray[3] := FColor3;
  FColorArray[4] := FColor4;
  FColorArray[5] := FColor5;
  FColorArray[6] := FColor6;
  FColorArray[7] := FColor7;
  FColorArray[8] := FColor8;
  FColorArray[9] := FColor9;

  FActive0 := True;
  FActive1 := True;
  FActive2 := True;
  FActiveArray[0] := FActive0;
  FActiveArray[1] := FActive1;
  FActiveArray[2] := FActive2;

  SetGradientColorArray;

  FGradientStyle := gsHorizontal_L;
  FRotation := 100;
  FCustomDiagonal := False;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Farbwerte setzen }
procedure TCHGradient.SetColor(Index: Integer; const Value: TColor);
begin
  case Index of
    0: FColor0 := Value;
    1: FColor1 := Value;
    2: FColor2 := Value;
    3: FColor3 := Value;
    4: FColor4 := Value;
    5: FColor5 := Value;
    6: FColor6 := Value;
    7: FColor7 := Value;
    8: FColor8 := Value;
    9: FColor9 := Value;
  end;

  FColorArray[Index] := Value;
  SetGradientColorArray;

  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ Farben aktivieren/deaktivieren }
procedure TCHGradient.SetActive(Index: Integer; Value: Boolean);
begin
  case Index of
    0: FActive0 := Value;
    1: FActive1 := Value;
    2: FActive2 := Value;
    3: FActive3 := Value;
    4: FActive4 := Value;
    5: FActive5 := Value;
    6: FActive6 := Value;
    7: FActive7 := Value;
    8: FActive8 := Value;
    9: FActive9 := Value;
  end;

  FActiveArray[Index] := Value;
  SetGradientColorArray;

  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.InitActiveArray;
var
  I : Integer;
begin
  for I := 0 to 9 do
  begin
    FActiveArray[I] := False;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.InitColorArray;
var
  I : Integer;
begin
  for I := 0 to 9 do
  begin
    FColorArray[I] := -1;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.SetGradientColorArray;
var
  I, Next : Integer;
begin
  FGradientColorArray := nil;
  Next := 1;
  for I := 0 to 9 do
  begin
    if FActiveArray[I] = True then
    begin
      SetLength(FGradientColorArray, Next);
      FGradientColorArray[Next -1] := FColorArray[I];
      Inc(Next);
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.SetGradientStyle(const Value: TGradientStyle);
begin
  if FGradientStyle <> Value then
  begin
    FGradientStyle := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGradient.SetRotation(const Value: TGradientStep);
begin
  if FRotation <> Value then
  begin
    FRotation := Value;
    DoChange(Self);
  end;
end;




{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ G l y p h }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHGlyph.Create;
begin
  inherited Create;

  FGlyphBmp := TBitmap.Create;
  FGlyphBmp.Width := 0;
  FGlyphBmp.Height := 0;
  FPosX := 0;
  FPosY := 0;
  FGlyphSpace := 5;
  FTransColor := clWhite;
  FTransMode := tmAuto;
  FGlyphAlignment := gaLeft;
  FGlyphAlignMode := gmCaption;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHGlyph.Destroy;
begin
  FGlyphBmp.Free;
  inherited Destroy;
end;

procedure TCHGlyph.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetGlyph(const Value: TBitmap);
begin
  FGlyphBmp.Assign(Value);
  DoChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetGlyphAlignment(const Value: TGlyphAlignment);
begin
  if FGlyphAlignment <> Value then
  begin
    FGlyphAlignment := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetGlyphAlignMode(const Value: TGlyphAlignMode);
begin
  if FGlyphAlignMode <> Value then
  begin
    FGlyphAlignMode := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetGlyphSpace(const Value: Integer);
begin
  if FGlyphSpace <> Value then
  begin
    FGlyphSpace:=Value;
    if FGlyphSpace < 0 then
      FGlyphSpace := 0;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetPosX(const Value: Integer);
begin
  if FPosX <> Value then
  begin
    FPosX := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetPosY(const Value: Integer);
begin
  if FPosY <> Value then
  begin
    FPosY := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetTransColor(const Value: TColor);
begin
  if FTransColor <> Value then
  begin
    FTransColor := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHGlyph.SetTransMode(const Value: TTransMode);
begin
  if FTransMode <> Value then
  begin
    FTransMode := Value;
    DoChange(Self);
  end;
end;

{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ B i t m a p }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBitmap.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHBitmap.Create;
begin
  inherited Create;

  FForegroundBmp := TBitmap.Create;
  FForegroundBmp.Width := 0;
  FForegroundBmp.Height := 0;
  FBmpMode := bmNormal;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHBitmap.Destroy;
begin
  FForegroundBmp.Free;
  inherited Destroy;
end;

procedure TCHBitmap.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBitmap.SetBmpMode(const Value: TBitmapMode);
begin
  if FBmpMode <> Value then
  begin
    FBmpMode := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHBitmap.SetForegroundBmp(const Value: TBitmap);
begin
  FForegroundBmp.Assign(Value);
  DoChange(Self);
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ F i l l }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFill.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHFill.Create;
begin
  inherited Create;

  FFillStyle := fsNormal;
  FTransparent := False;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFill.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFill.SetFillStyle(const Value: TFillStyle);
begin
  if FFillStyle <> Value then
  begin
    FFillStyle := Value;
    DoChange(Self);
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFill.SetTransparent(const Value: Boolean);
begin
  if FTransparent <> Value then
  begin
    FTransparent := Value;
    DoChange(Self);
  end;
end;


{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ F i l l  B}
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
{ +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHFillB.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
end;

constructor TCHFillB.Create;
begin
  inherited Create;

  FTransparent := False;
end;

procedure TCHFillB.DoChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TCHFillB.SetTransparent(const Value: Boolean);
begin
  if FTransparent <> Value then
  begin
    FTransparent := Value;
    DoChange(Self);
  end;
end;



end.
